/*
 * This file is part of Saenaru.
 *
 * Copyright (c) 1990-1998 Microsoft Corporation.
 * Copyright (c) 2006 Hye-Shik Chang <perky@i18n.org>.
 * Copyright (c) 2006 Won-Kyu Park <wkpark@kldp.org>.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Saenaru: saenaru/src/hanja.c,v 1.1 2006/10/24 11:32:49 wkpark Exp $
 */

#include <windows.h>

struct _ConvTbl 
{
  UINT key;
  UINT code;
};

typedef struct _ConvTbl ConvTbl;

/* conversion table for the KS X 1001 + KS X 1002 hanja characters */
static const ConvTbl hanja_to_hangul[] = {
 { 0x4E00, 0xC77C },
 { 0x4E01, 0xC815 },
 { 0x4E03, 0xCE60 },
 { 0x4E07, 0xB9CC },
 { 0x4E08, 0xC7A5 },
 { 0x4E09, 0xC0BC },
 { 0x4E0A, 0xC0C1 },
 { 0x4E0B, 0xD558 },
 { 0x4E0D, 0xBD80 },
 { 0x4E11, 0xCD95 },
 { 0x4E14, 0xCC28 },
 { 0x4E15, 0xBE44 },
 { 0x4E16, 0xC138 },
 { 0x4E18, 0xAD6C },
 { 0x4E19, 0xBCD1 },
 { 0x4E1E, 0xC2B9 },
 { 0x4E2D, 0xC911 },
 { 0x4E32, 0xACF6 },
 { 0x4E38, 0xD658 },
 { 0x4E39, 0xB2E8 },
 { 0x4E3B, 0xC8FC },
 { 0x4E42, 0xC608 },
 { 0x4E43, 0xB0B4 },
 { 0x4E45, 0xAD6C },
 { 0x4E4B, 0xC9C0 },
 { 0x4E4D, 0xC0AC },
 { 0x4E4E, 0xD638 },
 { 0x4E4F, 0xD54D },
 { 0x4E56, 0xAD34 },
 { 0x4E58, 0xC2B9 },
 { 0x4E59, 0xC744 },
 { 0x4E5D, 0xAD6C },
 { 0x4E5E, 0xAC78 },
 { 0x4E5F, 0xC57C },
 { 0x4E6B, 0xAC08 },
 { 0x4E6D, 0xB3CC },
 { 0x4E73, 0xC720 },
 { 0x4E76, 0xBCFC },
 { 0x4E77, 0xC0B4 },
 { 0x4E7E, 0xAC74 },
 { 0x4E82, 0xB09C },
 { 0x4E86, 0xB8CC },
 { 0x4E88, 0xC5EC },
 { 0x4E8B, 0xC0AC },
 { 0x4E8C, 0xC774 },
 { 0x4E8E, 0xC6B0 },
 { 0x4E90, 0xC6B8 },
 { 0x4E91, 0xC6B4 },
 { 0x4E92, 0xD638 },
 { 0x4E94, 0xC624 },
 { 0x4E95, 0xC815 },
 { 0x4E98, 0xAE0D },
 { 0x4E9B, 0xC0AC },
 { 0x4E9E, 0xC544 },
 { 0x4EA1, 0xB9DD },
 { 0x4EA2, 0xD56D },
 { 0x4EA4, 0xAD50 },
 { 0x4EA5, 0xD574 },
 { 0x4EA6, 0xC5ED },
 { 0x4EA8, 0xD615 },
 { 0x4EAB, 0xD5A5 },
 { 0x4EAC, 0xACBD },
 { 0x4EAD, 0xC815 },
 { 0x4EAE, 0xB7C9 },
 { 0x4EB6, 0xB2E8 },
 { 0x4EBA, 0xC778 },
 { 0x4EC0, 0xC2ED },
 { 0x4EC1, 0xC778 },
 { 0x4EC4, 0xCE21 },
 { 0x4EC7, 0xAD6C },
 { 0x4ECA, 0xAE08 },
 { 0x4ECB, 0xAC1C },
 { 0x4ECD, 0xC789 },
 { 0x4ED4, 0xC790 },
 { 0x4ED5, 0xC0AC },
 { 0x4ED6, 0xD0C0 },
 { 0x4ED7, 0xC7A5 },
 { 0x4ED8, 0xBD80 },
 { 0x4ED9, 0xC120 },
 { 0x4EDD, 0xB3D9 },
 { 0x4EDF, 0xCC9C },
 { 0x4EE3, 0xB300 },
 { 0x4EE4, 0xB839 },
 { 0x4EE5, 0xC774 },
 { 0x4EF0, 0xC559 },
 { 0x4EF2, 0xC911 },
 { 0x4EF6, 0xAC74 },
 { 0x4EF7, 0xAC1C },
 { 0x4EFB, 0xC784 },
 { 0x4F01, 0xAE30 },
 { 0x4F09, 0xD56D },
 { 0x4F0A, 0xC774 },
 { 0x4F0B, 0xAE09 },
 { 0x4F0D, 0xC624 },
 { 0x4F0E, 0xAE30 },
 { 0x4F0F, 0xBCF5 },
 { 0x4F10, 0xBC8C },
 { 0x4F11, 0xD734 },
 { 0x4F2F, 0xBC31 },
 { 0x4F34, 0xBC18 },
 { 0x4F36, 0xB839 },
 { 0x4F38, 0xC2E0 },
 { 0x4F3A, 0xC0AC },
 { 0x4F3C, 0xC0AC },
 { 0x4F3D, 0xAC00 },
 { 0x4F43, 0xC804 },
 { 0x4F46, 0xB2E8 },
 { 0x4F47, 0xC800 },
 { 0x4F48, 0xD3EC },
 { 0x4F4D, 0xC704 },
 { 0x4F4E, 0xC800 },
 { 0x4F4F, 0xC8FC },
 { 0x4F50, 0xC88C },
 { 0x4F51, 0xC6B0 },
 { 0x4F55, 0xD558 },
 { 0x4F59, 0xC5EC },
 { 0x4F5A, 0xC77C },
 { 0x4F5B, 0xBD88 },
 { 0x4F5C, 0xC791 },
 { 0x4F69, 0xD328 },
 { 0x4F6F, 0xC591 },
 { 0x4F70, 0xBC31 },
 { 0x4F73, 0xAC00 },
 { 0x4F76, 0xAE38 },
 { 0x4F7A, 0xC804 },
 { 0x4F7E, 0xC77C },
 { 0x4F7F, 0xC0AC },
 { 0x4F81, 0xC2E0 },
 { 0x4F83, 0xAC04 },
 { 0x4F84, 0xC9C8 },
 { 0x4F86, 0xB0B4 },
 { 0x4F88, 0xCE58 },
 { 0x4F8A, 0xAD11 },
 { 0x4F8B, 0xB840 },
 { 0x4F8D, 0xC2DC },
 { 0x4F8F, 0xC8FC },
 { 0x4F91, 0xC720 },
 { 0x4F96, 0xB95C },
 { 0x4F98, 0xCC28 },
 { 0x4F9B, 0xACF5 },
 { 0x4F9D, 0xC758 },
 { 0x4FAE, 0xBAA8 },
 { 0x4FAF, 0xD6C4 },
 { 0x4FB5, 0xCE68 },
 { 0x4FB6, 0xB824 },
 { 0x4FBF, 0xBCC0 },
 { 0x4FC2, 0xACC4 },
 { 0x4FC3, 0xCD09 },
 { 0x4FC4, 0xC544 },
 { 0x4FC9, 0xC624 },
 { 0x4FCA, 0xC900 },
 { 0x4FCE, 0xC870 },
 { 0x4FD1, 0xC6A9 },
 { 0x4FD3, 0xACBD },
 { 0x4FD4, 0xD604 },
 { 0x4FD7, 0xC18D },
 { 0x4FDA, 0xB9AC },
 { 0x4FDD, 0xBCF4 },
 { 0x4FDF, 0xC0AC },
 { 0x4FE0, 0xD611 },
 { 0x4FE1, 0xC2E0 },
 { 0x4FEE, 0xC218 },
 { 0x4FEF, 0xBD80 },
 { 0x4FF1, 0xAD6C },
 { 0x4FF3, 0xBC30 },
 { 0x4FF5, 0xD45C },
 { 0x4FF8, 0xBD09 },
 { 0x4FFA, 0xC5C4 },
 { 0x5002, 0xBCD1 },
 { 0x5006, 0xB7C9 },
 { 0x5009, 0xCC3D },
 { 0x500B, 0xAC1C },
 { 0x500D, 0xBC30 },
 { 0x5011, 0xBB38 },
 { 0x5012, 0xB3C4 },
 { 0x5016, 0xD589 },
 { 0x5019, 0xD6C4 },
 { 0x501A, 0xC758 },
 { 0x501C, 0xCC99 },
 { 0x501E, 0xACBD },
 { 0x501F, 0xCC28 },
 { 0x5021, 0xCC3D },
 { 0x5023, 0xBC29 },
 { 0x5024, 0xCE58 },
 { 0x5026, 0xAD8C },
 { 0x5027, 0xC885 },
 { 0x5028, 0xAC70 },
 { 0x502A, 0xC608 },
 { 0x502B, 0xB95C },
 { 0x502C, 0xD0C1 },
 { 0x502D, 0xC65C },
 { 0x503B, 0xC57C },
 { 0x5043, 0xC5B8 },
 { 0x5047, 0xAC00 },
 { 0x5048, 0xAC8C },
 { 0x5049, 0xC704 },
 { 0x504F, 0xD3B8 },
 { 0x5055, 0xD574 },
 { 0x505A, 0xC8FC },
 { 0x505C, 0xC815 },
 { 0x5065, 0xAC74 },
 { 0x5074, 0xCE21 },
 { 0x5075, 0xC815 },
 { 0x5076, 0xC6B0 },
 { 0x5078, 0xD22C },
 { 0x5080, 0xAD34 },
 { 0x5085, 0xBD80 },
 { 0x508D, 0xBC29 },
 { 0x5091, 0xAC78 },
 { 0x5098, 0xC0B0 },
 { 0x5099, 0xBE44 },
 { 0x50AC, 0xCD5C },
 { 0x50AD, 0xC6A9 },
 { 0x50B2, 0xC624 },
 { 0x50B3, 0xC804 },
 { 0x50B5, 0xCC44 },
 { 0x50B7, 0xC0C1 },
 { 0x50BE, 0xACBD },
 { 0x50C5, 0xADFC },
 { 0x50C9, 0xCCA8 },
 { 0x50CA, 0xC120 },
 { 0x50CF, 0xC0C1 },
 { 0x50D1, 0xAD50 },
 { 0x50D5, 0xBCF5 },
 { 0x50D6, 0xD76C },
 { 0x50DA, 0xB8CC },
 { 0x50DE, 0xC704 },
 { 0x50E5, 0xC694 },
 { 0x50E7, 0xC2B9 },
 { 0x50ED, 0xCC38 },
 { 0x50F9, 0xAC00 },
 { 0x50FB, 0xBCBD },
 { 0x50FF, 0xC0AC },
 { 0x5100, 0xC758 },
 { 0x5101, 0xC900 },
 { 0x5104, 0xC5B5 },
 { 0x5106, 0xACBD },
 { 0x5109, 0xAC80 },
 { 0x5112, 0xC720 },
 { 0x511F, 0xC0C1 },
 { 0x5121, 0xB8B0 },
 { 0x512A, 0xC6B0 },
 { 0x5132, 0xC800 },
 { 0x5137, 0xB824 },
 { 0x513A, 0xB098 },
 { 0x513C, 0xC5C4 },
 { 0x5140, 0xC62C },
 { 0x5141, 0xC724 },
 { 0x5143, 0xC6D0 },
 { 0x5144, 0xD615 },
 { 0x5145, 0xCDA9 },
 { 0x5146, 0xC870 },
 { 0x5147, 0xD749 },
 { 0x5148, 0xC120 },
 { 0x5149, 0xAD11 },
 { 0x514B, 0xADF9 },
 { 0x514C, 0xD0DC },
 { 0x514D, 0xBA74 },
 { 0x514E, 0xD1A0 },
 { 0x5152, 0xC544 },
 { 0x515C, 0xB450 },
 { 0x5162, 0xAE0D },
 { 0x5165, 0xC785 },
 { 0x5167, 0xB0B4 },
 { 0x5168, 0xC804 },
 { 0x5169, 0xB7C9 },
 { 0x516A, 0xC720 },
 { 0x516B, 0xD314 },
 { 0x516C, 0xACF5 },
 { 0x516D, 0xB959 },
 { 0x516E, 0xD61C },
 { 0x5171, 0xACF5 },
 { 0x5175, 0xBCD1 },
 { 0x5176, 0xAE30 },
 { 0x5177, 0xAD6C },
 { 0x5178, 0xC804 },
 { 0x517C, 0xACB8 },
 { 0x5180, 0xAE30 },
 { 0x5186, 0xC5D4 },
 { 0x518A, 0xCC45 },
 { 0x518D, 0xC7AC },
 { 0x5192, 0xBAA8 },
 { 0x5195, 0xBA74 },
 { 0x5197, 0xC6A9 },
 { 0x51A0, 0xAD00 },
 { 0x51A5, 0xBA85 },
 { 0x51AA, 0xBA71 },
 { 0x51AC, 0xB3D9 },
 { 0x51B6, 0xC57C },
 { 0x51B7, 0xB0C9 },
 { 0x51BD, 0xB82C },
 { 0x51C4, 0xCC98 },
 { 0x51C6, 0xC900 },
 { 0x51C9, 0xB7C9 },
 { 0x51CB, 0xC870 },
 { 0x51CC, 0xB2A5 },
 { 0x51CD, 0xB3D9 },
 { 0x51DC, 0xB2A0 },
 { 0x51DD, 0xC751 },
 { 0x51DE, 0xD76C },
 { 0x51E1, 0xBC94 },
 { 0x51F0, 0xD669 },
 { 0x51F1, 0xAC1C },
 { 0x51F6, 0xD749 },
 { 0x51F8, 0xCCA0 },
 { 0x51F9, 0xC694 },
 { 0x51FA, 0xCD9C },
 { 0x51FD, 0xD568 },
 { 0x5200, 0xB3C4 },
 { 0x5203, 0xC778 },
 { 0x5206, 0xBD84 },
 { 0x5207, 0xC808 },
 { 0x5208, 0xC608 },
 { 0x520A, 0xAC04 },
 { 0x520E, 0xBB38 },
 { 0x5211, 0xD615 },
 { 0x5217, 0xB82C },
 { 0x521D, 0xCD08 },
 { 0x5224, 0xD310 },
 { 0x5225, 0xBCC4 },
 { 0x5229, 0xB9AC },
 { 0x522A, 0xC0B0 },
 { 0x522E, 0xAD04 },
 { 0x5230, 0xB3C4 },
 { 0x5236, 0xC81C },
 { 0x5237, 0xC1C4 },
 { 0x5238, 0xAD8C },
 { 0x5239, 0xCC30 },
 { 0x523A, 0xC790 },
 { 0x523B, 0xAC01 },
 { 0x5243, 0xCCB4 },
 { 0x5247, 0xCE59 },
 { 0x524A, 0xC0AD },
 { 0x524B, 0xADF9 },
 { 0x524C, 0xB784 },
 { 0x524D, 0xC804 },
 { 0x5254, 0xCC99 },
 { 0x5256, 0xBD80 },
 { 0x525B, 0xAC15 },
 { 0x525D, 0xBC15 },
 { 0x5261, 0xC12C },
 { 0x5269, 0xC789 },
 { 0x526A, 0xC804 },
 { 0x526F, 0xBD80 },
 { 0x5272, 0xD560 },
 { 0x5275, 0xCC3D },
 { 0x527D, 0xD45C },
 { 0x527F, 0xCD08 },
 { 0x5283, 0xD68D },
 { 0x5287, 0xADF9 },
 { 0x5288, 0xBCBD },
 { 0x5289, 0xB958 },
 { 0x528D, 0xAC80 },
 { 0x5291, 0xC81C },
 { 0x5292, 0xAC80 },
 { 0x529B, 0xB825 },
 { 0x529F, 0xACF5 },
 { 0x52A0, 0xAC00 },
 { 0x52A3, 0xB82C },
 { 0x52A4, 0xADFC },
 { 0x52A9, 0xC870 },
 { 0x52AA, 0xB178 },
 { 0x52AB, 0xAC81 },
 { 0x52BE, 0xD575 },
 { 0x52C1, 0xACBD },
 { 0x52C3, 0xBC1C },
 { 0x52C5, 0xCE59 },
 { 0x52C7, 0xC6A9 },
 { 0x52C9, 0xBA74 },
 { 0x52CD, 0xACBD },
 { 0x52D2, 0xB291 },
 { 0x52D5, 0xB3D9 },
 { 0x52D6, 0xC6B1 },
 { 0x52D8, 0xAC10 },
 { 0x52D9, 0xBB34 },
 { 0x52DB, 0xD6C8 },
 { 0x52DD, 0xC2B9 },
 { 0x52DE, 0xB178 },
 { 0x52DF, 0xBAA8 },
 { 0x52E2, 0xC138 },
 { 0x52E3, 0xC801 },
 { 0x52E4, 0xADFC },
 { 0x52F3, 0xD6C8 },
 { 0x52F5, 0xB824 },
 { 0x52F8, 0xAD8C },
 { 0x52FA, 0xC791 },
 { 0x52FB, 0xADE0 },
 { 0x52FE, 0xAD6C },
 { 0x52FF, 0xBB3C },
 { 0x5305, 0xD3EC },
 { 0x5308, 0xD749 },
 { 0x530D, 0xD3EC },
 { 0x530F, 0xD3EC },
 { 0x5310, 0xBCF5 },
 { 0x5315, 0xBE44 },
 { 0x5316, 0xD654 },
 { 0x5317, 0xBC30 },
 { 0x5319, 0xC2DC },
 { 0x5320, 0xC7A5 },
 { 0x5321, 0xAD11 },
 { 0x5323, 0xAC11 },
 { 0x532A, 0xBE44 },
 { 0x532F, 0xD68C },
 { 0x5339, 0xD544 },
 { 0x533F, 0xB2C9 },
 { 0x5340, 0xAD6C },
 { 0x5341, 0xC2ED },
 { 0x5343, 0xCC9C },
 { 0x5344, 0xC785 },
 { 0x5347, 0xC2B9 },
 { 0x5348, 0xC624 },
 { 0x5349, 0xD6FC },
 { 0x534A, 0xBC18 },
 { 0x534D, 0xB9CC },
 { 0x5351, 0xBE44 },
 { 0x5352, 0xC878 },
 { 0x5353, 0xD0C1 },
 { 0x5354, 0xD611 },
 { 0x5357, 0xB0A8 },
 { 0x535A, 0xBC15 },
 { 0x535C, 0xBCF5 },
 { 0x535E, 0xBCC0 },
 { 0x5360, 0xC810 },
 { 0x5366, 0xAD18 },
 { 0x5368, 0xC124 },
 { 0x536F, 0xBB18 },
 { 0x5370, 0xC778 },
 { 0x5371, 0xC704 },
 { 0x5374, 0xAC01 },
 { 0x5375, 0xB09C },
 { 0x5377, 0xAD8C },
 { 0x537D, 0xC989 },
 { 0x537F, 0xACBD },
 { 0x5384, 0xC561 },
 { 0x5393, 0xC560 },
 { 0x5398, 0xB9AC },
 { 0x539A, 0xD6C4 },
 { 0x539F, 0xC6D0 },
 { 0x53A0, 0xCE21 },
 { 0x53A5, 0xAD90 },
 { 0x53A6, 0xD558 },
 { 0x53AD, 0xC5FC },
 { 0x53BB, 0xAC70 },
 { 0x53C3, 0xC0BC },
 { 0x53C8, 0xC6B0 },
 { 0x53C9, 0xCC28 },
 { 0x53CA, 0xAE09 },
 { 0x53CB, 0xC6B0 },
 { 0x53CD, 0xBC18 },
 { 0x53D4, 0xC219 },
 { 0x53D6, 0xCDE8 },
 { 0x53D7, 0xC218 },
 { 0x53DB, 0xBC18 },
 { 0x53E1, 0xC608 },
 { 0x53E2, 0xCD1D },
 { 0x53E3, 0xAD6C },
 { 0x53E4, 0xACE0 },
 { 0x53E5, 0xAD6C },
 { 0x53E9, 0xACE0 },
 { 0x53EA, 0xC9C0 },
 { 0x53EB, 0xADDC },
 { 0x53EC, 0xC18C },
 { 0x53ED, 0xD314 },
 { 0x53EF, 0xAC00 },
 { 0x53F0, 0xD0DC },
 { 0x53F1, 0xC9C8 },
 { 0x53F2, 0xC0AC },
 { 0x53F3, 0xC6B0 },
 { 0x53F8, 0xC0AC },
 { 0x5403, 0xD758 },
 { 0x5404, 0xAC01 },
 { 0x5408, 0xD569 },
 { 0x5409, 0xAE38 },
 { 0x540A, 0xC801 },
 { 0x540C, 0xB3D9 },
 { 0x540D, 0xBA85 },
 { 0x540E, 0xD6C4 },
 { 0x540F, 0xB9AC },
 { 0x5410, 0xD1A0 },
 { 0x5411, 0xD5A5 },
 { 0x541B, 0xAD70 },
 { 0x541D, 0xB9B0 },
 { 0x541F, 0xC74C },
 { 0x5420, 0xD3D0 },
 { 0x5426, 0xBD80 },
 { 0x5429, 0xBD84 },
 { 0x542B, 0xD568 },
 { 0x5433, 0xC624 },
 { 0x5438, 0xD761 },
 { 0x5439, 0xCDE8 },
 { 0x543B, 0xBB38 },
 { 0x543C, 0xD6C4 },
 { 0x543E, 0xC624 },
 { 0x5442, 0xB824 },
 { 0x5448, 0xC815 },
 { 0x544A, 0xACE0 },
 { 0x5451, 0xD0C4 },
 { 0x5468, 0xC8FC },
 { 0x546A, 0xC8FC },
 { 0x5471, 0xACE0 },
 { 0x5473, 0xBBF8 },
 { 0x5475, 0xAC00 },
 { 0x547B, 0xC2E0 },
 { 0x547C, 0xD638 },
 { 0x547D, 0xBA85 },
 { 0x5480, 0xC800 },
 { 0x5486, 0xD3EC },
 { 0x548C, 0xD654 },
 { 0x548E, 0xAD6C },
 { 0x5490, 0xBD80 },
 { 0x54A4, 0xD0C0 },
 { 0x54A8, 0xC790 },
 { 0x54AB, 0xC9C0 },
 { 0x54AC, 0xAD50 },
 { 0x54B3, 0xD574 },
 { 0x54B8, 0xD568 },
 { 0x54BD, 0xC5F4 },
 { 0x54C0, 0xC560 },
 { 0x54C1, 0xD488 },
 { 0x54C4, 0xD64D },
 { 0x54C8, 0xD569 },
 { 0x54C9, 0xC7AC },
 { 0x54E1, 0xC6D0 },
 { 0x54E5, 0xAC00 },
 { 0x54E8, 0xCD08 },
 { 0x54ED, 0xACE1 },
 { 0x54EE, 0xD6A8 },
 { 0x54F2, 0xCCA0 },
 { 0x54FA, 0xD3EC },
 { 0x5504, 0xD328 },
 { 0x5506, 0xC0AC },
 { 0x5507, 0xC9C4 },
 { 0x550E, 0xB9AC },
 { 0x5510, 0xB2F9 },
 { 0x551C, 0xB9D0 },
 { 0x552F, 0xC720 },
 { 0x5531, 0xCC3D },
 { 0x5535, 0xC554 },
 { 0x553E, 0xD0C0 },
 { 0x5544, 0xD0C1 },
 { 0x5546, 0xC0C1 },
 { 0x554F, 0xBB38 },
 { 0x5553, 0xACC4 },
 { 0x5556, 0xB2F4 },
 { 0x555E, 0xC544 },
 { 0x5563, 0xD568 },
 { 0x557C, 0xC81C },
 { 0x5580, 0xAC1D },
 { 0x5584, 0xC120 },
 { 0x5586, 0xCCA0 },
 { 0x5587, 0xB098 },
 { 0x5589, 0xD6C4 },
 { 0x558A, 0xD568 },
 { 0x5598, 0xCC9C },
 { 0x5599, 0xD6FC },
 { 0x559A, 0xD658 },
 { 0x559C, 0xD76C },
 { 0x559D, 0xAC08 },
 { 0x55A7, 0xD6E4 },
 { 0x55A9, 0xC720 },
 { 0x55AA, 0xC0C1 },
 { 0x55AB, 0xB07D },
 { 0x55AC, 0xAD50 },
 { 0x55AE, 0xB2E8 },
 { 0x55C5, 0xD6C4 },
 { 0x55C7, 0xC0C9 },
 { 0x55D4, 0xC9C4 },
 { 0x55DA, 0xC624 },
 { 0x55DC, 0xAE30 },
 { 0x55DF, 0xCC28 },
 { 0x55E3, 0xC0AC },
 { 0x55E4, 0xCE58 },
 { 0x55FD, 0xC218 },
 { 0x55FE, 0xC8FC },
 { 0x5606, 0xD0C4 },
 { 0x5609, 0xAC00 },
 { 0x5614, 0xAD6C },
 { 0x5617, 0xC0C1 },
 { 0x562F, 0xC18C },
 { 0x5632, 0xC870 },
 { 0x5634, 0xCDE8 },
 { 0x5636, 0xC2DC },
 { 0x5653, 0xD5C8 },
 { 0x5668, 0xAE30 },
 { 0x566B, 0xD76C },
 { 0x5674, 0xBD84 },
 { 0x5686, 0xD6A8 },
 { 0x56A5, 0xC5F0 },
 { 0x56AC, 0xBE48 },
 { 0x56AE, 0xD5A5 },
 { 0x56B4, 0xC5C4 },
 { 0x56BC, 0xC791 },
 { 0x56CA, 0xB0AD },
 { 0x56CD, 0xD76C },
 { 0x56D1, 0xCD09 },
 { 0x56DA, 0xC218 },
 { 0x56DB, 0xC0AC },
 { 0x56DE, 0xD68C },
 { 0x56E0, 0xC778 },
 { 0x56F0, 0xACE4 },
 { 0x56F9, 0xB839 },
 { 0x56FA, 0xACE0 },
 { 0x5703, 0xD3EC },
 { 0x5704, 0xC5B4 },
 { 0x5708, 0xAD8C },
 { 0x570B, 0xAD6D },
 { 0x570D, 0xC704 },
 { 0x5712, 0xC6D0 },
 { 0x5713, 0xC6D0 },
 { 0x5716, 0xB3C4 },
 { 0x5718, 0xB2E8 },
 { 0x571F, 0xD1A0 },
 { 0x5728, 0xC7AC },
 { 0x572D, 0xADDC },
 { 0x5730, 0xC9C0 },
 { 0x573B, 0xAE30 },
 { 0x5740, 0xC9C0 },
 { 0x5742, 0xD310 },
 { 0x5747, 0xADE0 },
 { 0x574A, 0xBC29 },
 { 0x574D, 0xB2F4 },
 { 0x574E, 0xAC10 },
 { 0x5750, 0xC88C },
 { 0x5751, 0xAC31 },
 { 0x5761, 0xD30C },
 { 0x5764, 0xACE4 },
 { 0x5766, 0xD0C4 },
 { 0x576A, 0xD3C9 },
 { 0x576E, 0xB300 },
 { 0x5770, 0xACBD },
 { 0x5775, 0xAD6C },
 { 0x577C, 0xD0C1 },
 { 0x5782, 0xC218 },
 { 0x5788, 0xB300 },
 { 0x578B, 0xD615 },
 { 0x5793, 0xD574 },
 { 0x57A0, 0xC740 },
 { 0x57A2, 0xAD6C },
 { 0x57A3, 0xC6D0 },
 { 0x57C3, 0xC560 },
 { 0x57C7, 0xC6A9 },
 { 0x57C8, 0xC900 },
 { 0x57CB, 0xB9E4 },
 { 0x57CE, 0xC131 },
 { 0x57DF, 0xC5ED },
 { 0x57E0, 0xBD80 },
 { 0x57F0, 0xCC44 },
 { 0x57F4, 0xC2DD },
 { 0x57F7, 0xC9D1 },
 { 0x57F9, 0xBC30 },
 { 0x57FA, 0xAE30 },
 { 0x57FC, 0xAE30 },
 { 0x5800, 0xAD74 },
 { 0x5802, 0xB2F9 },
 { 0x5805, 0xACAC },
 { 0x5806, 0xD1F4 },
 { 0x5808, 0xAC15 },
 { 0x5809, 0xC721 },
 { 0x580A, 0xC545 },
 { 0x581E, 0xCCA9 },
 { 0x5821, 0xBCF4 },
 { 0x5824, 0xC81C },
 { 0x5827, 0xC5F0 },
 { 0x582A, 0xAC10 },
 { 0x582F, 0xC694 },
 { 0x5830, 0xC5B8 },
 { 0x5831, 0xBCF4 },
 { 0x5834, 0xC7A5 },
 { 0x5835, 0xB3C4 },
 { 0x583A, 0xACC4 },
 { 0x584A, 0xAD34 },
 { 0x584B, 0xC601 },
 { 0x584F, 0xAC1C },
 { 0x5851, 0xC18C },
 { 0x5854, 0xD0D1 },
 { 0x5857, 0xB3C4 },
 { 0x5858, 0xB2F9 },
 { 0x585A, 0xCD1D },
 { 0x585E, 0xC0C8 },
 { 0x5861, 0xC804 },
 { 0x5862, 0xC624 },
 { 0x5864, 0xD6C8 },
 { 0x5875, 0xC9C4 },
 { 0x5879, 0xCC38 },
 { 0x587C, 0xC804 },
 { 0x587E, 0xC219 },
 { 0x5883, 0xACBD },
 { 0x5885, 0xC11C },
 { 0x5889, 0xC6A9 },
 { 0x5893, 0xBB18 },
 { 0x589C, 0xCD94 },
 { 0x589E, 0xC99D },
 { 0x589F, 0xD5C8 },
 { 0x58A8, 0xBB35 },
 { 0x58A9, 0xB3C8 },
 { 0x58AE, 0xD0C0 },
 { 0x58B3, 0xBD84 },
 { 0x58BA, 0xC624 },
 { 0x58BB, 0xC7A5 },
 { 0x58BE, 0xAC04 },
 { 0x58C1, 0xBCBD },
 { 0x58C5, 0xC639 },
 { 0x58C7, 0xB2E8 },
 { 0x58CE, 0xD6C8 },
 { 0x58D1, 0xD559 },
 { 0x58D3, 0xC555 },
 { 0x58D5, 0xD638 },
 { 0x58D8, 0xB204 },
 { 0x58D9, 0xAD11 },
 { 0x58DE, 0xAD34 },
 { 0x58DF, 0xB18D },
 { 0x58E4, 0xC591 },
 { 0x58EB, 0xC0AC },
 { 0x58EC, 0xC784 },
 { 0x58EF, 0xC7A5 },
 { 0x58F9, 0xC77C },
 { 0x58FA, 0xD638 },
 { 0x58FB, 0xC11C },
 { 0x58FD, 0xC218 },
 { 0x590F, 0xD558 },
 { 0x5914, 0xAE30 },
 { 0x5915, 0xC11D },
 { 0x5916, 0xC678 },
 { 0x5919, 0xC219 },
 { 0x591A, 0xB2E4 },
 { 0x591C, 0xC57C },
 { 0x5922, 0xBABD },
 { 0x5927, 0xB300 },
 { 0x5929, 0xCC9C },
 { 0x592A, 0xD0DC },
 { 0x592B, 0xBD80 },
 { 0x592D, 0xC694 },
 { 0x592E, 0xC559 },
 { 0x5931, 0xC2E4 },
 { 0x5937, 0xC774 },
 { 0x593E, 0xD611 },
 { 0x5944, 0xC5C4 },
 { 0x5947, 0xAE30 },
 { 0x5948, 0xB098 },
 { 0x5949, 0xBD09 },
 { 0x594E, 0xADDC },
 { 0x594F, 0xC8FC },
 { 0x5950, 0xD658 },
 { 0x5951, 0xACC4 },
 { 0x5954, 0xBD84 },
 { 0x5955, 0xD601 },
 { 0x5957, 0xD22C },
 { 0x595A, 0xD574 },
 { 0x5960, 0xC804 },
 { 0x5962, 0xC0AC },
 { 0x5967, 0xC624 },
 { 0x596A, 0xD0C8 },
 { 0x596B, 0xC724 },
 { 0x596C, 0xC7A5 },
 { 0x596D, 0xC11D },
 { 0x596E, 0xBD84 },
 { 0x5973, 0xB140 },
 { 0x5974, 0xB178 },
 { 0x5978, 0xAC04 },
 { 0x597D, 0xD638 },
 { 0x5982, 0xC5EC },
 { 0x5983, 0xBE44 },
 { 0x5984, 0xB9DD },
 { 0x598A, 0xC784 },
 { 0x5993, 0xAE30 },
 { 0x5996, 0xC694 },
 { 0x5997, 0xAE08 },
 { 0x5999, 0xBB18 },
 { 0x59A5, 0xD0C0 },
 { 0x59A8, 0xBC29 },
 { 0x59AC, 0xD22C },
 { 0x59B9, 0xB9E4 },
 { 0x59BB, 0xCC98 },
 { 0x59BE, 0xCCA9 },
 { 0x59C3, 0xC815 },
 { 0x59C6, 0xBAA8 },
 { 0x59C9, 0xC790 },
 { 0x59CB, 0xC2DC },
 { 0x59D0, 0xC800 },
 { 0x59D1, 0xACE0 },
 { 0x59D3, 0xC131 },
 { 0x59D4, 0xC704 },
 { 0x59D9, 0xC784 },
 { 0x59DA, 0xC694 },
 { 0x59DC, 0xAC15 },
 { 0x59DD, 0xC8FC },
 { 0x59E6, 0xAC04 },
 { 0x59E8, 0xC774 },
 { 0x59EA, 0xC9C8 },
 { 0x59EC, 0xD76C },
 { 0x59EE, 0xD56D },
 { 0x59F8, 0xC5F0 },
 { 0x59FB, 0xC778 },
 { 0x59FF, 0xC790 },
 { 0x5A01, 0xC704 },
 { 0x5A03, 0xC65C },
 { 0x5A11, 0xC0AC },
 { 0x5A18, 0xB0AD },
 { 0x5A1B, 0xC624 },
 { 0x5A1C, 0xB098 },
 { 0x5A1F, 0xC5F0 },
 { 0x5A20, 0xC2E0 },
 { 0x5A25, 0xC544 },
 { 0x5A29, 0xB9CC },
 { 0x5A36, 0xCDE8 },
 { 0x5A3C, 0xCC3D },
 { 0x5A41, 0xB8E8 },
 { 0x5A46, 0xD30C },
 { 0x5A49, 0xC644 },
 { 0x5A5A, 0xD63C },
 { 0x5A62, 0xBE44 },
 { 0x5A66, 0xBD80 },
 { 0x5A92, 0xB9E4 },
 { 0x5A9A, 0xBBF8 },
 { 0x5A9B, 0xC6D0 },
 { 0x5AA4, 0xC2DC },
 { 0x5AC1, 0xAC00 },
 { 0x5AC2, 0xC218 },
 { 0x5AC4, 0xC6D0 },
 { 0x5AC9, 0xC9C8 },
 { 0x5ACC, 0xD610 },
 { 0x5AE1, 0xC801 },
 { 0x5AE6, 0xD56D },
 { 0x5AE9, 0xB208 },
 { 0x5B05, 0xD654 },
 { 0x5B09, 0xD76C },
 { 0x5B0B, 0xC120 },
 { 0x5B0C, 0xAD50 },
 { 0x5B16, 0xD3D0 },
 { 0x5B2A, 0xBE48 },
 { 0x5B40, 0xC0C1 },
 { 0x5B43, 0xC591 },
 { 0x5B50, 0xC790 },
 { 0x5B51, 0xD608 },
 { 0x5B54, 0xACF5 },
 { 0x5B55, 0xC789 },
 { 0x5B57, 0xC790 },
 { 0x5B58, 0xC874 },
 { 0x5B5A, 0xBD80 },
 { 0x5B5C, 0xC790 },
 { 0x5B5D, 0xD6A8 },
 { 0x5B5F, 0xB9F9 },
 { 0x5B63, 0xACC4 },
 { 0x5B64, 0xACE0 },
 { 0x5B69, 0xD574 },
 { 0x5B6B, 0xC190 },
 { 0x5B70, 0xC219 },
 { 0x5B71, 0xC794 },
 { 0x5B75, 0xBD80 },
 { 0x5B78, 0xD559 },
 { 0x5B7A, 0xC720 },
 { 0x5B7C, 0xC5BC },
 { 0x5B85, 0xB301 },
 { 0x5B87, 0xC6B0 },
 { 0x5B88, 0xC218 },
 { 0x5B89, 0xC548 },
 { 0x5B8B, 0xC1A1 },
 { 0x5B8C, 0xC644 },
 { 0x5B8F, 0xAD49 },
 { 0x5B93, 0xBCF5 },
 { 0x5B95, 0xD0D5 },
 { 0x5B96, 0xD6A1 },
 { 0x5B97, 0xC885 },
 { 0x5B98, 0xAD00 },
 { 0x5B99, 0xC8FC },
 { 0x5B9A, 0xC815 },
 { 0x5B9B, 0xC644 },
 { 0x5B9C, 0xC758 },
 { 0x5BA2, 0xAC1D },
 { 0x5BA3, 0xC120 },
 { 0x5BA4, 0xC2E4 },
 { 0x5BA5, 0xC720 },
 { 0x5BA6, 0xD658 },
 { 0x5BAC, 0xC131 },
 { 0x5BAE, 0xAD81 },
 { 0x5BB0, 0xC7AC },
 { 0x5BB3, 0xD574 },
 { 0x5BB4, 0xC5F0 },
 { 0x5BB5, 0xC18C },
 { 0x5BB6, 0xAC00 },
 { 0x5BB8, 0xC2E0 },
 { 0x5BB9, 0xC6A9 },
 { 0x5BBF, 0xC219 },
 { 0x5BC0, 0xCC44 },
 { 0x5BC2, 0xC801 },
 { 0x5BC3, 0xC6D0 },
 { 0x5BC4, 0xAE30 },
 { 0x5BC5, 0xC778 },
 { 0x5BC6, 0xBC00 },
 { 0x5BC7, 0xAD6C },
 { 0x5BCC, 0xBD80 },
 { 0x5BD0, 0xB9E4 },
 { 0x5BD2, 0xD55C },
 { 0x5BD3, 0xC6B0 },
 { 0x5BD4, 0xC2DD },
 { 0x5BD7, 0xB155 },
 { 0x5BDE, 0xB9C9 },
 { 0x5BDF, 0xCC30 },
 { 0x5BE1, 0xACFC },
 { 0x5BE2, 0xCE68 },
 { 0x5BE4, 0xC624 },
 { 0x5BE5, 0xC694 },
 { 0x5BE6, 0xC2E4 },
 { 0x5BE7, 0xB155 },
 { 0x5BE8, 0xCC44 },
 { 0x5BE9, 0xC2EC },
 { 0x5BEB, 0xC0AC },
 { 0x5BEC, 0xAD00 },
 { 0x5BEE, 0xB8CC },
 { 0x5BEF, 0xC900 },
 { 0x5BF5, 0xCD1D },
 { 0x5BF6, 0xBCF4 },
 { 0x5BF8, 0xCD0C },
 { 0x5BFA, 0xC0AC },
 { 0x5C01, 0xBD09 },
 { 0x5C04, 0xC0AC },
 { 0x5C07, 0xC7A5 },
 { 0x5C08, 0xC804 },
 { 0x5C09, 0xC704 },
 { 0x5C0A, 0xC874 },
 { 0x5C0B, 0xC2EC },
 { 0x5C0D, 0xB300 },
 { 0x5C0E, 0xB3C4 },
 { 0x5C0F, 0xC18C },
 { 0x5C11, 0xC18C },
 { 0x5C16, 0xCCA8 },
 { 0x5C19, 0xC0C1 },
 { 0x5C24, 0xC6B0 },
 { 0x5C28, 0xBC29 },
 { 0x5C31, 0xCDE8 },
 { 0x5C38, 0xC2DC },
 { 0x5C39, 0xC724 },
 { 0x5C3A, 0xCC99 },
 { 0x5C3B, 0xACE0 },
 { 0x5C3C, 0xB2C8 },
 { 0x5C3E, 0xBBF8 },
 { 0x5C3F, 0xB1E8 },
 { 0x5C40, 0xAD6D },
 { 0x5C45, 0xAC70 },
 { 0x5C46, 0xACC4 },
 { 0x5C48, 0xAD74 },
 { 0x5C4B, 0xC625 },
 { 0x5C4D, 0xC2DC },
 { 0x5C4E, 0xC2DC },
 { 0x5C51, 0xC124 },
 { 0x5C55, 0xC804 },
 { 0x5C5B, 0xBCD1 },
 { 0x5C60, 0xB3C4 },
 { 0x5C62, 0xB204 },
 { 0x5C64, 0xCE35 },
 { 0x5C65, 0xB9AC },
 { 0x5C6C, 0xC18D },
 { 0x5C6F, 0xB454 },
 { 0x5C71, 0xC0B0 },
 { 0x5C79, 0xD758 },
 { 0x5C90, 0xAE30 },
 { 0x5C91, 0xC7A0 },
 { 0x5CA1, 0xAC15 },
 { 0x5CA9, 0xC554 },
 { 0x5CAB, 0xC218 },
 { 0x5CAC, 0xAC11 },
 { 0x5CB1, 0xB300 },
 { 0x5CB3, 0xC545 },
 { 0x5CB5, 0xD638 },
 { 0x5CB7, 0xBBFC },
 { 0x5CB8, 0xC548 },
 { 0x5CBA, 0xB839 },
 { 0x5CBE, 0xC810 },
 { 0x5CC0, 0xC218 },
 { 0x5CD9, 0xCE58 },
 { 0x5CE0, 0xC0C1 },
 { 0x5CE8, 0xC544 },
 { 0x5CEF, 0xBD09 },
 { 0x5CF0, 0xBD09 },
 { 0x5CF4, 0xD604 },
 { 0x5CF6, 0xB3C4 },
 { 0x5CFB, 0xC900 },
 { 0x5CFD, 0xD611 },
 { 0x5D07, 0xC22D },
 { 0x5D0D, 0xB798 },
 { 0x5D0E, 0xAE30 },
 { 0x5D11, 0xACE4 },
 { 0x5D14, 0xCD5C },
 { 0x5D16, 0xC560 },
 { 0x5D17, 0xAC15 },
 { 0x5D19, 0xB95C },
 { 0x5D27, 0xC22D },
 { 0x5D29, 0xBD95 },
 { 0x5D4B, 0xBBF8 },
 { 0x5D4C, 0xAC10 },
 { 0x5D50, 0xB0A8 },
 { 0x5D69, 0xC22D },
 { 0x5D6C, 0xC678 },
 { 0x5D6F, 0xCC28 },
 { 0x5D87, 0xAD6C },
 { 0x5D8B, 0xB3C4 },
 { 0x5D9D, 0xB4F1 },
 { 0x5DA0, 0xAD50 },
 { 0x5DA2, 0xC694 },
 { 0x5DAA, 0xC5C5 },
 { 0x5DB8, 0xC601 },
 { 0x5DBA, 0xB839 },
 { 0x5DBC, 0xC11C },
 { 0x5DBD, 0xC545 },
 { 0x5DCD, 0xC678 },
 { 0x5DD2, 0xB9CC },
 { 0x5DD6, 0xC554 },
 { 0x5DDD, 0xCC9C },
 { 0x5DDE, 0xC8FC },
 { 0x5DE1, 0xC21C },
 { 0x5DE2, 0xC18C },
 { 0x5DE5, 0xACF5 },
 { 0x5DE6, 0xC88C },
 { 0x5DE7, 0xAD50 },
 { 0x5DE8, 0xAC70 },
 { 0x5DEB, 0xBB34 },
 { 0x5DEE, 0xCC28 },
 { 0x5DF1, 0xAE30 },
 { 0x5DF2, 0xC774 },
 { 0x5DF3, 0xC0AC },
 { 0x5DF4, 0xD30C },
 { 0x5DF7, 0xD56D },
 { 0x5DFD, 0xC190 },
 { 0x5DFE, 0xAC74 },
 { 0x5E02, 0xC2DC },
 { 0x5E03, 0xD3EC },
 { 0x5E06, 0xBC94 },
 { 0x5E0C, 0xD76C },
 { 0x5E11, 0xD0D5 },
 { 0x5E16, 0xCCA9 },
 { 0x5E19, 0xC9C8 },
 { 0x5E1B, 0xBC31 },
 { 0x5E1D, 0xC81C },
 { 0x5E25, 0xC218 },
 { 0x5E2B, 0xC0AC },
 { 0x5E2D, 0xC11D },
 { 0x5E33, 0xC7A5 },
 { 0x5E36, 0xB300 },
 { 0x5E38, 0xC0C1 },
 { 0x5E3D, 0xBAA8 },
 { 0x5E3F, 0xD6C4 },
 { 0x5E40, 0xC815 },
 { 0x5E44, 0xC545 },
 { 0x5E45, 0xD3ED },
 { 0x5E47, 0xBC29 },
 { 0x5E4C, 0xD669 },
 { 0x5E55, 0xB9C9 },
 { 0x5E5F, 0xCE58 },
 { 0x5E61, 0xBC88 },
 { 0x5E62, 0xB2F9 },
 { 0x5E63, 0xD3D0 },
 { 0x5E72, 0xAC04 },
 { 0x5E73, 0xD3C9 },
 { 0x5E74, 0xB144 },
 { 0x5E77, 0xBCD1 },
 { 0x5E78, 0xD589 },
 { 0x5E79, 0xAC04 },
 { 0x5E7B, 0xD658 },
 { 0x5E7C, 0xC720 },
 { 0x5E7D, 0xC720 },
 { 0x5E7E, 0xAE30 },
 { 0x5E84, 0xC7A5 },
 { 0x5E87, 0xBE44 },
 { 0x5E8A, 0xC0C1 },
 { 0x5E8F, 0xC11C },
 { 0x5E95, 0xC800 },
 { 0x5E97, 0xC810 },
 { 0x5E9A, 0xACBD },
 { 0x5E9C, 0xBD80 },
 { 0x5EA0, 0xC0C1 },
 { 0x5EA6, 0xB3C4 },
 { 0x5EA7, 0xC88C },
 { 0x5EAB, 0xACE0 },
 { 0x5EAD, 0xC815 },
 { 0x5EB5, 0xC554 },
 { 0x5EB6, 0xC11C },
 { 0x5EB7, 0xAC15 },
 { 0x5EB8, 0xC6A9 },
 { 0x5EBE, 0xC720 },
 { 0x5EC2, 0xC0C1 },
 { 0x5EC8, 0xD558 },
 { 0x5EC9, 0xB834 },
 { 0x5ECA, 0xB0AD },
 { 0x5ED0, 0xAD6C },
 { 0x5ED3, 0xACFD },
 { 0x5ED6, 0xB8CC },
 { 0x5EDA, 0xC8FC },
 { 0x5EDB, 0xC804 },
 { 0x5EDF, 0xBB18 },
 { 0x5EE0, 0xCC3D },
 { 0x5EE2, 0xD3D0 },
 { 0x5EE3, 0xAD11 },
 { 0x5EEC, 0xB824 },
 { 0x5EF3, 0xCCAD },
 { 0x5EF6, 0xC5F0 },
 { 0x5EF7, 0xC815 },
 { 0x5EFA, 0xAC74 },
 { 0x5EFB, 0xD68C },
 { 0x5F01, 0xBCC0 },
 { 0x5F04, 0xB18D },
 { 0x5F0A, 0xD3D0 },
 { 0x5F0F, 0xC2DD },
 { 0x5F11, 0xC2DC },
 { 0x5F13, 0xAD81 },
 { 0x5F14, 0xC870 },
 { 0x5F15, 0xC778 },
 { 0x5F17, 0xBD88 },
 { 0x5F18, 0xD64D },
 { 0x5F1B, 0xC774 },
 { 0x5F1F, 0xC81C },
 { 0x5F26, 0xD604 },
 { 0x5F27, 0xD638 },
 { 0x5F29, 0xB178 },
 { 0x5F31, 0xC57D },
 { 0x5F35, 0xC7A5 },
 { 0x5F3A, 0xAC15 },
 { 0x5F3C, 0xD544 },
 { 0x5F48, 0xD0C4 },
 { 0x5F4A, 0xAC15 },
 { 0x5F4C, 0xBBF8 },
 { 0x5F4E, 0xB9CC },
 { 0x5F56, 0xB2E8 },
 { 0x5F57, 0xD61C },
 { 0x5F59, 0xD718 },
 { 0x5F5B, 0xC774 },
 { 0x5F62, 0xD615 },
 { 0x5F66, 0xC5B8 },
 { 0x5F67, 0xC6B1 },
 { 0x5F69, 0xCC44 },
 { 0x5F6A, 0xD45C },
 { 0x5F6B, 0xC870 },
 { 0x5F6C, 0xBE48 },
 { 0x5F6D, 0xD33D },
 { 0x5F70, 0xCC3D },
 { 0x5F71, 0xC601 },
 { 0x5F77, 0xBC29 },
 { 0x5F79, 0xC5ED },
 { 0x5F7C, 0xD53C },
 { 0x5F7F, 0xBD88 },
 { 0x5F80, 0xC655 },
 { 0x5F81, 0xC815 },
 { 0x5F85, 0xB300 },
 { 0x5F87, 0xC21C },
 { 0x5F8A, 0xD68C },
 { 0x5F8B, 0xB960 },
 { 0x5F8C, 0xD6C4 },
 { 0x5F90, 0xC11C },
 { 0x5F91, 0xACBD },
 { 0x5F92, 0xB3C4 },
 { 0x5F97, 0xB4DD },
 { 0x5F98, 0xBC30 },
 { 0x5F99, 0xC0AC },
 { 0x5F9E, 0xC885 },
 { 0x5FA0, 0xB798 },
 { 0x5FA1, 0xC5B4 },
 { 0x5FA8, 0xD669 },
 { 0x5FA9, 0xBCF5 },
 { 0x5FAA, 0xC21C },
 { 0x5FAE, 0xBBF8 },
 { 0x5FB5, 0xC9D5 },
 { 0x5FB7, 0xB355 },
 { 0x5FB9, 0xCCA0 },
 { 0x5FBD, 0xD718 },
 { 0x5FC3, 0xC2EC },
 { 0x5FC5, 0xD544 },
 { 0x5FCC, 0xAE30 },
 { 0x5FCD, 0xC778 },
 { 0x5FD6, 0xCD0C },
 { 0x5FD7, 0xC9C0 },
 { 0x5FD8, 0xB9DD },
 { 0x5FD9, 0xB9DD },
 { 0x5FE0, 0xCDA9 },
 { 0x5FEB, 0xCF8C },
 { 0x5FF5, 0xB150 },
 { 0x5FFD, 0xD640 },
 { 0x5FFF, 0xBD84 },
 { 0x600F, 0xC559 },
 { 0x6012, 0xB178 },
 { 0x6016, 0xD3EC },
 { 0x601C, 0xB839 },
 { 0x601D, 0xC0AC },
 { 0x6020, 0xD0DC },
 { 0x6021, 0xC774 },
 { 0x6025, 0xAE09 },
 { 0x6027, 0xC131 },
 { 0x6028, 0xC6D0 },
 { 0x602A, 0xAD34 },
 { 0x602F, 0xAC81 },
 { 0x6041, 0xC784 },
 { 0x6042, 0xC21C },
 { 0x6043, 0xC2DC },
 { 0x604D, 0xD669 },
 { 0x6050, 0xACF5 },
 { 0x6052, 0xD56D },
 { 0x6055, 0xC11C },
 { 0x6059, 0xC591 },
 { 0x605D, 0xAD04 },
 { 0x6062, 0xD68C },
 { 0x6063, 0xC790 },
 { 0x6064, 0xD73C },
 { 0x6065, 0xCE58 },
 { 0x6068, 0xD55C },
 { 0x6069, 0xC740 },
 { 0x606A, 0xAC01 },
 { 0x606C, 0xB150 },
 { 0x606D, 0xACF5 },
 { 0x606F, 0xC2DD },
 { 0x6070, 0xD761 },
 { 0x6085, 0xC5F4 },
 { 0x6089, 0xC2E4 },
 { 0x608C, 0xC81C },
 { 0x608D, 0xD55C },
 { 0x6094, 0xD68C },
 { 0x6096, 0xD328 },
 { 0x609A, 0xC1A1 },
 { 0x609B, 0xC804 },
 { 0x609F, 0xC624 },
 { 0x60A0, 0xC720 },
 { 0x60A3, 0xD658 },
 { 0x60A4, 0xCD1D },
 { 0x60A7, 0xB9AC },
 { 0x60B0, 0xC885 },
 { 0x60B2, 0xBE44 },
 { 0x60B3, 0xB355 },
 { 0x60B4, 0xCDCC },
 { 0x60B6, 0xBBFC },
 { 0x60B8, 0xACC4 },
 { 0x60BC, 0xB3C4 },
 { 0x60BD, 0xCC98 },
 { 0x60C5, 0xC815 },
 { 0x60C7, 0xB3C8 },
 { 0x60D1, 0xD639 },
 { 0x60DA, 0xD640 },
 { 0x60DC, 0xC11D },
 { 0x60DF, 0xC720 },
 { 0x60E0, 0xD61C },
 { 0x60E1, 0xC545 },
 { 0x60F0, 0xD0C0 },
 { 0x60F1, 0xB1CC },
 { 0x60F3, 0xC0C1 },
 { 0x60F6, 0xD669 },
 { 0x60F9, 0xC57C },
 { 0x60FA, 0xC131 },
 { 0x60FB, 0xCE21 },
 { 0x6101, 0xC218 },
 { 0x6106, 0xAC74 },
 { 0x6108, 0xC720 },
 { 0x6109, 0xC720 },
 { 0x610D, 0xBBFC },
 { 0x610E, 0xD345 },
 { 0x610F, 0xC758 },
 { 0x6115, 0xC545 },
 { 0x611A, 0xC6B0 },
 { 0x611B, 0xC560 },
 { 0x611F, 0xAC10 },
 { 0x6127, 0xAD34 },
 { 0x6130, 0xD669 },
 { 0x6134, 0xCC3D },
 { 0x6137, 0xAC1C },
 { 0x613C, 0xC2E0 },
 { 0x613E, 0xAC1C },
 { 0x613F, 0xC6D0 },
 { 0x6142, 0xC6A9 },
 { 0x6144, 0xB960 },
 { 0x6147, 0xC740 },
 { 0x6148, 0xC790 },
 { 0x614A, 0xACB8 },
 { 0x614B, 0xD0DC },
 { 0x614C, 0xD669 },
 { 0x6153, 0xD45C },
 { 0x6155, 0xBAA8 },
 { 0x6158, 0xCC38 },
 { 0x6159, 0xCC38 },
 { 0x615D, 0xD2B9 },
 { 0x615F, 0xD1B5 },
 { 0x6162, 0xB9CC },
 { 0x6163, 0xAD00 },
 { 0x6164, 0xAC01 },
 { 0x6167, 0xD61C },
 { 0x6168, 0xAC1C },
 { 0x616B, 0xC885 },
 { 0x616E, 0xB824 },
 { 0x6170, 0xC704 },
 { 0x6176, 0xACBD },
 { 0x6177, 0xAC15 },
 { 0x617D, 0xCC99 },
 { 0x617E, 0xC695 },
 { 0x6181, 0xCD1D },
 { 0x6182, 0xC6B0 },
 { 0x618A, 0xBE44 },
 { 0x618E, 0xC99D },
 { 0x6190, 0xB828 },
 { 0x6191, 0xBE59 },
 { 0x6194, 0xCD08 },
 { 0x6198, 0xD76C },
 { 0x6199, 0xD76C },
 { 0x619A, 0xD0C4 },
 { 0x61A4, 0xBD84 },
 { 0x61A7, 0xB3D9 },
 { 0x61A9, 0xAC8C },
 { 0x61AB, 0xBBFC },
 { 0x61AC, 0xACBD },
 { 0x61AE, 0xBB34 },
 { 0x61B2, 0xD5CC },
 { 0x61B6, 0xC5B5 },
 { 0x61BA, 0xB2F4 },
 { 0x61BE, 0xAC10 },
 { 0x61C3, 0xADFC },
 { 0x61C7, 0xAC04 },
 { 0x61C8, 0xD574 },
 { 0x61C9, 0xC751 },
 { 0x61CA, 0xC624 },
 { 0x61CB, 0xBB34 },
 { 0x61E6, 0xB098 },
 { 0x61F2, 0xC9D5 },
 { 0x61F6, 0xB098 },
 { 0x61F7, 0xD68C },
 { 0x61F8, 0xD604 },
 { 0x61FA, 0xCC38 },
 { 0x61FC, 0xAD6C },
 { 0x61FF, 0xC758 },
 { 0x6200, 0xB828 },
 { 0x6207, 0xB2F9 },
 { 0x6208, 0xACFC },
 { 0x620A, 0xBB34 },
 { 0x620C, 0xC220 },
 { 0x620D, 0xC218 },
 { 0x620E, 0xC735 },
 { 0x6210, 0xC131 },
 { 0x6211, 0xC544 },
 { 0x6212, 0xACC4 },
 { 0x6216, 0xD639 },
 { 0x621A, 0xCC99 },
 { 0x621F, 0xADF9 },
 { 0x6221, 0xAC10 },
 { 0x622A, 0xC808 },
 { 0x622E, 0xB959 },
 { 0x6230, 0xC804 },
 { 0x6231, 0xD76C },
 { 0x6234, 0xB300 },
 { 0x6236, 0xD638 },
 { 0x623E, 0xB824 },
 { 0x623F, 0xBC29 },
 { 0x6240, 0xC18C },
 { 0x6241, 0xD3B8 },
 { 0x6247, 0xC120 },
 { 0x6248, 0xD638 },
 { 0x6249, 0xBE44 },
 { 0x624B, 0xC218 },
 { 0x624D, 0xC7AC },
 { 0x6253, 0xD0C0 },
 { 0x6258, 0xD0C1 },
 { 0x626E, 0xBD84 },
 { 0x6271, 0xAE09 },
 { 0x6276, 0xBD80 },
 { 0x6279, 0xBE44 },
 { 0x627C, 0xC561 },
 { 0x627F, 0xC2B9 },
 { 0x6280, 0xAE30 },
 { 0x6284, 0xCD08 },
 { 0x6289, 0xACB0 },
 { 0x628A, 0xD30C },
 { 0x6291, 0xC5B5 },
 { 0x6292, 0xC11C },
 { 0x6295, 0xD22C },
 { 0x6297, 0xD56D },
 { 0x6298, 0xC808 },
 { 0x629B, 0xD3EC },
 { 0x62AB, 0xD53C },
 { 0x62B1, 0xD3EC },
 { 0x62B5, 0xC800 },
 { 0x62B9, 0xB9D0 },
 { 0x62BC, 0xC555 },
 { 0x62BD, 0xCD94 },
 { 0x62C2, 0xBD88 },
 { 0x62C7, 0xBB34 },
 { 0x62C8, 0xB150 },
 { 0x62C9, 0xB0A9 },
 { 0x62CC, 0xBC18 },
 { 0x62CD, 0xBC15 },
 { 0x62CF, 0xB098 },
 { 0x62D0, 0xAD34 },
 { 0x62D2, 0xAC70 },
 { 0x62D3, 0xCC99 },
 { 0x62D4, 0xBC1C },
 { 0x62D6, 0xD0C0 },
 { 0x62D7, 0xC694 },
 { 0x62D8, 0xAD6C },
 { 0x62D9, 0xC878 },
 { 0x62DB, 0xCD08 },
 { 0x62DC, 0xBC30 },
 { 0x62EC, 0xAD04 },
 { 0x62ED, 0xC2DD },
 { 0x62EE, 0xAE38 },
 { 0x62EF, 0xC99D },
 { 0x62F1, 0xACF5 },
 { 0x62F3, 0xAD8C },
 { 0x62F7, 0xACE0 },
 { 0x62FE, 0xC2B5 },
 { 0x62FF, 0xB098 },
 { 0x6301, 0xC9C0 },
 { 0x6307, 0xC9C0 },
 { 0x6309, 0xC548 },
 { 0x6311, 0xB3C4 },
 { 0x632B, 0xC88C },
 { 0x632F, 0xC9C4 },
 { 0x633A, 0xC815 },
 { 0x633B, 0xC5F0 },
 { 0x633D, 0xB9CC },
 { 0x633E, 0xD611 },
 { 0x6349, 0xCC29 },
 { 0x634C, 0xD314 },
 { 0x634F, 0xB0A0 },
 { 0x6350, 0xC5F0 },
 { 0x6355, 0xD3EC },
 { 0x6367, 0xBD09 },
 { 0x6368, 0xC0AC },
 { 0x636E, 0xAC70 },
 { 0x6372, 0xAD8C },
 { 0x6377, 0xCCA9 },
 { 0x637A, 0xB0A0 },
 { 0x637B, 0xB150 },
 { 0x637F, 0xC11C },
 { 0x6383, 0xC18C },
 { 0x6388, 0xC218 },
 { 0x6389, 0xB3C4 },
 { 0x638C, 0xC7A5 },
 { 0x6392, 0xBC30 },
 { 0x6396, 0xC561 },
 { 0x6398, 0xAD74 },
 { 0x639B, 0xAD18 },
 { 0x63A0, 0xB7B5 },
 { 0x63A1, 0xCC44 },
 { 0x63A2, 0xD0D0 },
 { 0x63A5, 0xC811 },
 { 0x63A7, 0xACF5 },
 { 0x63A8, 0xCD94 },
 { 0x63A9, 0xC5C4 },
 { 0x63AA, 0xC870 },
 { 0x63C0, 0xAC04 },
 { 0x63C4, 0xC720 },
 { 0x63C6, 0xADDC },
 { 0x63CF, 0xBB18 },
 { 0x63D0, 0xC81C },
 { 0x63D6, 0xC74D },
 { 0x63DA, 0xC591 },
 { 0x63DB, 0xD658 },
 { 0x63E1, 0xC545 },
 { 0x63ED, 0xAC8C },
 { 0x63EE, 0xD718 },
 { 0x63F4, 0xC6D0 },
 { 0x63F6, 0xC57C },
 { 0x63F7, 0xC0BD },
 { 0x640D, 0xC190 },
 { 0x640F, 0xBC15 },
 { 0x6414, 0xC18C },
 { 0x6416, 0xC694 },
 { 0x6417, 0xB3C4 },
 { 0x641C, 0xC218 },
 { 0x6422, 0xC9C4 },
 { 0x642C, 0xBC18 },
 { 0x642D, 0xD0D1 },
 { 0x643A, 0xD734 },
 { 0x643E, 0xCC29 },
 { 0x6458, 0xC801 },
 { 0x6460, 0xCD1D },
 { 0x6469, 0xB9C8 },
 { 0x646F, 0xC9C0 },
 { 0x6478, 0xBAA8 },
 { 0x6479, 0xBAA8 },
 { 0x647A, 0xC811 },
 { 0x6488, 0xB85C },
 { 0x6491, 0xD0F1 },
 { 0x6492, 0xC0B4 },
 { 0x6493, 0xC694 },
 { 0x649A, 0xB144 },
 { 0x649E, 0xB2F9 },
 { 0x64A4, 0xCCA0 },
 { 0x64A5, 0xBC1C },
 { 0x64AB, 0xBB34 },
 { 0x64AD, 0xD30C },
 { 0x64AE, 0xCD2C },
 { 0x64B0, 0xCC2C },
 { 0x64B2, 0xBC15 },
 { 0x64BB, 0xB2EC },
 { 0x64C1, 0xC639 },
 { 0x64C4, 0xB178 },
 { 0x64C5, 0xCC9C },
 { 0x64C7, 0xD0DD },
 { 0x64CA, 0xACA9 },
 { 0x64CD, 0xC870 },
 { 0x64CE, 0xACBD },
 { 0x64D2, 0xAE08 },
 { 0x64D4, 0xB2F4 },
 { 0x64D8, 0xBCBD },
 { 0x64DA, 0xAC70 },
 { 0x64E1, 0xB300 },
 { 0x64E2, 0xD0C1 },
 { 0x64E5, 0xB78C },
 { 0x64E6, 0xCC30 },
 { 0x64E7, 0xAC70 },
 { 0x64EC, 0xC758 },
 { 0x64F2, 0xCC99 },
 { 0x64F4, 0xD655 },
 { 0x64FA, 0xD30C },
 { 0x64FE, 0xC694 },
 { 0x6500, 0xBC18 },
 { 0x6504, 0xD130 },
 { 0x6518, 0xC591 },
 { 0x651D, 0xC12D },
 { 0x6523, 0xB828 },
 { 0x652A, 0xAD50 },
 { 0x652B, 0xD655 },
 { 0x652C, 0xB78C },
 { 0x652F, 0xC9C0 },
 { 0x6536, 0xC218 },
 { 0x6537, 0xACE0 },
 { 0x6538, 0xC720 },
 { 0x6539, 0xAC1C },
 { 0x653B, 0xACF5 },
 { 0x653E, 0xBC29 },
 { 0x653F, 0xC815 },
 { 0x6545, 0xACE0 },
 { 0x6548, 0xD6A8 },
 { 0x654D, 0xC11C },
 { 0x654E, 0xAD50 },
 { 0x654F, 0xBBFC },
 { 0x6551, 0xAD6C },
 { 0x6556, 0xC624 },
 { 0x6557, 0xD328 },
 { 0x655E, 0xCC3D },
 { 0x6562, 0xAC10 },
 { 0x6563, 0xC0B0 },
 { 0x6566, 0xB3C8 },
 { 0x656C, 0xACBD },
 { 0x656D, 0xC591 },
 { 0x6572, 0xACE0 },
 { 0x6574, 0xC815 },
 { 0x6575, 0xC801 },
 { 0x6577, 0xBD80 },
 { 0x6578, 0xC0AD },
 { 0x657E, 0xC120 },
 { 0x6582, 0xB834 },
 { 0x6583, 0xD3D0 },
 { 0x6585, 0xD6A8 },
 { 0x6587, 0xBB38 },
 { 0x658C, 0xBE48 },
 { 0x6590, 0xBE44 },
 { 0x6591, 0xBC18 },
 { 0x6597, 0xB450 },
 { 0x6599, 0xB8CC },
 { 0x659B, 0xACE1 },
 { 0x659C, 0xC0AC },
 { 0x659F, 0xC9D0 },
 { 0x65A1, 0xC54C },
 { 0x65A4, 0xADFC },
 { 0x65A5, 0xCC99 },
 { 0x65A7, 0xBD80 },
 { 0x65AB, 0xC791 },
 { 0x65AC, 0xCC38 },
 { 0x65AF, 0xC0AC },
 { 0x65B0, 0xC2E0 },
 { 0x65B7, 0xB2E8 },
 { 0x65B9, 0xBC29 },
 { 0x65BC, 0xC5B4 },
 { 0x65BD, 0xC2DC },
 { 0x65C1, 0xBC29 },
 { 0x65C5, 0xB824 },
 { 0x65CB, 0xC120 },
 { 0x65CC, 0xC815 },
 { 0x65CF, 0xC871 },
 { 0x65D2, 0xB958 },
 { 0x65D7, 0xAE30 },
 { 0x65E0, 0xBB34 },
 { 0x65E3, 0xAE30 },
 { 0x65E5, 0xC77C },
 { 0x65E6, 0xB2E8 },
 { 0x65E8, 0xC9C0 },
 { 0x65E9, 0xC870 },
 { 0x65EC, 0xC21C },
 { 0x65ED, 0xC6B1 },
 { 0x65F1, 0xD55C },
 { 0x65F4, 0xC6B0 },
 { 0x65FA, 0xC655 },
 { 0x65FB, 0xBBFC },
 { 0x65FC, 0xBBFC },
 { 0x65FD, 0xB3C8 },
 { 0x65FF, 0xC624 },
 { 0x6606, 0xACE4 },
 { 0x6607, 0xC2B9 },
 { 0x6609, 0xBC29 },
 { 0x660A, 0xD638 },
 { 0x660C, 0xCC3D },
 { 0x660E, 0xBA85 },
 { 0x660F, 0xD63C },
 { 0x6610, 0xBD84 },
 { 0x6611, 0xAE08 },
 { 0x6613, 0xC5ED },
 { 0x6614, 0xC11D },
 { 0x6615, 0xD754 },
 { 0x661E, 0xBCD1 },
 { 0x661F, 0xC131 },
 { 0x6620, 0xC601 },
 { 0x6625, 0xCD98 },
 { 0x6627, 0xB9E4 },
 { 0x6628, 0xC791 },
 { 0x662D, 0xC18C },
 { 0x662F, 0xC2DC },
 { 0x6630, 0xD558 },
 { 0x6631, 0xC6B1 },
 { 0x6634, 0xBB18 },
 { 0x6636, 0xCC3D },
 { 0x663A, 0xBCD1 },
 { 0x663B, 0xC559 },
 { 0x6641, 0xC870 },
 { 0x6642, 0xC2DC },
 { 0x6643, 0xD669 },
 { 0x6644, 0xD669 },
 { 0x6649, 0xC9C4 },
 { 0x664B, 0xC9C4 },
 { 0x664F, 0xC548 },
 { 0x6659, 0xC900 },
 { 0x665B, 0xD604 },
 { 0x665D, 0xC8FC },
 { 0x665E, 0xD76C },
 { 0x665F, 0xC131 },
 { 0x6664, 0xC624 },
 { 0x6665, 0xD658 },
 { 0x6666, 0xD68C },
 { 0x6667, 0xD638 },
 { 0x6668, 0xC2E0 },
 { 0x6669, 0xB9CC },
 { 0x666B, 0xD0C1 },
 { 0x666E, 0xBCF4 },
 { 0x666F, 0xACBD },
 { 0x6673, 0xC11D },
 { 0x6674, 0xCCAD },
 { 0x6676, 0xC815 },
 { 0x6677, 0xADC0 },
 { 0x6678, 0xC815 },
 { 0x667A, 0xC9C0 },
 { 0x6684, 0xD6E4 },
 { 0x6687, 0xAC00 },
 { 0x6688, 0xC6B4 },
 { 0x6689, 0xD718 },
 { 0x668E, 0xC601 },
 { 0x6690, 0xC704 },
 { 0x6691, 0xC11C },
 { 0x6696, 0xB09C },
 { 0x6697, 0xC554 },
 { 0x6698, 0xC591 },
 { 0x669D, 0xBA85 },
 { 0x66A0, 0xACE0 },
 { 0x66A2, 0xCC3D },
 { 0x66AB, 0xC7A0 },
 { 0x66AE, 0xBAA8 },
 { 0x66B2, 0xC7A5 },
 { 0x66B3, 0xD61C },
 { 0x66B4, 0xD3EC },
 { 0x66B9, 0xC12C },
 { 0x66BB, 0xACBD },
 { 0x66BE, 0xB3C8 },
 { 0x66C4, 0xC5FD },
 { 0x66C6, 0xB825 },
 { 0x66C7, 0xB2F4 },
 { 0x66C9, 0xD6A8 },
 { 0x66D6, 0xC560 },
 { 0x66D9, 0xC11C },
 { 0x66DC, 0xC694 },
 { 0x66DD, 0xD3ED },
 { 0x66E0, 0xAD11 },
 { 0x66E6, 0xD76C },
 { 0x66F0, 0xC648 },
 { 0x66F2, 0xACE1 },
 { 0x66F3, 0xC608 },
 { 0x66F4, 0xAC31 },
 { 0x66F7, 0xAC08 },
 { 0x66F8, 0xC11C },
 { 0x66F9, 0xC870 },
 { 0x66FA, 0xC870 },
 { 0x66FC, 0xB9CC },
 { 0x66FE, 0xC99D },
 { 0x66FF, 0xCCB4 },
 { 0x6700, 0xCD5C },
 { 0x6703, 0xD68C },
 { 0x6708, 0xC6D4 },
 { 0x6709, 0xC720 },
 { 0x670B, 0xBD95 },
 { 0x670D, 0xBCF5 },
 { 0x6714, 0xC0AD },
 { 0x6715, 0xC9D0 },
 { 0x6717, 0xB0AD },
 { 0x671B, 0xB9DD },
 { 0x671D, 0xC870 },
 { 0x671E, 0xAE30 },
 { 0x671F, 0xAE30 },
 { 0x6726, 0xBABD },
 { 0x6727, 0xB871 },
 { 0x6728, 0xBAA9 },
 { 0x672A, 0xBBF8 },
 { 0x672B, 0xB9D0 },
 { 0x672C, 0xBCF8 },
 { 0x672D, 0xCC30 },
 { 0x672E, 0xCD9C },
 { 0x6731, 0xC8FC },
 { 0x6734, 0xBC15 },
 { 0x6736, 0xD0C0 },
 { 0x673A, 0xADA4 },
 { 0x673D, 0xD6C4 },
 { 0x6746, 0xAC04 },
 { 0x6749, 0xC0BC },
 { 0x674E, 0xB9AC },
 { 0x674F, 0xD589 },
 { 0x6750, 0xC7AC },
 { 0x6751, 0xCD0C },
 { 0x6753, 0xD45C },
 { 0x6756, 0xC7A5 },
 { 0x675C, 0xB450 },
 { 0x675E, 0xAE30 },
 { 0x675F, 0xC18D },
 { 0x676D, 0xD56D },
 { 0x676F, 0xBC30 },
 { 0x6770, 0xAC78 },
 { 0x6771, 0xB3D9 },
 { 0x6773, 0xBB18 },
 { 0x6775, 0xC800 },
 { 0x6777, 0xD30C },
 { 0x677B, 0xB274 },
 { 0x677E, 0xC1A1 },
 { 0x677F, 0xD310 },
 { 0x6787, 0xBE44 },
 { 0x6789, 0xC655 },
 { 0x678B, 0xBC29 },
 { 0x678F, 0xB0A8 },
 { 0x6790, 0xC11D },
 { 0x6793, 0xB450 },
 { 0x6795, 0xCE68 },
 { 0x6797, 0xB9BC },
 { 0x679A, 0xB9E4 },
 { 0x679C, 0xACFC },
 { 0x679D, 0xC9C0 },
 { 0x67AF, 0xACE0 },
 { 0x67B0, 0xD3C9 },
 { 0x67B3, 0xC9C0 },
 { 0x67B6, 0xAC00 },
 { 0x67B7, 0xAC00 },
 { 0x67B8, 0xAD6C },
 { 0x67BE, 0xC2DC },
 { 0x67C4, 0xBCD1 },
 { 0x67CF, 0xBC31 },
 { 0x67D0, 0xBAA8 },
 { 0x67D1, 0xAC10 },
 { 0x67D2, 0xCE60 },
 { 0x67D3, 0xC5FC },
 { 0x67D4, 0xC720 },
 { 0x67DA, 0xC720 },
 { 0x67DD, 0xD0C1 },
 { 0x67E9, 0xAD6C },
 { 0x67EC, 0xAC04 },
 { 0x67EF, 0xAC00 },
 { 0x67F0, 0xB0B4 },
 { 0x67F1, 0xC8FC },
 { 0x67F3, 0xB958 },
 { 0x67F4, 0xC2DC },
 { 0x67F5, 0xCC45 },
 { 0x67F6, 0xC0AC },
 { 0x67FB, 0xC0AC },
 { 0x67FE, 0xC815 },
 { 0x6812, 0xC21C },
 { 0x6813, 0xC804 },
 { 0x6816, 0xC11C },
 { 0x6817, 0xB960 },
 { 0x6821, 0xAD50 },
 { 0x6822, 0xBC31 },
 { 0x682A, 0xC8FC },
 { 0x682F, 0xC6B1 },
 { 0x6838, 0xD575 },
 { 0x6839, 0xADFC },
 { 0x683C, 0xACA9 },
 { 0x683D, 0xC7AC },
 { 0x6840, 0xAC78 },
 { 0x6841, 0xD56D },
 { 0x6842, 0xACC4 },
 { 0x6843, 0xB3C4 },
 { 0x6848, 0xC548 },
 { 0x684E, 0xC9C8 },
 { 0x6850, 0xB3D9 },
 { 0x6851, 0xC0C1 },
 { 0x6853, 0xD658 },
 { 0x6854, 0xAE38 },
 { 0x686D, 0xC9C4 },
 { 0x6876, 0xD1B5 },
 { 0x687F, 0xAC04 },
 { 0x6881, 0xB7C9 },
 { 0x6885, 0xB9E4 },
 { 0x688F, 0xACE1 },
 { 0x6893, 0xC7AC },
 { 0x6894, 0xCE58 },
 { 0x6897, 0xACBD },
 { 0x689D, 0xC870 },
 { 0x689F, 0xD6A8 },
 { 0x68A1, 0xC644 },
 { 0x68A2, 0xCD08 },
 { 0x68A7, 0xC624 },
 { 0x68A8, 0xB9AC },
 { 0x68AD, 0xC0AC },
 { 0x68AF, 0xC81C },
 { 0x68B0, 0xACC4 },
 { 0x68B1, 0xACE4 },
 { 0x68B3, 0xC18C },
 { 0x68B5, 0xBC94 },
 { 0x68B6, 0xBBF8 },
 { 0x68C4, 0xAE30 },
 { 0x68C5, 0xBCD1 },
 { 0x68C9, 0xBA74 },
 { 0x68CB, 0xAE30 },
 { 0x68CD, 0xACE4 },
 { 0x68D2, 0xBD09 },
 { 0x68D5, 0xC885 },
 { 0x68D7, 0xC870 },
 { 0x68D8, 0xADF9 },
 { 0x68DA, 0xBD95 },
 { 0x68DF, 0xB3D9 },
 { 0x68E0, 0xB2F9 },
 { 0x68E7, 0xC794 },
 { 0x68E8, 0xACC4 },
 { 0x68EE, 0xC0BC },
 { 0x68F2, 0xC11C },
 { 0x68F9, 0xB3C4 },
 { 0x68FA, 0xAD00 },
 { 0x6900, 0xC644 },
 { 0x6905, 0xC758 },
 { 0x690D, 0xC2DD },
 { 0x690E, 0xCD94 },
 { 0x6912, 0xCD08 },
 { 0x6927, 0xBA85 },
 { 0x6930, 0xC57C },
 { 0x693D, 0xC5F0 },
 { 0x693F, 0xCD98 },
 { 0x694A, 0xC591 },
 { 0x6953, 0xD48D },
 { 0x6954, 0xC124 },
 { 0x6955, 0xD0C0 },
 { 0x6957, 0xAC74 },
 { 0x6959, 0xBB34 },
 { 0x695A, 0xCD08 },
 { 0x695E, 0xB989 },
 { 0x6960, 0xB0A8 },
 { 0x6961, 0xC720 },
 { 0x6962, 0xC720 },
 { 0x6963, 0xBBF8 },
 { 0x6968, 0xC815 },
 { 0x696B, 0xC999 },
 { 0x696D, 0xC5C5 },
 { 0x696E, 0xC800 },
 { 0x696F, 0xC21C },
 { 0x6975, 0xADF9 },
 { 0x6977, 0xD574 },
 { 0x6978, 0xCD94 },
 { 0x6979, 0xC601 },
 { 0x6995, 0xC6A9 },
 { 0x699B, 0xC9C4 },
 { 0x699C, 0xBC29 },
 { 0x69A5, 0xD669 },
 { 0x69A7, 0xBE44 },
 { 0x69AE, 0xC601 },
 { 0x69B4, 0xB958 },
 { 0x69BB, 0xD0D1 },
 { 0x69C1, 0xACE0 },
 { 0x69C3, 0xBC18 },
 { 0x69CB, 0xAD6C },
 { 0x69CC, 0xD1F4 },
 { 0x69CD, 0xCC3D },
 { 0x69D0, 0xAD34 },
 { 0x69E8, 0xACFD },
 { 0x69EA, 0xAC1C },
 { 0x69FB, 0xADDC },
 { 0x69FD, 0xC870 },
 { 0x69FF, 0xADFC },
 { 0x6A02, 0xB099 },
 { 0x6A0A, 0xBC88 },
 { 0x6A11, 0xB7C9 },
 { 0x6A13, 0xB204 },
 { 0x6A17, 0xC800 },
 { 0x6A19, 0xD45C },
 { 0x6A1E, 0xCD94 },
 { 0x6A1F, 0xC7A5 },
 { 0x6A21, 0xBAA8 },
 { 0x6A23, 0xC591 },
 { 0x6A35, 0xCD08 },
 { 0x6A38, 0xBC15 },
 { 0x6A39, 0xC218 },
 { 0x6A3A, 0xD654 },
 { 0x6A3D, 0xC900 },
 { 0x6A44, 0xAC10 },
 { 0x6A48, 0xC694 },
 { 0x6A4B, 0xAD50 },
 { 0x6A52, 0xC6B4 },
 { 0x6A53, 0xC21C },
 { 0x6A58, 0xADE4 },
 { 0x6A59, 0xB4F1 },
 { 0x6A5F, 0xAE30 },
 { 0x6A61, 0xC0C1 },
 { 0x6A6B, 0xD6A1 },
 { 0x6A80, 0xB2E8 },
 { 0x6A84, 0xACA9 },
 { 0x6A89, 0xC815 },
 { 0x6A8D, 0xC5B5 },
 { 0x6A8E, 0xAE08 },
 { 0x6A97, 0xBCBD },
 { 0x6A9C, 0xD68C },
 { 0x6AA2, 0xAC80 },
 { 0x6AA3, 0xC7A5 },
 { 0x6AB3, 0xBE48 },
 { 0x6ABB, 0xD568 },
 { 0x6AC2, 0xB3C4 },
 { 0x6AC3, 0xADA4 },
 { 0x6AD3, 0xB178 },
 { 0x6ADA, 0xB824 },
 { 0x6ADB, 0xC990 },
 { 0x6AF6, 0xD5CC },
 { 0x6AFB, 0xC575 },
 { 0x6B04, 0xB09C },
 { 0x6B0A, 0xAD8C },
 { 0x6B0C, 0xC7A5 },
 { 0x6B12, 0xB780 },
 { 0x6B16, 0xB78C },
 { 0x6B20, 0xD760 },
 { 0x6B21, 0xCC28 },
 { 0x6B23, 0xD754 },
 { 0x6B32, 0xC695 },
 { 0x6B3A, 0xAE30 },
 { 0x6B3D, 0xD760 },
 { 0x6B3E, 0xAD00 },
 { 0x6B46, 0xD760 },
 { 0x6B47, 0xD5D0 },
 { 0x6B4C, 0xAC00 },
 { 0x6B4E, 0xD0C4 },
 { 0x6B50, 0xAD6C },
 { 0x6B5F, 0xC5EC },
 { 0x6B61, 0xD658 },
 { 0x6B62, 0xC9C0 },
 { 0x6B63, 0xC815 },
 { 0x6B64, 0xCC28 },
 { 0x6B65, 0xBCF4 },
 { 0x6B66, 0xBB34 },
 { 0x6B6A, 0xC65C },
 { 0x6B72, 0xC138 },
 { 0x6B77, 0xB825 },
 { 0x6B78, 0xADC0 },
 { 0x6B7B, 0xC0AC },
 { 0x6B7F, 0xBAB0 },
 { 0x6B83, 0xC559 },
 { 0x6B84, 0xC9C4 },
 { 0x6B86, 0xD0DC },
 { 0x6B89, 0xC21C },
 { 0x6B8A, 0xC218 },
 { 0x6B96, 0xC2DD },
 { 0x6B98, 0xC794 },
 { 0x6B9E, 0xC6B4 },
 { 0x6BAE, 0xB834 },
 { 0x6BAF, 0xBE48 },
 { 0x6BB2, 0xC12C },
 { 0x6BB5, 0xB2E8 },
 { 0x6BB7, 0xC740 },
 { 0x6BBA, 0xC0B4 },
 { 0x6BBC, 0xAC01 },
 { 0x6BBF, 0xC804 },
 { 0x6BC1, 0xD6FC },
 { 0x6BC5, 0xC758 },
 { 0x6BC6, 0xAD6C },
 { 0x6BCB, 0xBB34 },
 { 0x6BCD, 0xBAA8 },
 { 0x6BCF, 0xB9E4 },
 { 0x6BD2, 0xB3C5 },
 { 0x6BD3, 0xC721 },
 { 0x6BD4, 0xBE44 },
 { 0x6BD6, 0xBE44 },
 { 0x6BD7, 0xBE44 },
 { 0x6BD8, 0xBE44 },
 { 0x6BDB, 0xBAA8 },
 { 0x6BEB, 0xD638 },
 { 0x6BEC, 0xAD6C },
 { 0x6C08, 0xC804 },
 { 0x6C0F, 0xC528 },
 { 0x6C11, 0xBBFC },
 { 0x6C13, 0xB9F9 },
 { 0x6C23, 0xAE30 },
 { 0x6C34, 0xC218 },
 { 0x6C37, 0xBE59 },
 { 0x6C38, 0xC601 },
 { 0x6C3E, 0xBC94 },
 { 0x6C40, 0xC815 },
 { 0x6C41, 0xC999 },
 { 0x6C42, 0xAD6C },
 { 0x6C4E, 0xBC94 },
 { 0x6C50, 0xC11D },
 { 0x6C55, 0xC0B0 },
 { 0x6C57, 0xD55C },
 { 0x6C5A, 0xC624 },
 { 0x6C5D, 0xC5EC },
 { 0x6C5E, 0xD64D },
 { 0x6C5F, 0xAC15 },
 { 0x6C60, 0xC9C0 },
 { 0x6C68, 0xACE8 },
 { 0x6C6A, 0xC655 },
 { 0x6C6D, 0xC608 },
 { 0x6C70, 0xD0DC },
 { 0x6C72, 0xAE09 },
 { 0x6C76, 0xBB38 },
 { 0x6C7A, 0xACB0 },
 { 0x6C7D, 0xAE30 },
 { 0x6C7E, 0xBD84 },
 { 0x6C81, 0xC2EC },
 { 0x6C82, 0xAE30 },
 { 0x6C83, 0xC625 },
 { 0x6C85, 0xC6D0 },
 { 0x6C86, 0xD56D },
 { 0x6C87, 0xC5F0 },
 { 0x6C88, 0xC2EC },
 { 0x6C8C, 0xB3C8 },
 { 0x6C90, 0xBAA9 },
 { 0x6C92, 0xBAB0 },
 { 0x6C93, 0xB2F5 },
 { 0x6C94, 0xBA74 },
 { 0x6C95, 0xBB3C },
 { 0x6C96, 0xCDA9 },
 { 0x6C99, 0xC0AC },
 { 0x6C9A, 0xC9C0 },
 { 0x6C9B, 0xD328 },
 { 0x6CAB, 0xB9D0 },
 { 0x6CAE, 0xC800 },
 { 0x6CB3, 0xD558 },
 { 0x6CB8, 0xBE44 },
 { 0x6CB9, 0xC720 },
 { 0x6CBB, 0xCE58 },
 { 0x6CBC, 0xC18C },
 { 0x6CBD, 0xACE0 },
 { 0x6CBE, 0xCCA8 },
 { 0x6CBF, 0xC5F0 },
 { 0x6CC1, 0xD669 },
 { 0x6CC2, 0xD615 },
 { 0x6CC4, 0xC124 },
 { 0x6CC9, 0xCC9C },
 { 0x6CCA, 0xBC15 },
 { 0x6CCC, 0xBE44 },
 { 0x6CD3, 0xD64D },
 { 0x6CD5, 0xBC95 },
 { 0x6CD7, 0xC0AC },
 { 0x6CDB, 0xBC94 },
 { 0x6CE1, 0xD3EC },
 { 0x6CE2, 0xD30C },
 { 0x6CE3, 0xC74D },
 { 0x6CE5, 0xB2C8 },
 { 0x6CE8, 0xC8FC },
 { 0x6CEB, 0xD604 },
 { 0x6CEE, 0xBC18 },
 { 0x6CEF, 0xBBFC },
 { 0x6CF0, 0xD0DC },
 { 0x6CF3, 0xC601 },
 { 0x6D0B, 0xC591 },
 { 0x6D0C, 0xB82C },
 { 0x6D11, 0xBCF4 },
 { 0x6D17, 0xC138 },
 { 0x6D19, 0xC218 },
 { 0x6D1B, 0xB099 },
 { 0x6D1E, 0xB3D9 },
 { 0x6D25, 0xC9C4 },
 { 0x6D27, 0xC720 },
 { 0x6D29, 0xC124 },
 { 0x6D2A, 0xD64D },
 { 0x6D32, 0xC8FC },
 { 0x6D35, 0xC21C },
 { 0x6D36, 0xD749 },
 { 0x6D38, 0xAD11 },
 { 0x6D39, 0xC6D0 },
 { 0x6D3B, 0xD65C },
 { 0x6D3D, 0xD761 },
 { 0x6D3E, 0xD30C },
 { 0x6D41, 0xB958 },
 { 0x6D59, 0xC808 },
 { 0x6D5A, 0xC900 },
 { 0x6D5C, 0xBE48 },
 { 0x6D63, 0xC644 },
 { 0x6D66, 0xD3EC },
 { 0x6D69, 0xD638 },
 { 0x6D6A, 0xB0AD },
 { 0x6D6C, 0xB9AC },
 { 0x6D6E, 0xBD80 },
 { 0x6D74, 0xC695 },
 { 0x6D77, 0xD574 },
 { 0x6D78, 0xCE68 },
 { 0x6D79, 0xD611 },
 { 0x6D7F, 0xD328 },
 { 0x6D85, 0xC5F4 },
 { 0x6D87, 0xACBD },
 { 0x6D88, 0xC18C },
 { 0x6D89, 0xC12D },
 { 0x6D8C, 0xC6A9 },
 { 0x6D8D, 0xD6A8 },
 { 0x6D8E, 0xC5F0 },
 { 0x6D91, 0xC18D },
 { 0x6D93, 0xC5F0 },
 { 0x6D95, 0xCCB4 },
 { 0x6DAF, 0xC560 },
 { 0x6DB2, 0xC561 },
 { 0x6DB5, 0xD568 },
 { 0x6DC0, 0xC815 },
 { 0x6DC3, 0xAD8C },
 { 0x6DC4, 0xCE58 },
 { 0x6DC5, 0xC11D },
 { 0x6DC6, 0xD6A8 },
 { 0x6DC7, 0xAE30 },
 { 0x6DCB, 0xB9BC },
 { 0x6DCF, 0xD638 },
 { 0x6DD1, 0xC219 },
 { 0x6DD8, 0xB3C4 },
 { 0x6DD9, 0xC885 },
 { 0x6DDA, 0xB204 },
 { 0x6DDE, 0xC1A1 },
 { 0x6DE1, 0xB2F4 },
 { 0x6DE8, 0xC815 },
 { 0x6DEA, 0xB95C },
 { 0x6DEB, 0xC74C },
 { 0x6DEE, 0xD68C },
 { 0x6DF1, 0xC2EC },
 { 0x6DF3, 0xC21C },
 { 0x6DF5, 0xC5F0 },
 { 0x6DF7, 0xD63C },
 { 0x6DF8, 0xCCAD },
 { 0x6DF9, 0xC5C4 },
 { 0x6DFA, 0xCC9C },
 { 0x6DFB, 0xCCA8 },
 { 0x6E17, 0xC0BC },
 { 0x6E19, 0xD658 },
 { 0x6E1A, 0xC800 },
 { 0x6E1B, 0xAC10 },
 { 0x6E1F, 0xC815 },
 { 0x6E20, 0xAC70 },
 { 0x6E21, 0xB3C4 },
 { 0x6E23, 0xC0AC },
 { 0x6E24, 0xBC1C },
 { 0x6E25, 0xC545 },
 { 0x6E26, 0xC640 },
 { 0x6E2B, 0xC124 },
 { 0x6E2C, 0xCE21 },
 { 0x6E2D, 0xC704 },
 { 0x6E2F, 0xD56D },
 { 0x6E32, 0xC120 },
 { 0x6E34, 0xAC08 },
 { 0x6E36, 0xC601 },
 { 0x6E38, 0xC720 },
 { 0x6E3A, 0xBB18 },
 { 0x6E3C, 0xBBF8 },
 { 0x6E3D, 0xC7AC },
 { 0x6E3E, 0xD63C },
 { 0x6E43, 0xBC30 },
 { 0x6E44, 0xBBF8 },
 { 0x6E4A, 0xC8FC },
 { 0x6E4D, 0xB2E8 },
 { 0x6E56, 0xD638 },
 { 0x6E58, 0xC0C1 },
 { 0x6E5B, 0xB2F4 },
 { 0x6E5C, 0xC2DD },
 { 0x6E5E, 0xC815 },
 { 0x6E5F, 0xD669 },
 { 0x6E67, 0xC6A9 },
 { 0x6E6B, 0xCD94 },
 { 0x6E6E, 0xC778 },
 { 0x6E6F, 0xD0D5 },
 { 0x6E72, 0xC6D0 },
 { 0x6E73, 0xB0A8 },
 { 0x6E7A, 0xBCF4 },
 { 0x6E90, 0xC6D0 },
 { 0x6E96, 0xC900 },
 { 0x6E9C, 0xB958 },
 { 0x6E9D, 0xAD6C },
 { 0x6E9F, 0xBA85 },
 { 0x6EA2, 0xC77C },
 { 0x6EA5, 0xBD80 },
 { 0x6EAA, 0xACC4 },
 { 0x6EAB, 0xC628 },
 { 0x6EAF, 0xC18C },
 { 0x6EB1, 0xC9C4 },
 { 0x6EB6, 0xC6A9 },
 { 0x6EBA, 0xB2C9 },
 { 0x6EC2, 0xBC29 },
 { 0x6EC4, 0xCC3D },
 { 0x6EC5, 0xBA78 },
 { 0x6EC9, 0xD669 },
 { 0x6ECB, 0xC790 },
 { 0x6ECC, 0xCC99 },
 { 0x6ECE, 0xD615 },
 { 0x6ED1, 0xACE8 },
 { 0x6ED3, 0xC7AC },
 { 0x6ED4, 0xB3C4 },
 { 0x6EEF, 0xCCB4 },
 { 0x6EF4, 0xC801 },
 { 0x6EF8, 0xD638 },
 { 0x6EFE, 0xACE4 },
 { 0x6EFF, 0xB9CC },
 { 0x6F01, 0xC5B4 },
 { 0x6F02, 0xD45C },
 { 0x6F06, 0xCE60 },
 { 0x6F0F, 0xB204 },
 { 0x6F11, 0xAC1C },
 { 0x6F14, 0xC5F0 },
 { 0x6F15, 0xC870 },
 { 0x6F20, 0xB9C9 },
 { 0x6F22, 0xD55C },
 { 0x6F23, 0xB828 },
 { 0x6F2B, 0xB9CC },
 { 0x6F2C, 0xC9C0 },
 { 0x6F31, 0xC218 },
 { 0x6F32, 0xCC3D },
 { 0x6F38, 0xC810 },
 { 0x6F3F, 0xC7A5 },
 { 0x6F41, 0xC601 },
 { 0x6F51, 0xBC1C },
 { 0x6F54, 0xACB0 },
 { 0x6F57, 0xC9D1 },
 { 0x6F58, 0xBC18 },
 { 0x6F5A, 0xC219 },
 { 0x6F5B, 0xC7A0 },
 { 0x6F5E, 0xB85C },
 { 0x6F5F, 0xC11D },
 { 0x6F62, 0xD669 },
 { 0x6F64, 0xC724 },
 { 0x6F6D, 0xB2F4 },
 { 0x6F6E, 0xC870 },
 { 0x6F70, 0xADA4 },
 { 0x6F7A, 0xC794 },
 { 0x6F7C, 0xB3D9 },
 { 0x6F7D, 0xBCF4 },
 { 0x6F7E, 0xB9B0 },
 { 0x6F81, 0xC0BD },
 { 0x6F84, 0xC9D5 },
 { 0x6F88, 0xCCA0 },
 { 0x6F8D, 0xC8FC },
 { 0x6F8E, 0xD33D },
 { 0x6F90, 0xC6B4 },
 { 0x6F94, 0xD638 },
 { 0x6F97, 0xAC04 },
 { 0x6FA3, 0xD55C },
 { 0x6FA4, 0xD0DD },
 { 0x6FA7, 0xB840 },
 { 0x6FAE, 0xD68C },
 { 0x6FAF, 0xCC2C },
 { 0x6FB1, 0xC804 },
 { 0x6FB3, 0xC624 },
 { 0x6FB9, 0xB2F4 },
 { 0x6FBE, 0xB2EC },
 { 0x6FC0, 0xACA9 },
 { 0x6FC1, 0xD0C1 },
 { 0x6FC2, 0xB834 },
 { 0x6FC3, 0xB18D },
 { 0x6FCA, 0xC608 },
 { 0x6FD5, 0xC2B5 },
 { 0x6FDA, 0xC601 },
 { 0x6FDF, 0xC81C },
 { 0x6FE0, 0xD638 },
 { 0x6FE1, 0xC720 },
 { 0x6FE4, 0xB3C4 },
 { 0x6FE9, 0xD638 },
 { 0x6FEB, 0xB0A8 },
 { 0x6FEC, 0xC900 },
 { 0x6FEF, 0xD0C1 },
 { 0x6FF1, 0xBE48 },
 { 0x6FFE, 0xB824 },
 { 0x7001, 0xC591 },
 { 0x7005, 0xD615 },
 { 0x7006, 0xB3C5 },
 { 0x7009, 0xC0AC },
 { 0x700B, 0xC2EC },
 { 0x700F, 0xB958 },
 { 0x7011, 0xD3ED },
 { 0x7015, 0xBE48 },
 { 0x7018, 0xB85C },
 { 0x701A, 0xD55C },
 { 0x701B, 0xC601 },
 { 0x701C, 0xC735 },
 { 0x701D, 0xB825 },
 { 0x701E, 0xC815 },
 { 0x701F, 0xC18C },
 { 0x7023, 0xD574 },
 { 0x7027, 0xB871 },
 { 0x7028, 0xB8B0 },
 { 0x702F, 0xC601 },
 { 0x7037, 0xC775 },
 { 0x703E, 0xB780 },
 { 0x704C, 0xAD00 },
 { 0x7050, 0xD615 },
 { 0x7051, 0xC1C4 },
 { 0x7058, 0xD0C4 },
 { 0x705D, 0xD638 },
 { 0x7063, 0xB9CC },
 { 0x706B, 0xD654 },
 { 0x7070, 0xD68C },
 { 0x7078, 0xAD6C },
 { 0x707C, 0xC791 },
 { 0x707D, 0xC7AC },
 { 0x7085, 0xACBD },
 { 0x708A, 0xCDE8 },
 { 0x708E, 0xC5FC },
 { 0x7092, 0xCD08 },
 { 0x7098, 0xD754 },
 { 0x7099, 0xC790 },
 { 0x709A, 0xAD11 },
 { 0x70A1, 0xC815 },
 { 0x70A4, 0xC18C },
 { 0x70AB, 0xD604 },
 { 0x70AC, 0xAC70 },
 { 0x70AD, 0xD0C4 },
 { 0x70AF, 0xD615 },
 { 0x70B3, 0xBCD1 },
 { 0x70B7, 0xC8FC },
 { 0x70B8, 0xC791 },
 { 0x70B9, 0xC810 },
 { 0x70C8, 0xB82C },
 { 0x70CB, 0xD734 },
 { 0x70CF, 0xC624 },
 { 0x70D8, 0xD64D },
 { 0x70D9, 0xB099 },
 { 0x70DD, 0xC99D },
 { 0x70DF, 0xC5F0 },
 { 0x70F1, 0xACBD },
 { 0x70F9, 0xD33D },
 { 0x70FD, 0xBD09 },
 { 0x7104, 0xD6C8 },
 { 0x7109, 0xC5B8 },
 { 0x710C, 0xC900 },
 { 0x7119, 0xBC30 },
 { 0x711A, 0xBD84 },
 { 0x711E, 0xB3C8 },
 { 0x7121, 0xBB34 },
 { 0x7126, 0xCD08 },
 { 0x7130, 0xC5FC },
 { 0x7136, 0xC5F0 },
 { 0x7147, 0xD718 },
 { 0x7149, 0xB828 },
 { 0x714A, 0xD6E4 },
 { 0x714C, 0xD669 },
 { 0x714E, 0xC804 },
 { 0x7150, 0xC601 },
 { 0x7156, 0xB09C },
 { 0x7159, 0xC5F0 },
 { 0x715C, 0xC6B1 },
 { 0x715E, 0xC0B4 },
 { 0x7164, 0xB9E4 },
 { 0x7165, 0xD658 },
 { 0x7166, 0xD6C4 },
 { 0x7167, 0xC870 },
 { 0x7169, 0xBC88 },
 { 0x716C, 0xC591 },
 { 0x716E, 0xC790 },
 { 0x717D, 0xC120 },
 { 0x7184, 0xC2DD },
 { 0x7189, 0xC6B4 },
 { 0x718A, 0xC6C5 },
 { 0x718F, 0xD6C8 },
 { 0x7192, 0xD615 },
 { 0x7194, 0xC6A9 },
 { 0x7199, 0xD76C },
 { 0x719F, 0xC219 },
 { 0x71A2, 0xBD09 },
 { 0x71AC, 0xC624 },
 { 0x71B1, 0xC5F4 },
 { 0x71B9, 0xD76C },
 { 0x71BA, 0xD76C },
 { 0x71BE, 0xCE58 },
 { 0x71C1, 0xC5FD },
 { 0x71C3, 0xC5F0 },
 { 0x71C8, 0xB4F1 },
 { 0x71C9, 0xB3C8 },
 { 0x71CE, 0xB8CC },
 { 0x71D0, 0xB9B0 },
 { 0x71D2, 0xC18C },
 { 0x71D4, 0xBC88 },
 { 0x71D5, 0xC5F0 },
 { 0x71DF, 0xC601 },
 { 0x71E5, 0xC870 },
 { 0x71E6, 0xCC2C },
 { 0x71E7, 0xC218 },
 { 0x71ED, 0xCD09 },
 { 0x71EE, 0xC12D },
 { 0x71FB, 0xD6C8 },
 { 0x71FC, 0xC2E0 },
 { 0x71FE, 0xB3C4 },
 { 0x71FF, 0xC694 },
 { 0x7200, 0xD601 },
 { 0x7206, 0xD3ED },
 { 0x7210, 0xB178 },
 { 0x721B, 0xB09C },
 { 0x722A, 0xC870 },
 { 0x722C, 0xD30C },
 { 0x722D, 0xC7C1 },
 { 0x7230, 0xC6D0 },
 { 0x7232, 0xC704 },
 { 0x7235, 0xC791 },
 { 0x7236, 0xBD80 },
 { 0x723A, 0xC57C },
 { 0x723B, 0xD6A8 },
 { 0x723D, 0xC0C1 },
 { 0x723E, 0xC774 },
 { 0x7240, 0xC0C1 },
 { 0x7246, 0xC7A5 },
 { 0x7247, 0xD3B8 },
 { 0x7248, 0xD310 },
 { 0x724C, 0xD328 },
 { 0x7252, 0xCCA9 },
 { 0x7258, 0xB3C5 },
 { 0x7259, 0xC544 },
 { 0x725B, 0xC6B0 },
 { 0x725D, 0xBE48 },
 { 0x725F, 0xBAA8 },
 { 0x7261, 0xBAA8 },
 { 0x7262, 0xB1CC },
 { 0x7267, 0xBAA9 },
 { 0x7269, 0xBB3C },
 { 0x7272, 0xC0DD },
 { 0x7279, 0xD2B9 },
 { 0x727D, 0xACAC },
 { 0x7280, 0xC11C },
 { 0x7281, 0xB9AC },
 { 0x72A2, 0xB3C5 },
 { 0x72A7, 0xD76C },
 { 0x72AC, 0xACAC },
 { 0x72AF, 0xBC94 },
 { 0x72C0, 0xC0C1 },
 { 0x72C2, 0xAD11 },
 { 0x72C4, 0xC801 },
 { 0x72CE, 0xC555 },
 { 0x72D0, 0xD638 },
 { 0x72D7, 0xAD6C },
 { 0x72D9, 0xC800 },
 { 0x72E1, 0xAD50 },
 { 0x72E9, 0xC218 },
 { 0x72F8, 0xB9AC },
 { 0x72F9, 0xD611 },
 { 0x72FC, 0xB0AD },
 { 0x72FD, 0xD328 },
 { 0x730A, 0xC608 },
 { 0x7316, 0xCC3D },
 { 0x731B, 0xB9F9 },
 { 0x731C, 0xC2DC },
 { 0x731D, 0xC878 },
 { 0x7325, 0xC678 },
 { 0x7329, 0xC131 },
 { 0x732A, 0xC800 },
 { 0x732B, 0xBB18 },
 { 0x7336, 0xC720 },
 { 0x7337, 0xC720 },
 { 0x733E, 0xD65C },
 { 0x733F, 0xC6D0 },
 { 0x7344, 0xC625 },
 { 0x7345, 0xC0AC },
 { 0x7350, 0xC7A5 },
 { 0x7352, 0xC624 },
 { 0x7357, 0xAD90 },
 { 0x7368, 0xB3C5 },
 { 0x736A, 0xD68C },
 { 0x7370, 0xC601 },
 { 0x7372, 0xD68D },
 { 0x7375, 0xB835 },
 { 0x7378, 0xC218 },
 { 0x737A, 0xB2EC },
 { 0x737B, 0xD5CC },
 { 0x7384, 0xD604 },
 { 0x7386, 0xC790 },
 { 0x7387, 0xB960 },
 { 0x7389, 0xC625 },
 { 0x738B, 0xC655 },
 { 0x738E, 0xC815 },
 { 0x7394, 0xCC9C },
 { 0x7396, 0xAD6C },
 { 0x7397, 0xC6B0 },
 { 0x7398, 0xAE30 },
 { 0x739F, 0xBBFC },
 { 0x73A7, 0xC724 },
 { 0x73A9, 0xC644 },
 { 0x73AD, 0xBE48 },
 { 0x73B2, 0xB839 },
 { 0x73B3, 0xB300 },
 { 0x73B9, 0xD604 },
 { 0x73C0, 0xBC15 },
 { 0x73C2, 0xAC00 },
 { 0x73C9, 0xBBFC },
 { 0x73CA, 0xC0B0 },
 { 0x73CC, 0xD544 },
 { 0x73CD, 0xC9C4 },
 { 0x73CF, 0xAC01 },
 { 0x73D6, 0xAD11 },
 { 0x73D9, 0xACF5 },
 { 0x73DD, 0xD6C4 },
 { 0x73DE, 0xB099 },
 { 0x73E0, 0xC8FC },
 { 0x73E3, 0xC21C },
 { 0x73E4, 0xBCF4 },
 { 0x73E5, 0xC774 },
 { 0x73E6, 0xD5A5 },
 { 0x73E9, 0xD615 },
 { 0x73EA, 0xADDC },
 { 0x73ED, 0xBC18 },
 { 0x73F7, 0xBB34 },
 { 0x73F9, 0xC131 },
 { 0x73FD, 0xC815 },
 { 0x73FE, 0xD604 },
 { 0x7401, 0xC120 },
 { 0x7403, 0xAD6C },
 { 0x7405, 0xB791 },
 { 0x7406, 0xB9AC },
 { 0x7407, 0xC218 },
 { 0x7409, 0xB958 },
 { 0x7413, 0xC644 },
 { 0x741B, 0xCE68 },
 { 0x7420, 0xC804 },
 { 0x7421, 0xC219 },
 { 0x7422, 0xD0C1 },
 { 0x7425, 0xD638 },
 { 0x7426, 0xAE30 },
 { 0x7428, 0xACE4 },
 { 0x742A, 0xAE30 },
 { 0x742B, 0xBD09 },
 { 0x742C, 0xC644 },
 { 0x742E, 0xC885 },
 { 0x742F, 0xAD00 },
 { 0x7430, 0xC5FC },
 { 0x7433, 0xB9BC },
 { 0x7434, 0xAE08 },
 { 0x7435, 0xBE44 },
 { 0x7436, 0xD30C },
 { 0x7438, 0xD0C1 },
 { 0x743A, 0xBC95 },
 { 0x743F, 0xD63C },
 { 0x7440, 0xC6B0 },
 { 0x7441, 0xBAA8 },
 { 0x7443, 0xCD98 },
 { 0x7444, 0xC120 },
 { 0x744B, 0xC704 },
 { 0x7455, 0xD558 },
 { 0x7457, 0xC6D0 },
 { 0x7459, 0xB178 },
 { 0x745A, 0xD638 },
 { 0x745B, 0xC601 },
 { 0x745C, 0xC720 },
 { 0x745E, 0xC11C },
 { 0x745F, 0xC2AC },
 { 0x7460, 0xB958 },
 { 0x7462, 0xC6A9 },
 { 0x7464, 0xC694 },
 { 0x7465, 0xC628 },
 { 0x7468, 0xC9C4 },
 { 0x7469, 0xC601 },
 { 0x746A, 0xB9C8 },
 { 0x746F, 0xB791 },
 { 0x747E, 0xADFC },
 { 0x7482, 0xAE30 },
 { 0x7483, 0xB9AC },
 { 0x7487, 0xC120 },
 { 0x7489, 0xB828 },
 { 0x748B, 0xC7A5 },
 { 0x7498, 0xB9B0 },
 { 0x749C, 0xD669 },
 { 0x749E, 0xBC15 },
 { 0x749F, 0xACBD },
 { 0x74A1, 0xC9C4 },
 { 0x74A3, 0xAE30 },
 { 0x74A5, 0xACBD },
 { 0x74A7, 0xBCBD },
 { 0x74A8, 0xCC2C },
 { 0x74AA, 0xC870 },
 { 0x74B0, 0xD658 },
 { 0x74B2, 0xC218 },
 { 0x74B5, 0xC5EC },
 { 0x74B9, 0xC219 },
 { 0x74BD, 0xC0C8 },
 { 0x74BF, 0xC120 },
 { 0x74C6, 0xC9C8 },
 { 0x74CA, 0xACBD },
 { 0x74CF, 0xB871 },
 { 0x74D4, 0xC601 },
 { 0x74D8, 0xAD00 },
 { 0x74DA, 0xCC2C },
 { 0x74DC, 0xACFC },
 { 0x74E0, 0xD638 },
 { 0x74E2, 0xD45C },
 { 0x74E3, 0xD310 },
 { 0x74E6, 0xC640 },
 { 0x74EE, 0xC639 },
 { 0x74F7, 0xC790 },
 { 0x7501, 0xBCD1 },
 { 0x7504, 0xACAC },
 { 0x7511, 0xC99D },
 { 0x7515, 0xC639 },
 { 0x7518, 0xAC10 },
 { 0x751A, 0xC2EC },
 { 0x751B, 0xCCA8 },
 { 0x751F, 0xC0DD },
 { 0x7523, 0xC0B0 },
 { 0x7525, 0xC0DD },
 { 0x7526, 0xC18C },
 { 0x7528, 0xC6A9 },
 { 0x752B, 0xBCF4 },
 { 0x752C, 0xC6A9 },
 { 0x7530, 0xC804 },
 { 0x7531, 0xC720 },
 { 0x7532, 0xAC11 },
 { 0x7533, 0xC2E0 },
 { 0x7537, 0xB0A8 },
 { 0x7538, 0xC804 },
 { 0x753A, 0xC815 },
 { 0x7547, 0xADE0 },
 { 0x754C, 0xACC4 },
 { 0x754F, 0xC678 },
 { 0x7551, 0xC804 },
 { 0x7553, 0xB2F5 },
 { 0x7554, 0xBC18 },
 { 0x7559, 0xB958 },
 { 0x755B, 0xC9C4 },
 { 0x755C, 0xCD95 },
 { 0x755D, 0xBB34 },
 { 0x7562, 0xD544 },
 { 0x7565, 0xB7B5 },
 { 0x7566, 0xD734 },
 { 0x756A, 0xBC88 },
 { 0x756F, 0xC900 },
 { 0x7570, 0xB9AC },
 { 0x7575, 0xD654 },
 { 0x7576, 0xB2F9 },
 { 0x7578, 0xAE30 },
 { 0x757A, 0xAC15 },
 { 0x757F, 0xAE30 },
 { 0x7586, 0xAC15 },
 { 0x7587, 0xC8FC },
 { 0x758A, 0xCCA9 },
 { 0x758B, 0xD544 },
 { 0x758E, 0xC18C },
 { 0x758F, 0xC18C },
 { 0x7591, 0xC758 },
 { 0x759D, 0xC0B0 },
 { 0x75A5, 0xAC1C },
 { 0x75AB, 0xC5ED },
 { 0x75B1, 0xD3EC },
 { 0x75B2, 0xD53C },
 { 0x75B3, 0xAC10 },
 { 0x75B5, 0xC790 },
 { 0x75B8, 0xB2EC },
 { 0x75B9, 0xC9C4 },
 { 0x75BC, 0xB3D9 },
 { 0x75BD, 0xC800 },
 { 0x75BE, 0xC9C8 },
 { 0x75C2, 0xAC00 },
 { 0x75C5, 0xBCD1 },
 { 0x75C7, 0xC99D },
 { 0x75CD, 0xC774 },
 { 0x75D2, 0xC591 },
 { 0x75D4, 0xCE58 },
 { 0x75D5, 0xD754 },
 { 0x75D8, 0xB450 },
 { 0x75D9, 0xACBD },
 { 0x75DB, 0xD1B5 },
 { 0x75E2, 0xB9AC },
 { 0x75F0, 0xB2F4 },
 { 0x75F2, 0xB9C8 },
 { 0x75F4, 0xCE58 },
 { 0x75FA, 0xBE44 },
 { 0x75FC, 0xACE0 },
 { 0x7600, 0xC5B4 },
 { 0x760D, 0xC591 },
 { 0x7619, 0xC18C },
 { 0x761F, 0xC628 },
 { 0x7620, 0xCC99 },
 { 0x7621, 0xCC3D },
 { 0x7622, 0xBC18 },
 { 0x7624, 0xB958 },
 { 0x7626, 0xC218 },
 { 0x763B, 0xB8E8 },
 { 0x7642, 0xB8CC },
 { 0x764C, 0xC554 },
 { 0x764E, 0xAC04 },
 { 0x7652, 0xC720 },
 { 0x7656, 0xBCBD },
 { 0x7661, 0xCE58 },
 { 0x7664, 0xC808 },
 { 0x7669, 0xB098 },
 { 0x766C, 0xC120 },
 { 0x7670, 0xC639 },
 { 0x7672, 0xC804 },
 { 0x7678, 0xACC4 },
 { 0x767B, 0xB4F1 },
 { 0x767C, 0xBC1C },
 { 0x767D, 0xBC31 },
 { 0x767E, 0xBC31 },
 { 0x7684, 0xC801 },
 { 0x7686, 0xAC1C },
 { 0x7687, 0xD669 },
 { 0x768E, 0xAD50 },
 { 0x7690, 0xACE0 },
 { 0x7693, 0xD638 },
 { 0x76AE, 0xD53C },
 { 0x76BA, 0xCD94 },
 { 0x76BF, 0xBA85 },
 { 0x76C2, 0xC6B0 },
 { 0x76C3, 0xBC30 },
 { 0x76C6, 0xBD84 },
 { 0x76C8, 0xC601 },
 { 0x76CA, 0xC775 },
 { 0x76D2, 0xD569 },
 { 0x76D6, 0xAC1C },
 { 0x76DB, 0xC131 },
 { 0x76DC, 0xB3C4 },
 { 0x76DE, 0xC794 },
 { 0x76DF, 0xB9F9 },
 { 0x76E1, 0xC9C4 },
 { 0x76E3, 0xAC10 },
 { 0x76E4, 0xBC18 },
 { 0x76E7, 0xB178 },
 { 0x76EE, 0xBAA9 },
 { 0x76F2, 0xB9F9 },
 { 0x76F4, 0xC9C1 },
 { 0x76F8, 0xC0C1 },
 { 0x76FC, 0xBC18 },
 { 0x76FE, 0xC21C },
 { 0x7701, 0xC0DD },
 { 0x7704, 0xBA74 },
 { 0x7708, 0xD0D0 },
 { 0x7709, 0xBBF8 },
 { 0x770B, 0xAC04 },
 { 0x771E, 0xC9C4 },
 { 0x7720, 0xBA74 },
 { 0x7729, 0xD604 },
 { 0x7737, 0xAD8C },
 { 0x7738, 0xBAA8 },
 { 0x773A, 0xC870 },
 { 0x773C, 0xC548 },
 { 0x7740, 0xCC29 },
 { 0x774D, 0xD604 },
 { 0x775B, 0xC815 },
 { 0x7761, 0xC218 },
 { 0x7763, 0xB3C5 },
 { 0x7766, 0xBAA9 },
 { 0x776B, 0xCCA9 },
 { 0x7779, 0xB3C4 },
 { 0x777E, 0xACE0 },
 { 0x777F, 0xC608 },
 { 0x778B, 0xC9C4 },
 { 0x7791, 0xBA85 },
 { 0x779E, 0xB9CC },
 { 0x77A5, 0xBCC4 },
 { 0x77AC, 0xC21C },
 { 0x77AD, 0xB8CC },
 { 0x77B0, 0xAC10 },
 { 0x77B3, 0xB3D9 },
 { 0x77BB, 0xCCA8 },
 { 0x77BC, 0xAC80 },
 { 0x77BF, 0xAD6C },
 { 0x77D7, 0xCD09 },
 { 0x77DB, 0xBAA8 },
 { 0x77DC, 0xAE0D },
 { 0x77E2, 0xC2DC },
 { 0x77E3, 0xC758 },
 { 0x77E5, 0xC9C0 },
 { 0x77E9, 0xAD6C },
 { 0x77ED, 0xB2E8 },
 { 0x77EE, 0xC65C },
 { 0x77EF, 0xAD50 },
 { 0x77F3, 0xC11D },
 { 0x7802, 0xC0AC },
 { 0x7812, 0xBE44 },
 { 0x7825, 0xC9C0 },
 { 0x7826, 0xCC44 },
 { 0x7827, 0xCE68 },
 { 0x782C, 0xB9BD },
 { 0x7832, 0xD3EC },
 { 0x7834, 0xD30C },
 { 0x7845, 0xADDC },
 { 0x784F, 0xC5F0 },
 { 0x785D, 0xCD08 },
 { 0x786B, 0xB958 },
 { 0x786C, 0xACBD },
 { 0x786F, 0xC5F0 },
 { 0x787C, 0xBD95 },
 { 0x7881, 0xAE30 },
 { 0x7887, 0xC815 },
 { 0x788C, 0xB179 },
 { 0x788D, 0xC560 },
 { 0x788E, 0xC1C4 },
 { 0x7891, 0xBE44 },
 { 0x7897, 0xC644 },
 { 0x78A3, 0xAC08 },
 { 0x78A7, 0xBCBD },
 { 0x78A9, 0xC11D },
 { 0x78BA, 0xD655 },
 { 0x78BB, 0xD655 },
 { 0x78BC, 0xB9C8 },
 { 0x78C1, 0xC790 },
 { 0x78C5, 0xBC29 },
 { 0x78CA, 0xB1CC },
 { 0x78CB, 0xCC28 },
 { 0x78CE, 0xACC4 },
 { 0x78D0, 0xBC18 },
 { 0x78E8, 0xB9C8 },
 { 0x78EC, 0xACBD },
 { 0x78EF, 0xAE30 },
 { 0x78F5, 0xAC04 },
 { 0x78FB, 0xBC18 },
 { 0x7901, 0xCD08 },
 { 0x790E, 0xCD08 },
 { 0x7916, 0xC5EC },
 { 0x792A, 0xB824 },
 { 0x792B, 0xB825 },
 { 0x792C, 0xBC18 },
 { 0x793A, 0xC2DC },
 { 0x793E, 0xC0AC },
 { 0x7940, 0xC0AC },
 { 0x7941, 0xAE30 },
 { 0x7947, 0xAE30 },
 { 0x7948, 0xAE30 },
 { 0x7949, 0xC9C0 },
 { 0x7950, 0xC6B0 },
 { 0x7956, 0xC870 },
 { 0x7957, 0xC9C0 },
 { 0x795A, 0xC870 },
 { 0x795B, 0xAC70 },
 { 0x795C, 0xD638 },
 { 0x795D, 0xCD95 },
 { 0x795E, 0xC2E0 },
 { 0x7960, 0xC0AC },
 { 0x7965, 0xC0C1 },
 { 0x7968, 0xD45C },
 { 0x796D, 0xC81C },
 { 0x797A, 0xAE30 },
 { 0x797F, 0xB179 },
 { 0x7981, 0xAE08 },
 { 0x798D, 0xD654 },
 { 0x798E, 0xC815 },
 { 0x798F, 0xBCF5 },
 { 0x7991, 0xC6B0 },
 { 0x79A6, 0xC5B4 },
 { 0x79A7, 0xD76C },
 { 0x79AA, 0xC120 },
 { 0x79AE, 0xB840 },
 { 0x79B1, 0xB3C4 },
 { 0x79B3, 0xC591 },
 { 0x79B9, 0xC6B0 },
 { 0x79BD, 0xAE08 },
 { 0x79BE, 0xD654 },
 { 0x79BF, 0xB3C5 },
 { 0x79C0, 0xC218 },
 { 0x79C1, 0xC0AC },
 { 0x79C9, 0xBCD1 },
 { 0x79CA, 0xB144 },
 { 0x79CB, 0xCD94 },
 { 0x79D1, 0xACFC },
 { 0x79D2, 0xCD08 },
 { 0x79D5, 0xBE44 },
 { 0x79D8, 0xBE44 },
 { 0x79DF, 0xC870 },
 { 0x79E4, 0xCE6D },
 { 0x79E6, 0xC9C4 },
 { 0x79E7, 0xC559 },
 { 0x79E9, 0xC9C8 },
 { 0x79FB, 0xC774 },
 { 0x7A00, 0xD76C },
 { 0x7A05, 0xC138 },
 { 0x7A08, 0xAC04 },
 { 0x7A0B, 0xC815 },
 { 0x7A0D, 0xCD08 },
 { 0x7A14, 0xC784 },
 { 0x7A17, 0xD328 },
 { 0x7A19, 0xC9C1 },
 { 0x7A1A, 0xCE58 },
 { 0x7A1C, 0xB2A5 },
 { 0x7A1F, 0xD488 },
 { 0x7A20, 0xC870 },
 { 0x7A2E, 0xC885 },
 { 0x7A31, 0xCE6D },
 { 0x7A36, 0xC6B1 },
 { 0x7A37, 0xC9C1 },
 { 0x7A3B, 0xB3C4 },
 { 0x7A3C, 0xAC00 },
 { 0x7A3D, 0xACC4 },
 { 0x7A3F, 0xACE0 },
 { 0x7A40, 0xACE1 },
 { 0x7A46, 0xBAA9 },
 { 0x7A49, 0xCE58 },
 { 0x7A4D, 0xC801 },
 { 0x7A4E, 0xC601 },
 { 0x7A57, 0xC218 },
 { 0x7A61, 0xC0C9 },
 { 0x7A62, 0xC608 },
 { 0x7A69, 0xC628 },
 { 0x7A6B, 0xD655 },
 { 0x7A70, 0xC591 },
 { 0x7A74, 0xD608 },
 { 0x7A76, 0xAD6C },
 { 0x7A79, 0xAD81 },
 { 0x7A7A, 0xACF5 },
 { 0x7A7D, 0xC815 },
 { 0x7A7F, 0xCC9C },
 { 0x7A81, 0xB3CC },
 { 0x7A84, 0xCC29 },
 { 0x7A88, 0xC694 },
 { 0x7A92, 0xC9C8 },
 { 0x7A93, 0xCC3D },
 { 0x7A95, 0xC870 },
 { 0x7A98, 0xAD70 },
 { 0x7A9F, 0xAD74 },
 { 0x7AA9, 0xC640 },
 { 0x7AAA, 0xC640 },
 { 0x7AAE, 0xAD81 },
 { 0x7AAF, 0xC694 },
 { 0x7ABA, 0xADDC },
 { 0x7AC4, 0xCC2C },
 { 0x7AC5, 0xADDC },
 { 0x7AC7, 0xB450 },
 { 0x7ACA, 0xC808 },
 { 0x7ACB, 0xB9BD },
 { 0x7AD7, 0xBB18 },
 { 0x7AD9, 0xCC38 },
 { 0x7ADD, 0xBCD1 },
 { 0x7ADF, 0xACBD },
 { 0x7AE0, 0xC7A5 },
 { 0x7AE3, 0xC900 },
 { 0x7AE5, 0xB3D9 },
 { 0x7AEA, 0xC218 },
 { 0x7AED, 0xAC08 },
 { 0x7AEF, 0xB2E8 },
 { 0x7AF6, 0xACBD },
 { 0x7AF9, 0xC8FD },
 { 0x7AFA, 0xCD95 },
 { 0x7AFF, 0xAC04 },
 { 0x7B0F, 0xD640 },
 { 0x7B11, 0xC18C },
 { 0x7B19, 0xC0DD },
 { 0x7B1B, 0xC801 },
 { 0x7B1E, 0xD0DC },
 { 0x7B20, 0xB9BD },
 { 0x7B26, 0xBD80 },
 { 0x7B2C, 0xC81C },
 { 0x7B2D, 0xB839 },
 { 0x7B39, 0xC138 },
 { 0x7B46, 0xD544 },
 { 0x7B49, 0xB4F1 },
 { 0x7B4B, 0xADFC },
 { 0x7B4C, 0xC804 },
 { 0x7B4D, 0xC21C },
 { 0x7B4F, 0xBC8C },
 { 0x7B50, 0xAD11 },
 { 0x7B51, 0xCD95 },
 { 0x7B52, 0xD1B5 },
 { 0x7B54, 0xB2F5 },
 { 0x7B56, 0xCC45 },
 { 0x7B60, 0xADE0 },
 { 0x7B6C, 0xC131 },
 { 0x7B6E, 0xC11C },
 { 0x7B75, 0xC5F0 },
 { 0x7B7D, 0xC624 },
 { 0x7B87, 0xAC1C },
 { 0x7B8B, 0xC804 },
 { 0x7B8F, 0xC7C1 },
 { 0x7B94, 0xBC15 },
 { 0x7B95, 0xAE30 },
 { 0x7B97, 0xC0B0 },
 { 0x7B9A, 0xCC28 },
 { 0x7B9D, 0xACB8 },
 { 0x7BA1, 0xAD00 },
 { 0x7BAD, 0xC804 },
 { 0x7BB1, 0xC0C1 },
 { 0x7BB4, 0xC7A0 },
 { 0x7BB8, 0xC800 },
 { 0x7BC0, 0xC808 },
 { 0x7BC1, 0xD669 },
 { 0x7BC4, 0xBC94 },
 { 0x7BC6, 0xC804 },
 { 0x7BC7, 0xD3B8 },
 { 0x7BC9, 0xCD95 },
 { 0x7BD2, 0xC2DD },
 { 0x7BE0, 0xC18C },
 { 0x7BE4, 0xB3C5 },
 { 0x7BE9, 0xC0AC },
 { 0x7C07, 0xC871 },
 { 0x7C12, 0xCC2C },
 { 0x7C1E, 0xB2E8 },
 { 0x7C21, 0xAC04 },
 { 0x7C27, 0xD669 },
 { 0x7C2A, 0xC7A0 },
 { 0x7C2B, 0xC18C },
 { 0x7C3D, 0xCCA8 },
 { 0x7C3E, 0xB834 },
 { 0x7C3F, 0xBD80 },
 { 0x7C43, 0xB78C },
 { 0x7C4C, 0xC8FC },
 { 0x7C4D, 0xC801 },
 { 0x7C60, 0xB18D },
 { 0x7C64, 0xCCA8 },
 { 0x7C6C, 0xB9AC },
 { 0x7C73, 0xBBF8 },
 { 0x7C83, 0xBE44 },
 { 0x7C89, 0xBD84 },
 { 0x7C92, 0xB9BD },
 { 0x7C95, 0xBC15 },
 { 0x7C97, 0xC870 },
 { 0x7C98, 0xC810 },
 { 0x7C9F, 0xC18D },
 { 0x7CA5, 0xC8FD },
 { 0x7CA7, 0xC7A5 },
 { 0x7CAE, 0xB7C9 },
 { 0x7CB1, 0xB7C9 },
 { 0x7CB2, 0xCC2C },
 { 0x7CB3, 0xAC31 },
 { 0x7CB9, 0xC218 },
 { 0x7CBE, 0xC815 },
 { 0x7CCA, 0xD638 },
 { 0x7CD6, 0xB2F9 },
 { 0x7CDE, 0xBD84 },
 { 0x7CDF, 0xC870 },
 { 0x7CE0, 0xAC15 },
 { 0x7CE7, 0xB7C9 },
 { 0x7CFB, 0xACC4 },
 { 0x7CFE, 0xADDC },
 { 0x7D00, 0xAE30 },
 { 0x7D02, 0xC8FC },
 { 0x7D04, 0xC57D },
 { 0x7D05, 0xD64D },
 { 0x7D06, 0xC6B0 },
 { 0x7D07, 0xD758 },
 { 0x7D08, 0xD658 },
 { 0x7D0A, 0xBB38 },
 { 0x7D0B, 0xBB38 },
 { 0x7D0D, 0xB0A9 },
 { 0x7D10, 0xB274 },
 { 0x7D14, 0xC21C },
 { 0x7D17, 0xC0AC },
 { 0x7D18, 0xAD49 },
 { 0x7D19, 0xC9C0 },
 { 0x7D1A, 0xAE09 },
 { 0x7D1B, 0xBD84 },
 { 0x7D20, 0xC18C },
 { 0x7D21, 0xBC29 },
 { 0x7D22, 0xC0AD },
 { 0x7D2B, 0xC790 },
 { 0x7D2C, 0xC8FC },
 { 0x7D2E, 0xCC30 },
 { 0x7D2F, 0xB204 },
 { 0x7D30, 0xC138 },
 { 0x7D33, 0xC2E0 },
 { 0x7D35, 0xC800 },
 { 0x7D39, 0xC18C },
 { 0x7D3A, 0xAC10 },
 { 0x7D42, 0xC885 },
 { 0x7D43, 0xD604 },
 { 0x7D44, 0xC870 },
 { 0x7D45, 0xACBD },
 { 0x7D46, 0xBC18 },
 { 0x7D50, 0xACB0 },
 { 0x7D5E, 0xAD50 },
 { 0x7D61, 0xB77D },
 { 0x7D62, 0xD604 },
 { 0x7D66, 0xAE09 },
 { 0x7D68, 0xC735 },
 { 0x7D6A, 0xC778 },
 { 0x7D6E, 0xC11C },
 { 0x7D71, 0xD1B5 },
 { 0x7D72, 0xC0AC },
 { 0x7D73, 0xAC15 },
 { 0x7D76, 0xC808 },
 { 0x7D79, 0xACAC },
 { 0x7D7F, 0xAD6C },
 { 0x7D8E, 0xC815 },
 { 0x7D8F, 0xC218 },
 { 0x7D93, 0xACBD },
 { 0x7D9C, 0xC885 },
 { 0x7DA0, 0xB179 },
 { 0x7DA2, 0xC8FC },
 { 0x7DAC, 0xC218 },
 { 0x7DAD, 0xC720 },
 { 0x7DB1, 0xAC15 },
 { 0x7DB2, 0xB9DD },
 { 0x7DB4, 0xCCA0 },
 { 0x7DB5, 0xCC44 },
 { 0x7DB8, 0xB95C },
 { 0x7DBA, 0xAE30 },
 { 0x7DBB, 0xD0C4 },
 { 0x7DBD, 0xC791 },
 { 0x7DBE, 0xB2A5 },
 { 0x7DBF, 0xBA74 },
 { 0x7DC7, 0xCE58 },
 { 0x7DCA, 0xAE34 },
 { 0x7DCB, 0xBE44 },
 { 0x7DD6, 0xC11C },
 { 0x7DD8, 0xD568 },
 { 0x7DDA, 0xC120 },
 { 0x7DDD, 0xC9D1 },
 { 0x7DDE, 0xB2E8 },
 { 0x7DE0, 0xCCB4 },
 { 0x7DE1, 0xBBFC },
 { 0x7DE3, 0xC5F0 },
 { 0x7DE8, 0xD3B8 },
 { 0x7DE9, 0xC644 },
 { 0x7DEC, 0xBA74 },
 { 0x7DEF, 0xC704 },
 { 0x7DF4, 0xB828 },
 { 0x7DFB, 0xCE58 },
 { 0x7E09, 0xC9C4 },
 { 0x7E0A, 0xC561 },
 { 0x7E15, 0xC628 },
 { 0x7E1B, 0xBC15 },
 { 0x7E1D, 0xC9C4 },
 { 0x7E1E, 0xD638 },
 { 0x7E1F, 0xC695 },
 { 0x7E21, 0xC7AC },
 { 0x7E23, 0xD604 },
 { 0x7E2B, 0xBD09 },
 { 0x7E2E, 0xCD95 },
 { 0x7E2F, 0xC5F0 },
 { 0x7E31, 0xC885 },
 { 0x7E37, 0xB204 },
 { 0x7E3D, 0xCD1D },
 { 0x7E3E, 0xC801 },
 { 0x7E41, 0xBC88 },
 { 0x7E43, 0xBD95 },
 { 0x7E46, 0xBB34 },
 { 0x7E47, 0xC694 },
 { 0x7E52, 0xC99D },
 { 0x7E54, 0xC9C1 },
 { 0x7E55, 0xC120 },
 { 0x7E5E, 0xC694 },
 { 0x7E61, 0xC218 },
 { 0x7E69, 0xC2B9 },
 { 0x7E6A, 0xD68C },
 { 0x7E6B, 0xACC4 },
 { 0x7E6D, 0xACAC },
 { 0x7E70, 0xC870 },
 { 0x7E79, 0xC5ED },
 { 0x7E7C, 0xACC4 },
 { 0x7E82, 0xCC2C },
 { 0x7E8C, 0xC18D },
 { 0x7E8F, 0xC804 },
 { 0x7E93, 0xC601 },
 { 0x7E96, 0xC12C },
 { 0x7E98, 0xCC2C },
 { 0x7E9B, 0xB3C5 },
 { 0x7E9C, 0xB78C },
 { 0x7F36, 0xBD80 },
 { 0x7F38, 0xD56D },
 { 0x7F3A, 0xACB0 },
 { 0x7F4C, 0xC575 },
 { 0x7F50, 0xAD00 },
 { 0x7F54, 0xB9DD },
 { 0x7F55, 0xD55C },
 { 0x7F6A, 0xC8C4 },
 { 0x7F6B, 0xAD18 },
 { 0x7F6E, 0xCE58 },
 { 0x7F70, 0xBC8C },
 { 0x7F72, 0xC11C },
 { 0x7F75, 0xB9E4 },
 { 0x7F77, 0xD30C },
 { 0x7F79, 0xB9AC },
 { 0x7F85, 0xB098 },
 { 0x7F88, 0xAE30 },
 { 0x7F8A, 0xC591 },
 { 0x7F8C, 0xAC15 },
 { 0x7F8E, 0xBBF8 },
 { 0x7F94, 0xACE0 },
 { 0x7F9A, 0xB839 },
 { 0x7F9E, 0xC218 },
 { 0x7FA4, 0xAD70 },
 { 0x7FA8, 0xC120 },
 { 0x7FA9, 0xC758 },
 { 0x7FB2, 0xD76C },
 { 0x7FB8, 0xB9AC },
 { 0x7FB9, 0xAC31 },
 { 0x7FBD, 0xC6B0 },
 { 0x7FC1, 0xC639 },
 { 0x7FC5, 0xC2DC },
 { 0x7FCA, 0xC775 },
 { 0x7FCC, 0xC775 },
 { 0x7FCE, 0xB839 },
 { 0x7FD2, 0xC2B5 },
 { 0x7FD4, 0xC0C1 },
 { 0x7FD5, 0xD761 },
 { 0x7FDF, 0xC801 },
 { 0x7FE0, 0xCDE8 },
 { 0x7FE1, 0xBE44 },
 { 0x7FE9, 0xD3B8 },
 { 0x7FEB, 0xC644 },
 { 0x7FF0, 0xD55C },
 { 0x7FF9, 0xAD50 },
 { 0x7FFC, 0xC775 },
 { 0x8000, 0xC694 },
 { 0x8001, 0xB178 },
 { 0x8003, 0xACE0 },
 { 0x8005, 0xC790 },
 { 0x8006, 0xAE30 },
 { 0x8009, 0xAD6C },
 { 0x800C, 0xC774 },
 { 0x8010, 0xB0B4 },
 { 0x8015, 0xACBD },
 { 0x8017, 0xBAA8 },
 { 0x8018, 0xC6B4 },
 { 0x802D, 0xAE30 },
 { 0x8033, 0xC774 },
 { 0x8036, 0xC57C },
 { 0x803D, 0xD0D0 },
 { 0x803F, 0xACBD },
 { 0x8043, 0xB2F4 },
 { 0x8046, 0xB839 },
 { 0x804A, 0xB8CC },
 { 0x8056, 0xC131 },
 { 0x8058, 0xBE59 },
 { 0x805A, 0xCDE8 },
 { 0x805E, 0xBB38 },
 { 0x806F, 0xB828 },
 { 0x8070, 0xCD1D },
 { 0x8072, 0xC131 },
 { 0x8073, 0xC6A9 },
 { 0x8077, 0xC9C1 },
 { 0x807D, 0xCCAD },
 { 0x807E, 0xB18D },
 { 0x807F, 0xC728 },
 { 0x8084, 0xC774 },
 { 0x8085, 0xC219 },
 { 0x8086, 0xC0AC },
 { 0x8087, 0xC870 },
 { 0x8089, 0xC721 },
 { 0x808B, 0xB291 },
 { 0x808C, 0xAE30 },
 { 0x8096, 0xCD08 },
 { 0x809B, 0xD56D },
 { 0x809D, 0xAC04 },
 { 0x80A1, 0xACE0 },
 { 0x80A2, 0xC9C0 },
 { 0x80A5, 0xBE44 },
 { 0x80A9, 0xACAC },
 { 0x80AA, 0xBC29 },
 { 0x80AF, 0xAE0D },
 { 0x80B1, 0xAD49 },
 { 0x80B2, 0xC721 },
 { 0x80B4, 0xD6A8 },
 { 0x80BA, 0xD3D0 },
 { 0x80C3, 0xC704 },
 { 0x80C4, 0xC8FC },
 { 0x80CC, 0xBC30 },
 { 0x80CE, 0xD0DC },
 { 0x80DA, 0xBC30 },
 { 0x80DB, 0xAC11 },
 { 0x80DE, 0xD3EC },
 { 0x80E1, 0xD638 },
 { 0x80E4, 0xC724 },
 { 0x80E5, 0xC11C },
 { 0x80F1, 0xAD11 },
 { 0x80F4, 0xB3D9 },
 { 0x80F8, 0xD749 },
 { 0x80FD, 0xB2A5 },
 { 0x8102, 0xC9C0 },
 { 0x8105, 0xD611 },
 { 0x8106, 0xCDE8 },
 { 0x8107, 0xD611 },
 { 0x8108, 0xB9E5 },
 { 0x810A, 0xCC99 },
 { 0x8118, 0xC644 },
 { 0x811A, 0xAC01 },
 { 0x811B, 0xACBD },
 { 0x8123, 0xC21C },
 { 0x8129, 0xC218 },
 { 0x812B, 0xD0C8 },
 { 0x812F, 0xD3EC },
 { 0x8139, 0xCC3D },
 { 0x813E, 0xBE44 },
 { 0x814B, 0xC561 },
 { 0x814E, 0xC2E0 },
 { 0x8150, 0xBD80 },
 { 0x8151, 0xBD80 },
 { 0x8154, 0xAC15 },
 { 0x8155, 0xC644 },
 { 0x8165, 0xC131 },
 { 0x8166, 0xB1CC },
 { 0x816B, 0xC885 },
 { 0x8170, 0xC694 },
 { 0x8171, 0xAC74 },
 { 0x8178, 0xC7A5 },
 { 0x8179, 0xBCF5 },
 { 0x817A, 0xC120 },
 { 0x817F, 0xD1F4 },
 { 0x8180, 0xBC29 },
 { 0x8188, 0xACA9 },
 { 0x818A, 0xBC15 },
 { 0x818F, 0xACE0 },
 { 0x819A, 0xBD80 },
 { 0x819C, 0xB9C9 },
 { 0x819D, 0xC2AC },
 { 0x81A0, 0xAD50 },
 { 0x81A3, 0xC9C8 },
 { 0x81A8, 0xD33D },
 { 0x81B3, 0xC120 },
 { 0x81B5, 0xCDCC },
 { 0x81BA, 0xC751 },
 { 0x81BD, 0xB2F4 },
 { 0x81BE, 0xD68C },
 { 0x81BF, 0xB18D },
 { 0x81C0, 0xB454 },
 { 0x81C2, 0xBE44 },
 { 0x81C6, 0xC5B5 },
 { 0x81CD, 0xC81C },
 { 0x81D8, 0xB0A9 },
 { 0x81DF, 0xC7A5 },
 { 0x81E3, 0xC2E0 },
 { 0x81E5, 0xC640 },
 { 0x81E7, 0xC7A5 },
 { 0x81E8, 0xB9BC },
 { 0x81EA, 0xC790 },
 { 0x81ED, 0xCDE8 },
 { 0x81F3, 0xC9C0 },
 { 0x81F4, 0xCE58 },
 { 0x81FA, 0xB300 },
 { 0x81FB, 0xC9C4 },
 { 0x81FC, 0xAD6C },
 { 0x81FE, 0xC720 },
 { 0x8205, 0xAD6C },
 { 0x8207, 0xC5EC },
 { 0x8208, 0xD765 },
 { 0x820A, 0xAD6C },
 { 0x820C, 0xC124 },
 { 0x820D, 0xC0AC },
 { 0x8212, 0xC11C },
 { 0x821B, 0xCC9C },
 { 0x821C, 0xC21C },
 { 0x821E, 0xBB34 },
 { 0x821F, 0xC8FC },
 { 0x8221, 0xAC15 },
 { 0x822A, 0xD56D },
 { 0x822B, 0xBC29 },
 { 0x822C, 0xBC18 },
 { 0x8235, 0xD0C0 },
 { 0x8236, 0xBC15 },
 { 0x8237, 0xD604 },
 { 0x8239, 0xC120 },
 { 0x8240, 0xBD80 },
 { 0x8245, 0xC5EC },
 { 0x8247, 0xC815 },
 { 0x8259, 0xCC3D },
 { 0x8264, 0xC758 },
 { 0x8266, 0xD568 },
 { 0x826E, 0xAC04 },
 { 0x826F, 0xB7C9 },
 { 0x8271, 0xAC04 },
 { 0x8272, 0xC0C9 },
 { 0x8276, 0xC5FC },
 { 0x8278, 0xCD08 },
 { 0x827E, 0xC560 },
 { 0x828B, 0xC6B0 },
 { 0x828D, 0xC791 },
 { 0x828E, 0xAD81 },
 { 0x8292, 0xB9DD },
 { 0x8299, 0xBD80 },
 { 0x829A, 0xB454 },
 { 0x829D, 0xC9C0 },
 { 0x829F, 0xC0BC },
 { 0x82A5, 0xAC1C },
 { 0x82A6, 0xD638 },
 { 0x82A9, 0xAE08 },
 { 0x82AC, 0xBD84 },
 { 0x82AD, 0xD30C },
 { 0x82AE, 0xC608 },
 { 0x82AF, 0xC2EC },
 { 0x82B1, 0xD654 },
 { 0x82B3, 0xBC29 },
 { 0x82B7, 0xC9C0 },
 { 0x82B8, 0xC6B4 },
 { 0x82B9, 0xADFC },
 { 0x82BB, 0xCD94 },
 { 0x82BC, 0xBAA8 },
 { 0x82BD, 0xC544 },
 { 0x82BF, 0xC789 },
 { 0x82D1, 0xC6D0 },
 { 0x82D2, 0xC5FC },
 { 0x82D4, 0xD0DC },
 { 0x82D5, 0xCD08 },
 { 0x82D7, 0xBB18 },
 { 0x82DB, 0xAC00 },
 { 0x82DE, 0xD3EC },
 { 0x82DF, 0xAD6C },
 { 0x82E1, 0xC774 },
 { 0x82E5, 0xC57C },
 { 0x82E6, 0xACE0 },
 { 0x82E7, 0xC800 },
 { 0x82F1, 0xC601 },
 { 0x82FD, 0xACE0 },
 { 0x82FE, 0xD544 },
 { 0x8301, 0xC904 },
 { 0x8302, 0xBB34 },
 { 0x8303, 0xBC94 },
 { 0x8304, 0xAC00 },
 { 0x8305, 0xBAA8 },
 { 0x8309, 0xB9D0 },
 { 0x8317, 0xBA85 },
 { 0x8328, 0xC790 },
 { 0x832B, 0xB9DD },
 { 0x832F, 0xBCF5 },
 { 0x8331, 0xC218 },
 { 0x8334, 0xD68C },
 { 0x8335, 0xC778 },
 { 0x8336, 0xB2E4 },
 { 0x8338, 0xC6A9 },
 { 0x8339, 0xC5EC },
 { 0x8340, 0xC21C },
 { 0x8347, 0xD589 },
 { 0x8349, 0xCD08 },
 { 0x834A, 0xD615 },
 { 0x834F, 0xC784 },
 { 0x8351, 0xC774 },
 { 0x8352, 0xD669 },
 { 0x8373, 0xB450 },
 { 0x8377, 0xD558 },
 { 0x837B, 0xC801 },
 { 0x8389, 0xB9AC },
 { 0x838A, 0xC7A5 },
 { 0x838E, 0xC0AC },
 { 0x8396, 0xACBD },
 { 0x8398, 0xC2E0 },
 { 0x839E, 0xC644 },
 { 0x83A2, 0xD611 },
 { 0x83A9, 0xBD80 },
 { 0x83AA, 0xC544 },
 { 0x83AB, 0xB9C9 },
 { 0x83BD, 0xB9DD },
 { 0x83C1, 0xCCAD },
 { 0x83C5, 0xAD00 },
 { 0x83C9, 0xB179 },
 { 0x83CA, 0xAD6D },
 { 0x83CC, 0xADE0 },
 { 0x83D3, 0xACFC },
 { 0x83D6, 0xCC3D },
 { 0x83DC, 0xCC44 },
 { 0x83E9, 0xBCF4 },
 { 0x83EB, 0xADFC },
 { 0x83EF, 0xD654 },
 { 0x83F0, 0xACE0 },
 { 0x83F1, 0xB2A5 },
 { 0x83F2, 0xBE44 },
 { 0x83F4, 0xC554 },
 { 0x83F9, 0xC800 },
 { 0x83FD, 0xC219 },
 { 0x8403, 0xCDCC },
 { 0x8404, 0xB3C4 },
 { 0x840A, 0xB798 },
 { 0x840C, 0xB9F9 },
 { 0x840D, 0xD3C9 },
 { 0x840E, 0xC704 },
 { 0x8429, 0xCD94 },
 { 0x842C, 0xB9CC },
 { 0x8431, 0xD6E4 },
 { 0x8438, 0xC720 },
 { 0x843D, 0xB099 },
 { 0x8449, 0xC12D },
 { 0x8457, 0xC800 },
 { 0x845B, 0xAC08 },
 { 0x8461, 0xD3EC },
 { 0x8463, 0xB3D9 },
 { 0x8466, 0xC704 },
 { 0x846B, 0xD638 },
 { 0x846C, 0xC7A5 },
 { 0x846F, 0xC57D },
 { 0x8475, 0xADDC },
 { 0x847A, 0xC999 },
 { 0x8490, 0xC218 },
 { 0x8494, 0xC2DC },
 { 0x8499, 0xBABD },
 { 0x849C, 0xC0B0 },
 { 0x84A1, 0xBC29 },
 { 0x84B2, 0xD3EC },
 { 0x84B8, 0xC99D },
 { 0x84BB, 0xC57D },
 { 0x84BC, 0xCC3D },
 { 0x84BF, 0xD638 },
 { 0x84C0, 0xC190 },
 { 0x84C2, 0xBA85 },
 { 0x84C4, 0xCD95 },
 { 0x84C6, 0xC11D },
 { 0x84C9, 0xC6A9 },
 { 0x84CB, 0xAC1C },
 { 0x84CD, 0xC2DC },
 { 0x84D1, 0xC0AC },
 { 0x84DA, 0xC218 },
 { 0x84EC, 0xBD09 },
 { 0x84EE, 0xB828 },
 { 0x84F4, 0xC21C },
 { 0x84FC, 0xB8CC },
 { 0x8511, 0xBA78 },
 { 0x8513, 0xB9CC },
 { 0x8514, 0xBCF5 },
 { 0x8517, 0xC790 },
 { 0x8518, 0xC0BC },
 { 0x851A, 0xC6B8 },
 { 0x851E, 0xB8E8 },
 { 0x8521, 0xCC44 },
 { 0x8523, 0xC7A5 },
 { 0x8525, 0xCD1D },
 { 0x852C, 0xC18C },
 { 0x852D, 0xC74C },
 { 0x852F, 0xC9C4 },
 { 0x853D, 0xD3D0 },
 { 0x853F, 0xC704 },
 { 0x8541, 0xB2F4 },
 { 0x8543, 0xBC88 },
 { 0x8549, 0xCD08 },
 { 0x854E, 0xAD50 },
 { 0x8553, 0xC6B4 },
 { 0x8559, 0xD61C },
 { 0x8563, 0xC21C },
 { 0x8568, 0xAD90 },
 { 0x8569, 0xD0D5 },
 { 0x856A, 0xBB34 },
 { 0x856D, 0xC18C },
 { 0x8584, 0xBC15 },
 { 0x8587, 0xBBF8 },
 { 0x858F, 0xC758 },
 { 0x8591, 0xAC15 },
 { 0x8594, 0xC7A5 },
 { 0x859B, 0xC124 },
 { 0x85A6, 0xCC9C },
 { 0x85A8, 0xD6D9 },
 { 0x85A9, 0xC0B4 },
 { 0x85AA, 0xC2E0 },
 { 0x85AF, 0xC11C },
 { 0x85B0, 0xD6C8 },
 { 0x85BA, 0xC81C },
 { 0x85C1, 0xACE0 },
 { 0x85C9, 0xC790 },
 { 0x85CD, 0xB0A8 },
 { 0x85CE, 0xC2E0 },
 { 0x85CF, 0xC7A5 },
 { 0x85D5, 0xC6B0 },
 { 0x85DC, 0xB824 },
 { 0x85DD, 0xC608 },
 { 0x85E4, 0xB4F1 },
 { 0x85E5, 0xC57D },
 { 0x85E9, 0xBC88 },
 { 0x85EA, 0xC218 },
 { 0x85F7, 0xC800 },
 { 0x85FA, 0xB9B0 },
 { 0x85FB, 0xC870 },
 { 0x85FF, 0xACFD },
 { 0x8602, 0xC608 },
 { 0x8606, 0xB178 },
 { 0x8607, 0xC18C },
 { 0x860A, 0xC628 },
 { 0x8616, 0xC5BC },
 { 0x8617, 0xBCBD },
 { 0x861A, 0xC120 },
 { 0x862D, 0xB09C },
 { 0x863F, 0xB098 },
 { 0x864E, 0xD638 },
 { 0x8650, 0xD559 },
 { 0x8654, 0xAC74 },
 { 0x8655, 0xCC98 },
 { 0x865B, 0xD5C8 },
 { 0x865C, 0xB178 },
 { 0x865E, 0xC6B0 },
 { 0x865F, 0xD638 },
 { 0x8667, 0xD734 },
 { 0x8679, 0xD64D },
 { 0x868A, 0xBB38 },
 { 0x868C, 0xBC29 },
 { 0x8693, 0xC778 },
 { 0x86A3, 0xACF5 },
 { 0x86A4, 0xC870 },
 { 0x86A9, 0xCE58 },
 { 0x86C7, 0xC0AC },
 { 0x86CB, 0xB2E8 },
 { 0x86D4, 0xD68C },
 { 0x86D9, 0xC640 },
 { 0x86DB, 0xC8FC },
 { 0x86DF, 0xAD50 },
 { 0x86E4, 0xD569 },
 { 0x86ED, 0xC9C8 },
 { 0x86FE, 0xC544 },
 { 0x8700, 0xCD09 },
 { 0x8702, 0xBD09 },
 { 0x8703, 0xC2E0 },
 { 0x8708, 0xC624 },
 { 0x8718, 0xC9C0 },
 { 0x871A, 0xBE44 },
 { 0x871C, 0xBC00 },
 { 0x874E, 0xAC08 },
 { 0x8755, 0xC2DD },
 { 0x8757, 0xD669 },
 { 0x875F, 0xC704 },
 { 0x8766, 0xD558 },
 { 0x8768, 0xC2AC },
 { 0x8774, 0xD638 },
 { 0x8776, 0xC811 },
 { 0x8778, 0xC640 },
 { 0x8782, 0xB791 },
 { 0x878D, 0xC735 },
 { 0x879F, 0xBA85 },
 { 0x87A2, 0xD615 },
 { 0x87B3, 0xB2F9 },
 { 0x87BA, 0xB098 },
 { 0x87C4, 0xCE69 },
 { 0x87E0, 0xBC18 },
 { 0x87EC, 0xC120 },
 { 0x87EF, 0xC694 },
 { 0x87F2, 0xCDA9 },
 { 0x87F9, 0xD574 },
 { 0x87FB, 0xC758 },
 { 0x87FE, 0xC12C },
 { 0x8805, 0xC2B9 },
 { 0x881F, 0xB0A9 },
 { 0x8822, 0xC900 },
 { 0x8823, 0xB824 },
 { 0x8831, 0xACE0 },
 { 0x8836, 0xC7A0 },
 { 0x883B, 0xB9CC },
 { 0x8840, 0xD608 },
 { 0x8846, 0xC911 },
 { 0x884C, 0xD56D },
 { 0x884D, 0xC5F0 },
 { 0x8852, 0xD604 },
 { 0x8853, 0xC220 },
 { 0x8857, 0xAC00 },
 { 0x8859, 0xC544 },
 { 0x885B, 0xC704 },
 { 0x885D, 0xCDA9 },
 { 0x8861, 0xD615 },
 { 0x8862, 0xAD6C },
 { 0x8863, 0xC758 },
 { 0x8868, 0xD45C },
 { 0x886B, 0xC0BC },
 { 0x8870, 0xC1E0 },
 { 0x8872, 0xB0A9 },
 { 0x8877, 0xCDA9 },
 { 0x887E, 0xAE08 },
 { 0x887F, 0xAE08 },
 { 0x8881, 0xC6D0 },
 { 0x8882, 0xBA8C },
 { 0x8888, 0xAC00 },
 { 0x888B, 0xB300 },
 { 0x888D, 0xD3EC },
 { 0x8892, 0xB2E8 },
 { 0x8896, 0xC218 },
 { 0x8897, 0xC9C4 },
 { 0x889E, 0xACE4 },
 { 0x88AB, 0xD53C },
 { 0x88B4, 0xACE0 },
 { 0x88C1, 0xC7AC },
 { 0x88C2, 0xB82C },
 { 0x88CF, 0xB9AC },
 { 0x88D4, 0xC608 },
 { 0x88D5, 0xC720 },
 { 0x88D9, 0xAD70 },
 { 0x88DC, 0xBCF4 },
 { 0x88DD, 0xC7A5 },
 { 0x88DF, 0xC0AC },
 { 0x88E1, 0xB9AC },
 { 0x88E8, 0xBE44 },
 { 0x88F3, 0xC0C1 },
 { 0x88F4, 0xBC30 },
 { 0x88F5, 0xBC30 },
 { 0x88F8, 0xB098 },
 { 0x88FD, 0xC81C },
 { 0x8907, 0xBCF5 },
 { 0x8910, 0xAC08 },
 { 0x8912, 0xD3EC },
 { 0x8913, 0xBCF4 },
 { 0x8918, 0xC704 },
 { 0x8919, 0xBC30 },
 { 0x8925, 0xC695 },
 { 0x892A, 0xD1F4 },
 { 0x8936, 0xC2B5 },
 { 0x8938, 0xB8E8 },
 { 0x893B, 0xC124 },
 { 0x8941, 0xAC15 },
 { 0x8944, 0xC591 },
 { 0x895F, 0xAE08 },
 { 0x8964, 0xB0A8 },
 { 0x896A, 0xB9D0 },
 { 0x8972, 0xC2B5 },
 { 0x897F, 0xC11C },
 { 0x8981, 0xC694 },
 { 0x8983, 0xB2F4 },
 { 0x8986, 0xBCF5 },
 { 0x8987, 0xD328 },
 { 0x898B, 0xACAC },
 { 0x898F, 0xADDC },
 { 0x8993, 0xBA71 },
 { 0x8996, 0xC2DC },
 { 0x89A1, 0xACA9 },
 { 0x89A9, 0xB3C4 },
 { 0x89AA, 0xCE5C },
 { 0x89B2, 0xADFC },
 { 0x89BA, 0xAC01 },
 { 0x89BD, 0xB78C },
 { 0x89C0, 0xAD00 },
 { 0x89D2, 0xAC01 },
 { 0x89E3, 0xD574 },
 { 0x89F4, 0xC0C1 },
 { 0x89F8, 0xCD09 },
 { 0x8A00, 0xC5B8 },
 { 0x8A02, 0xC815 },
 { 0x8A03, 0xBD80 },
 { 0x8A08, 0xACC4 },
 { 0x8A0A, 0xC2E0 },
 { 0x8A0C, 0xD64D },
 { 0x8A0E, 0xD1A0 },
 { 0x8A13, 0xD6C8 },
 { 0x8A16, 0xD758 },
 { 0x8A17, 0xD0C1 },
 { 0x8A18, 0xAE30 },
 { 0x8A1B, 0xC640 },
 { 0x8A1D, 0xC544 },
 { 0x8A1F, 0xC1A1 },
 { 0x8A23, 0xACB0 },
 { 0x8A25, 0xB20C },
 { 0x8A2A, 0xBC29 },
 { 0x8A2D, 0xC124 },
 { 0x8A31, 0xD5C8 },
 { 0x8A34, 0xC18C },
 { 0x8A36, 0xAC00 },
 { 0x8A3A, 0xC9C4 },
 { 0x8A3B, 0xC8FC },
 { 0x8A50, 0xC0AC },
 { 0x8A54, 0xC870 },
 { 0x8A55, 0xD3C9 },
 { 0x8A5B, 0xC800 },
 { 0x8A5E, 0xC0AC },
 { 0x8A60, 0xC601 },
 { 0x8A62, 0xC21C },
 { 0x8A63, 0xC608 },
 { 0x8A66, 0xC2DC },
 { 0x8A69, 0xC2DC },
 { 0x8A6D, 0xADA4 },
 { 0x8A6E, 0xC804 },
 { 0x8A70, 0xD790 },
 { 0x8A71, 0xD654 },
 { 0x8A72, 0xD574 },
 { 0x8A73, 0xC0C1 },
 { 0x8A75, 0xC120 },
 { 0x8A79, 0xCCA8 },
 { 0x8A85, 0xC8FC },
 { 0x8A87, 0xACFC },
 { 0x8A8C, 0xC9C0 },
 { 0x8A8D, 0xC778 },
 { 0x8A93, 0xC11C },
 { 0x8A95, 0xD0C4 },
 { 0x8A98, 0xC720 },
 { 0x8A9E, 0xC5B4 },
 { 0x8AA0, 0xC131 },
 { 0x8AA1, 0xACC4 },
 { 0x8AA3, 0xBB34 },
 { 0x8AA4, 0xC624 },
 { 0x8AA5, 0xACE0 },
 { 0x8AA6, 0xC1A1 },
 { 0x8AA8, 0xD68C },
 { 0x8AAA, 0xC124 },
 { 0x8AB0, 0xC218 },
 { 0x8AB2, 0xACFC },
 { 0x8AB9, 0xBE44 },
 { 0x8ABC, 0xC758 },
 { 0x8ABE, 0xC740 },
 { 0x8ABF, 0xC870 },
 { 0x8AC2, 0xCCA8 },
 { 0x8AC4, 0xC21C },
 { 0x8AC7, 0xB2F4 },
 { 0x8ACB, 0xCCAD },
 { 0x8ACD, 0xC7C1 },
 { 0x8ACF, 0xCD94 },
 { 0x8AD2, 0xB7C9 },
 { 0x8AD6, 0xB17C },
 { 0x8ADB, 0xC720 },
 { 0x8ADC, 0xCCA9 },
 { 0x8AE1, 0xC2DC },
 { 0x8AE6, 0xCCB4 },
 { 0x8AE7, 0xD574 },
 { 0x8AEA, 0xC815 },
 { 0x8AEB, 0xAC04 },
 { 0x8AED, 0xC720 },
 { 0x8AEE, 0xC790 },
 { 0x8AF1, 0xD718 },
 { 0x8AF6, 0xC2EC },
 { 0x8AF7, 0xD48D },
 { 0x8AF8, 0xC81C },
 { 0x8AFA, 0xC5B8 },
 { 0x8AFE, 0xB099 },
 { 0x8B00, 0xBAA8 },
 { 0x8B01, 0xC54C },
 { 0x8B02, 0xC704 },
 { 0x8B04, 0xB4F1 },
 { 0x8B0E, 0xBBF8 },
 { 0x8B10, 0xBC00 },
 { 0x8B14, 0xD559 },
 { 0x8B16, 0xC18D },
 { 0x8B17, 0xBC29 },
 { 0x8B19, 0xACB8 },
 { 0x8B1A, 0xC775 },
 { 0x8B1B, 0xAC15 },
 { 0x8B1D, 0xC0AC },
 { 0x8B20, 0xC694 },
 { 0x8B28, 0xBAA8 },
 { 0x8B2B, 0xC801 },
 { 0x8B2C, 0xB958 },
 { 0x8B33, 0xAD6C },
 { 0x8B39, 0xADFC },
 { 0x8B41, 0xD654 },
 { 0x8B49, 0xC99D },
 { 0x8B4E, 0xD73C },
 { 0x8B4F, 0xAE30 },
 { 0x8B58, 0xC2DD },
 { 0x8B5A, 0xB2F4 },
 { 0x8B5C, 0xBCF4 },
 { 0x8B66, 0xACBD },
 { 0x8B6C, 0xBE44 },
 { 0x8B6F, 0xC5ED },
 { 0x8B70, 0xC758 },
 { 0x8B74, 0xACAC },
 { 0x8B77, 0xD638 },
 { 0x8B7D, 0xC608 },
 { 0x8B80, 0xB3C5 },
 { 0x8B8A, 0xBCC0 },
 { 0x8B90, 0xC218 },
 { 0x8B92, 0xCC38 },
 { 0x8B93, 0xC591 },
 { 0x8B96, 0xCC38 },
 { 0x8B9A, 0xCC2C },
 { 0x8C37, 0xACE1 },
 { 0x8C3F, 0xACC4 },
 { 0x8C41, 0xD65C },
 { 0x8C46, 0xB450 },
 { 0x8C48, 0xAC1C },
 { 0x8C4A, 0xD48D },
 { 0x8C4C, 0xC644 },
 { 0x8C55, 0xC2DC },
 { 0x8C5A, 0xB3C8 },
 { 0x8C61, 0xC0C1 },
 { 0x8C6A, 0xD638 },
 { 0x8C6B, 0xC608 },
 { 0x8C79, 0xD45C },
 { 0x8C7A, 0xC2DC },
 { 0x8C82, 0xCD08 },
 { 0x8C8A, 0xB9E5 },
 { 0x8C8C, 0xBAA8 },
 { 0x8C9D, 0xD328 },
 { 0x8C9E, 0xC815 },
 { 0x8CA0, 0xBD80 },
 { 0x8CA1, 0xC7AC },
 { 0x8CA2, 0xACF5 },
 { 0x8CA7, 0xBE48 },
 { 0x8CA8, 0xD654 },
 { 0x8CA9, 0xD310 },
 { 0x8CAA, 0xD0D0 },
 { 0x8CAB, 0xAD00 },
 { 0x8CAC, 0xCC45 },
 { 0x8CAF, 0xC800 },
 { 0x8CB0, 0xC138 },
 { 0x8CB3, 0xC774 },
 { 0x8CB4, 0xADC0 },
 { 0x8CB6, 0xD3C4 },
 { 0x8CB7, 0xB9E4 },
 { 0x8CB8, 0xB300 },
 { 0x8CBB, 0xBE44 },
 { 0x8CBC, 0xCCA9 },
 { 0x8CBD, 0xC774 },
 { 0x8CBF, 0xBB34 },
 { 0x8CC0, 0xD558 },
 { 0x8CC1, 0xBD84 },
 { 0x8CC2, 0xB1CC },
 { 0x8CC3, 0xC784 },
 { 0x8CC4, 0xD68C },
 { 0x8CC7, 0xC790 },
 { 0x8CC8, 0xAC00 },
 { 0x8CCA, 0xC801 },
 { 0x8CD1, 0xC9C4 },
 { 0x8CD3, 0xBE48 },
 { 0x8CDA, 0xB8B0 },
 { 0x8CDC, 0xC0AC },
 { 0x8CDE, 0xC0C1 },
 { 0x8CE0, 0xBC30 },
 { 0x8CE2, 0xD604 },
 { 0x8CE3, 0xB9E4 },
 { 0x8CE4, 0xCC9C },
 { 0x8CE6, 0xBD80 },
 { 0x8CEA, 0xC9C8 },
 { 0x8CED, 0xB3C4 },
 { 0x8CF4, 0xB8B0 },
 { 0x8CFB, 0xBD80 },
 { 0x8CFC, 0xAD6C },
 { 0x8CFD, 0xC0C8 },
 { 0x8D04, 0xC9C0 },
 { 0x8D05, 0xCDE8 },
 { 0x8D07, 0xC724 },
 { 0x8D08, 0xC99D },
 { 0x8D0A, 0xCC2C },
 { 0x8D0D, 0xC12C },
 { 0x8D13, 0xC7A5 },
 { 0x8D16, 0xC18D },
 { 0x8D64, 0xC801 },
 { 0x8D66, 0xC0AC },
 { 0x8D6B, 0xD601 },
 { 0x8D70, 0xC8FC },
 { 0x8D73, 0xADDC },
 { 0x8D74, 0xBD80 },
 { 0x8D77, 0xAE30 },
 { 0x8D85, 0xCD08 },
 { 0x8D8A, 0xC6D4 },
 { 0x8D99, 0xC870 },
 { 0x8DA3, 0xCDE8 },
 { 0x8DA8, 0xCD94 },
 { 0x8DB3, 0xC871 },
 { 0x8DBA, 0xBD80 },
 { 0x8DBE, 0xC9C0 },
 { 0x8DC6, 0xD0DC },
 { 0x8DCB, 0xBC1C },
 { 0x8DCC, 0xC9C8 },
 { 0x8DCF, 0xAC00 },
 { 0x8DDB, 0xD30C },
 { 0x8DDD, 0xAC70 },
 { 0x8DE1, 0xC801 },
 { 0x8DE3, 0xC120 },
 { 0x8DE8, 0xACFC },
 { 0x8DEF, 0xB178 },
 { 0x8DF3, 0xB3C4 },
 { 0x8E0A, 0xC6A9 },
 { 0x8E0F, 0xB2F5 },
 { 0x8E10, 0xCC9C },
 { 0x8E1E, 0xAC70 },
 { 0x8E2A, 0xC885 },
 { 0x8E30, 0xC720 },
 { 0x8E35, 0xC885 },
 { 0x8E42, 0xC720 },
 { 0x8E44, 0xC81C },
 { 0x8E47, 0xAC74 },
 { 0x8E48, 0xB3C4 },
 { 0x8E49, 0xCC28 },
 { 0x8E4A, 0xD61C },
 { 0x8E59, 0xCD95 },
 { 0x8E5F, 0xC801 },
 { 0x8E60, 0xCC99 },
 { 0x8E74, 0xCD95 },
 { 0x8E76, 0xAD90 },
 { 0x8E81, 0xC870 },
 { 0x8E87, 0xC800 },
 { 0x8E8A, 0xC8FC },
 { 0x8E8D, 0xC57D },
 { 0x8EAA, 0xB9B0 },
 { 0x8EAB, 0xC2E0 },
 { 0x8EAC, 0xAD81 },
 { 0x8EC0, 0xAD6C },
 { 0x8ECA, 0xAC70 },
 { 0x8ECB, 0xC54C },
 { 0x8ECC, 0xADA4 },
 { 0x8ECD, 0xAD70 },
 { 0x8ED2, 0xD5CC },
 { 0x8EDF, 0xC5F0 },
 { 0x8EEB, 0xC9C4 },
 { 0x8EF8, 0xCD95 },
 { 0x8EFB, 0xAC00 },
 { 0x8EFE, 0xC2DD },
 { 0x8F03, 0xAD50 },
 { 0x8F05, 0xB85C },
 { 0x8F09, 0xC7AC },
 { 0x8F12, 0xCCA9 },
 { 0x8F13, 0xB9CC },
 { 0x8F14, 0xBCF4 },
 { 0x8F15, 0xACBD },
 { 0x8F1B, 0xB7C9 },
 { 0x8F1C, 0xCE58 },
 { 0x8F1D, 0xD718 },
 { 0x8F1E, 0xB9DD },
 { 0x8F1F, 0xCCA0 },
 { 0x8F26, 0xB828 },
 { 0x8F27, 0xBCD1 },
 { 0x8F29, 0xBC30 },
 { 0x8F2A, 0xB95C },
 { 0x8F2F, 0xC9D1 },
 { 0x8F33, 0xC8FC },
 { 0x8F38, 0xC218 },
 { 0x8F39, 0xBCF5 },
 { 0x8F3B, 0xBCF5 },
 { 0x8F3E, 0xC804 },
 { 0x8F3F, 0xC5EC },
 { 0x8F44, 0xD560 },
 { 0x8F45, 0xC6D0 },
 { 0x8F49, 0xC804 },
 { 0x8F4D, 0xCCA0 },
 { 0x8F4E, 0xAD50 },
 { 0x8F5D, 0xC5EC },
 { 0x8F5F, 0xAD49 },
 { 0x8F62, 0xB825 },
 { 0x8F9B, 0xC2E0 },
 { 0x8F9C, 0xACE0 },
 { 0x8FA3, 0xB784 },
 { 0x8FA6, 0xD310 },
 { 0x8FA8, 0xBCC0 },
 { 0x8FAD, 0xC0AC },
 { 0x8FAF, 0xBCC0 },
 { 0x8FB0, 0xC2E0 },
 { 0x8FB1, 0xC695 },
 { 0x8FB2, 0xB18D },
 { 0x8FC2, 0xC6B0 },
 { 0x8FC5, 0xC2E0 },
 { 0x8FCE, 0xC601 },
 { 0x8FD1, 0xADFC },
 { 0x8FD4, 0xBC18 },
 { 0x8FE6, 0xAC00 },
 { 0x8FEA, 0xC801 },
 { 0x8FEB, 0xBC15 },
 { 0x8FED, 0xC9C8 },
 { 0x8FF0, 0xC220 },
 { 0x8FF2, 0xAC81 },
 { 0x8FF7, 0xBBF8 },
 { 0x8FF9, 0xC801 },
 { 0x8FFD, 0xCD94 },
 { 0x9000, 0xD1F4 },
 { 0x9001, 0xC1A1 },
 { 0x9002, 0xAD04 },
 { 0x9003, 0xB3C4 },
 { 0x9005, 0xD6C4 },
 { 0x9006, 0xC5ED },
 { 0x9008, 0xD615 },
 { 0x900B, 0xD3EC },
 { 0x900D, 0xC18C },
 { 0x900F, 0xD22C },
 { 0x9010, 0xCD95 },
 { 0x9011, 0xAD6C },
 { 0x9014, 0xB3C4 },
 { 0x9015, 0xACBD },
 { 0x9017, 0xB450 },
 { 0x9019, 0xC800 },
 { 0x901A, 0xD1B5 },
 { 0x901D, 0xC11C },
 { 0x901E, 0xB839 },
 { 0x901F, 0xC18D },
 { 0x9020, 0xC870 },
 { 0x9021, 0xC900 },
 { 0x9022, 0xBD09 },
 { 0x9023, 0xB828 },
 { 0x902E, 0xCCB4 },
 { 0x9031, 0xC8FC },
 { 0x9032, 0xC9C4 },
 { 0x9035, 0xADDC },
 { 0x9038, 0xC77C },
 { 0x903C, 0xD54D },
 { 0x903E, 0xC720 },
 { 0x9041, 0xB454 },
 { 0x9042, 0xC218 },
 { 0x9047, 0xC6B0 },
 { 0x904A, 0xC720 },
 { 0x904B, 0xC6B4 },
 { 0x904D, 0xD3B8 },
 { 0x904E, 0xACFC },
 { 0x9050, 0xD558 },
 { 0x9051, 0xD669 },
 { 0x9053, 0xB3C4 },
 { 0x9054, 0xB2EC },
 { 0x9055, 0xC704 },
 { 0x9059, 0xC694 },
 { 0x905C, 0xC190 },
 { 0x905D, 0xB2F5 },
 { 0x905E, 0xCCB4 },
 { 0x9060, 0xC6D0 },
 { 0x9061, 0xC18C },
 { 0x9063, 0xACAC },
 { 0x9069, 0xC801 },
 { 0x906D, 0xC870 },
 { 0x906E, 0xCC28 },
 { 0x906F, 0xB454 },
 { 0x9072, 0xC9C0 },
 { 0x9075, 0xC900 },
 { 0x9077, 0xCC9C },
 { 0x9078, 0xC120 },
 { 0x907A, 0xC720 },
 { 0x907C, 0xB8CC },
 { 0x907D, 0xAC70 },
 { 0x907F, 0xD53C },
 { 0x9080, 0xC694 },
 { 0x9081, 0xB9E4 },
 { 0x9082, 0xD574 },
 { 0x9083, 0xC218 },
 { 0x9084, 0xD658 },
 { 0x9087, 0xC774 },
 { 0x9088, 0xB9C9 },
 { 0x908A, 0xBCC0 },
 { 0x908F, 0xB098 },
 { 0x9091, 0xC74D },
 { 0x9095, 0xC639 },
 { 0x9099, 0xB9DD },
 { 0x90A2, 0xD615 },
 { 0x90A3, 0xB098 },
 { 0x90A6, 0xBC29 },
 { 0x90A8, 0xCD0C },
 { 0x90AA, 0xC0AC },
 { 0x90AF, 0xAC10 },
 { 0x90B0, 0xD0DC },
 { 0x90B1, 0xAD6C },
 { 0x90B5, 0xC18C },
 { 0x90B8, 0xC800 },
 { 0x90C1, 0xC6B1 },
 { 0x90CA, 0xAD50 },
 { 0x90DE, 0xB791 },
 { 0x90E1, 0xAD70 },
 { 0x90E8, 0xBD80 },
 { 0x90ED, 0xACFD },
 { 0x90F5, 0xC6B0 },
 { 0x90FD, 0xB3C4 },
 { 0x9102, 0xC545 },
 { 0x9112, 0xCD94 },
 { 0x9115, 0xD5A5 },
 { 0x9119, 0xBE44 },
 { 0x9127, 0xB4F1 },
 { 0x912D, 0xC815 },
 { 0x9132, 0xB2E8 },
 { 0x9149, 0xC720 },
 { 0x914A, 0xC815 },
 { 0x914B, 0xCD94 },
 { 0x914C, 0xC791 },
 { 0x914D, 0xBC30 },
 { 0x914E, 0xC8FC },
 { 0x9152, 0xC8FC },
 { 0x9162, 0xCD08 },
 { 0x9169, 0xBA85 },
 { 0x916A, 0xB099 },
 { 0x916C, 0xC218 },
 { 0x9175, 0xD6A8 },
 { 0x9177, 0xD639 },
 { 0x9178, 0xC0B0 },
 { 0x9187, 0xC21C },
 { 0x9189, 0xCDE8 },
 { 0x918B, 0xCD08 },
 { 0x918D, 0xC81C },
 { 0x9192, 0xC131 },
 { 0x919C, 0xCD94 },
 { 0x91AB, 0xC758 },
 { 0x91AC, 0xC7A5 },
 { 0x91AE, 0xCD08 },
 { 0x91AF, 0xD61C },
 { 0x91B1, 0xBC1C },
 { 0x91B4, 0xB840 },
 { 0x91B5, 0xAC39 },
 { 0x91C0, 0xC591 },
 { 0x91C7, 0xCC44 },
 { 0x91C9, 0xC720 },
 { 0x91CB, 0xC11D },
 { 0x91CC, 0xB9AC },
 { 0x91CD, 0xC911 },
 { 0x91CE, 0xC57C },
 { 0x91CF, 0xB7C9 },
 { 0x91D0, 0xB9AC },
 { 0x91D1, 0xAE08 },
 { 0x91D7, 0xC1E0 },
 { 0x91D8, 0xC815 },
 { 0x91DC, 0xBD80 },
 { 0x91DD, 0xCE68 },
 { 0x91E3, 0xC870 },
 { 0x91E7, 0xCC9C },
 { 0x91EA, 0xC6B0 },
 { 0x91F5, 0xCC44 },
 { 0x920D, 0xB454 },
 { 0x9210, 0xAC80 },
 { 0x9211, 0xD310 },
 { 0x9212, 0xC0BD },
 { 0x9217, 0xC724 },
 { 0x921E, 0xADE0 },
 { 0x9234, 0xB839 },
 { 0x923A, 0xC625 },
 { 0x923F, 0xC804 },
 { 0x9240, 0xAC11 },
 { 0x9245, 0xAC70 },
 { 0x9249, 0xD604 },
 { 0x9257, 0xACB8 },
 { 0x925B, 0xC5F0 },
 { 0x925E, 0xC6D4 },
 { 0x9262, 0xBC1C },
 { 0x9264, 0xAD6C },
 { 0x9265, 0xC220 },
 { 0x9266, 0xC815 },
 { 0x9280, 0xC740 },
 { 0x9283, 0xCD1D },
 { 0x9285, 0xB3D9 },
 { 0x9291, 0xC120 },
 { 0x9293, 0xC804 },
 { 0x9296, 0xC218 },
 { 0x9298, 0xBA85 },
 { 0x929C, 0xD568 },
 { 0x92B3, 0xC608 },
 { 0x92B6, 0xAD6C },
 { 0x92B7, 0xC18C },
 { 0x92B9, 0xC218 },
 { 0x92CC, 0xC815 },
 { 0x92CF, 0xD611 },
 { 0x92D2, 0xBD09 },
 { 0x92E4, 0xC11C },
 { 0x92EA, 0xD3EC },
 { 0x92F8, 0xAC70 },
 { 0x92FC, 0xAC15 },
 { 0x9304, 0xB179 },
 { 0x9310, 0xCD94 },
 { 0x9318, 0xCD94 },
 { 0x931A, 0xC7C1 },
 { 0x931E, 0xC21C },
 { 0x931F, 0xB2F4 },
 { 0x9320, 0xC815 },
 { 0x9321, 0xAE30 },
 { 0x9322, 0xC804 },
 { 0x9324, 0xAE30 },
 { 0x9326, 0xAE08 },
 { 0x9328, 0xBB18 },
 { 0x932B, 0xC11D },
 { 0x932E, 0xACE0 },
 { 0x932F, 0xCC29 },
 { 0x9348, 0xC601 },
 { 0x934A, 0xB828 },
 { 0x934B, 0xACFC },
 { 0x934D, 0xB3C4 },
 { 0x9354, 0xC545 },
 { 0x935B, 0xB2E8 },
 { 0x936E, 0xC720 },
 { 0x9375, 0xAC74 },
 { 0x937C, 0xCE68 },
 { 0x937E, 0xC885 },
 { 0x938C, 0xACB8 },
 { 0x9394, 0xC6A9 },
 { 0x9396, 0xC1C4 },
 { 0x939A, 0xCD94 },
 { 0x93A3, 0xD615 },
 { 0x93A7, 0xAC1C },
 { 0x93AC, 0xD638 },
 { 0x93AD, 0xC9C4 },
 { 0x93B0, 0xC77C },
 { 0x93C3, 0xC871 },
 { 0x93D1, 0xC801 },
 { 0x93DE, 0xC6A9 },
 { 0x93E1, 0xACBD },
 { 0x93E4, 0xB8E8 },
 { 0x93F6, 0xC9D1 },
 { 0x9404, 0xD6A1 },
 { 0x9418, 0xC885 },
 { 0x9425, 0xC120 },
 { 0x942B, 0xC804 },
 { 0x9435, 0xCCA0 },
 { 0x9438, 0xD0C1 },
 { 0x9444, 0xC8FC },
 { 0x9451, 0xAC10 },
 { 0x9452, 0xAC10 },
 { 0x945B, 0xAD11 },
 { 0x947D, 0xCC2C },
 { 0x947F, 0xCC29 },
 { 0x9577, 0xC7A5 },
 { 0x9580, 0xBB38 },
 { 0x9583, 0xC12C },
 { 0x9589, 0xD3D0 },
 { 0x958B, 0xAC1C },
 { 0x958F, 0xC724 },
 { 0x9591, 0xD55C },
 { 0x9592, 0xD55C },
 { 0x9593, 0xAC04 },
 { 0x9594, 0xBBFC },
 { 0x9598, 0xAC11 },
 { 0x95A3, 0xAC01 },
 { 0x95A4, 0xD569 },
 { 0x95A5, 0xBC8C },
 { 0x95A8, 0xADDC },
 { 0x95AD, 0xB824 },
 { 0x95B1, 0xC5F4 },
 { 0x95BB, 0xC5FC },
 { 0x95BC, 0xC54C },
 { 0x95C7, 0xC554 },
 { 0x95CA, 0xD65C },
 { 0x95D4, 0xD569 },
 { 0x95D5, 0xAD90 },
 { 0x95D6, 0xD2C8 },
 { 0x95DC, 0xAD00 },
 { 0x95E1, 0xCC9C },
 { 0x95E2, 0xBCBD },
 { 0x961C, 0xBD80 },
 { 0x9621, 0xCC9C },
 { 0x962A, 0xD310 },
 { 0x962E, 0xC644 },
 { 0x9632, 0xBC29 },
 { 0x963B, 0xC870 },
 { 0x963F, 0xC544 },
 { 0x9640, 0xD0C0 },
 { 0x9642, 0xD53C },
 { 0x9644, 0xBD80 },
 { 0x964B, 0xB204 },
 { 0x964C, 0xB9E5 },
 { 0x964D, 0xAC15 },
 { 0x9650, 0xD55C },
 { 0x965B, 0xD3D0 },
 { 0x965C, 0xD569 },
 { 0x965D, 0xC12C },
 { 0x965E, 0xC2B9 },
 { 0x965F, 0xCC99 },
 { 0x9662, 0xC6D0 },
 { 0x9663, 0xC9C4 },
 { 0x9664, 0xC81C },
 { 0x966A, 0xBC30 },
 { 0x9670, 0xC74C },
 { 0x9673, 0xC9C4 },
 { 0x9675, 0xB2A5 },
 { 0x9676, 0xB3C4 },
 { 0x9677, 0xD568 },
 { 0x9678, 0xB959 },
 { 0x967D, 0xC591 },
 { 0x9685, 0xC6B0 },
 { 0x9686, 0xB96D },
 { 0x968A, 0xB300 },
 { 0x968B, 0xC218 },
 { 0x968D, 0xD669 },
 { 0x968E, 0xACC4 },
 { 0x9694, 0xACA9 },
 { 0x9695, 0xC6B4 },
 { 0x9698, 0xC560 },
 { 0x9699, 0xADF9 },
 { 0x969B, 0xC81C },
 { 0x969C, 0xC7A5 },
 { 0x96A3, 0xB9B0 },
 { 0x96A7, 0xC218 },
 { 0x96A8, 0xC218 },
 { 0x96AA, 0xD5D8 },
 { 0x96B1, 0xC740 },
 { 0x96B7, 0xB840 },
 { 0x96BB, 0xCC99 },
 { 0x96C0, 0xC791 },
 { 0x96C1, 0xC548 },
 { 0x96C4, 0xC6C5 },
 { 0x96C5, 0xC544 },
 { 0x96C6, 0xC9D1 },
 { 0x96C7, 0xACE0 },
 { 0x96C9, 0xCE58 },
 { 0x96CB, 0xC900 },
 { 0x96CC, 0xC790 },
 { 0x96CD, 0xC639 },
 { 0x96CE, 0xC800 },
 { 0x96D5, 0xC870 },
 { 0x96D6, 0xC218 },
 { 0x96D9, 0xC30D },
 { 0x96DB, 0xCD94 },
 { 0x96DC, 0xC7A1 },
 { 0x96E2, 0xB9AC },
 { 0x96E3, 0xB09C },
 { 0x96E8, 0xC6B0 },
 { 0x96E9, 0xC6B0 },
 { 0x96EA, 0xC124 },
 { 0x96EF, 0xBB38 },
 { 0x96F0, 0xBD84 },
 { 0x96F2, 0xC6B4 },
 { 0x96F6, 0xB839 },
 { 0x96F7, 0xB1CC },
 { 0x96F9, 0xBC15 },
 { 0x96FB, 0xC804 },
 { 0x9700, 0xC218 },
 { 0x9706, 0xC815 },
 { 0x9707, 0xC9C4 },
 { 0x9711, 0xC810 },
 { 0x9713, 0xC608 },
 { 0x9716, 0xB9BC },
 { 0x9719, 0xC601 },
 { 0x971C, 0xC0C1 },
 { 0x971E, 0xD558 },
 { 0x9727, 0xBB34 },
 { 0x9730, 0xC0B0 },
 { 0x9732, 0xB178 },
 { 0x9739, 0xBCBD },
 { 0x973D, 0xC81C },
 { 0x9742, 0xB825 },
 { 0x9744, 0xC560 },
 { 0x9748, 0xB839 },
 { 0x9751, 0xCCAD },
 { 0x9756, 0xC815 },
 { 0x975C, 0xC815 },
 { 0x975E, 0xBE44 },
 { 0x9761, 0xBBF8 },
 { 0x9762, 0xBA74 },
 { 0x9769, 0xD601 },
 { 0x976D, 0xC778 },
 { 0x9774, 0xD654 },
 { 0x9777, 0xC778 },
 { 0x977A, 0xB9D0 },
 { 0x978B, 0xD61C },
 { 0x978D, 0xC548 },
 { 0x978F, 0xACF5 },
 { 0x97A0, 0xAD6D },
 { 0x97A8, 0xAC08 },
 { 0x97AB, 0xAD6D },
 { 0x97AD, 0xD3B8 },
 { 0x97C6, 0xCC9C },
 { 0x97CB, 0xC704 },
 { 0x97D3, 0xD55C },
 { 0x97DC, 0xB3C4 },
 { 0x97F3, 0xC74C },
 { 0x97F6, 0xC18C },
 { 0x97FB, 0xC6B4 },
 { 0x97FF, 0xD5A5 },
 { 0x9800, 0xD638 },
 { 0x9801, 0xD608 },
 { 0x9802, 0xC815 },
 { 0x9803, 0xACBD },
 { 0x9805, 0xD56D },
 { 0x9806, 0xC21C },
 { 0x9808, 0xC218 },
 { 0x980A, 0xC6B1 },
 { 0x980C, 0xC1A1 },
 { 0x9810, 0xC608 },
 { 0x9811, 0xC644 },
 { 0x9812, 0xBC18 },
 { 0x9813, 0xB3C8 },
 { 0x9817, 0xD30C },
 { 0x9818, 0xB839 },
 { 0x982D, 0xB450 },
 { 0x9830, 0xD611 },
 { 0x9838, 0xACBD },
 { 0x9839, 0xD1F4 },
 { 0x983B, 0xBE48 },
 { 0x9846, 0xACFC },
 { 0x984C, 0xC81C },
 { 0x984D, 0xC561 },
 { 0x984E, 0xC545 },
 { 0x9854, 0xC548 },
 { 0x9858, 0xC6D0 },
 { 0x985A, 0xC804 },
 { 0x985E, 0xB958 },
 { 0x9865, 0xD638 },
 { 0x9867, 0xACE0 },
 { 0x986B, 0xC804 },
 { 0x986F, 0xD604 },
 { 0x98A8, 0xD48D },
 { 0x98AF, 0xC0BD },
 { 0x98B1, 0xD0DC },
 { 0x98C4, 0xD45C },
 { 0x98C7, 0xD45C },
 { 0x98DB, 0xBE44 },
 { 0x98DC, 0xBC88 },
 { 0x98DF, 0xC2DD },
 { 0x98E1, 0xC190 },
 { 0x98E2, 0xAE30 },
 { 0x98ED, 0xCE59 },
 { 0x98EE, 0xC74C },
 { 0x98EF, 0xBC18 },
 { 0x98F4, 0xC774 },
 { 0x98FC, 0xC0AC },
 { 0x98FD, 0xD3EC },
 { 0x98FE, 0xC2DD },
 { 0x9903, 0xAD50 },
 { 0x9909, 0xD5A5 },
 { 0x990A, 0xC591 },
 { 0x990C, 0xC774 },
 { 0x9910, 0xCC2C },
 { 0x9913, 0xC544 },
 { 0x9918, 0xC5EC },
 { 0x991E, 0xC804 },
 { 0x9920, 0xBCD1 },
 { 0x9928, 0xAD00 },
 { 0x9945, 0xB9CC },
 { 0x9949, 0xADFC },
 { 0x994B, 0xADA4 },
 { 0x994C, 0xCC2C },
 { 0x994D, 0xC120 },
 { 0x9951, 0xAE30 },
 { 0x9952, 0xC694 },
 { 0x9954, 0xC639 },
 { 0x9957, 0xD5A5 },
 { 0x9996, 0xC218 },
 { 0x9999, 0xD5A5 },
 { 0x999D, 0xD544 },
 { 0x99A5, 0xBCF5 },
 { 0x99A8, 0xD615 },
 { 0x99AC, 0xB9C8 },
 { 0x99AD, 0xC5B4 },
 { 0x99AE, 0xD48D },
 { 0x99B1, 0xD0C0 },
 { 0x99B3, 0xCE58 },
 { 0x99B4, 0xC21C },
 { 0x99B9, 0xC77C },
 { 0x99C1, 0xBC15 },
 { 0x99D0, 0xC8FC },
 { 0x99D1, 0xB178 },
 { 0x99D2, 0xAD6C },
 { 0x99D5, 0xAC00 },
 { 0x99D9, 0xBD80 },
 { 0x99DD, 0xD0C0 },
 { 0x99DF, 0xC0AC },
 { 0x99ED, 0xD574 },
 { 0x99F1, 0xB099 },
 { 0x99FF, 0xC900 },
 { 0x9A01, 0xBE59 },
 { 0x9A08, 0xBCD1 },
 { 0x9A0E, 0xAE30 },
 { 0x9A0F, 0xAE30 },
 { 0x9A19, 0xD3B8 },
 { 0x9A2B, 0xAC74 },
 { 0x9A30, 0xB4F1 },
 { 0x9A36, 0xCD94 },
 { 0x9A37, 0xC18C },
 { 0x9A40, 0xB9E5 },
 { 0x9A43, 0xD45C },
 { 0x9A45, 0xAD6C },
 { 0x9A4D, 0xD6A8 },
 { 0x9A55, 0xAD50 },
 { 0x9A57, 0xD5D8 },
 { 0x9A5A, 0xACBD },
 { 0x9A5B, 0xC5ED },
 { 0x9A5F, 0xCDE8 },
 { 0x9A62, 0xB824 },
 { 0x9A65, 0xAE30 },
 { 0x9A69, 0xD658 },
 { 0x9A6A, 0xB824 },
 { 0x9AA8, 0xACE8 },
 { 0x9AB8, 0xD574 },
 { 0x9AD3, 0xC218 },
 { 0x9AD4, 0xCCB4 },
 { 0x9AD8, 0xACE0 },
 { 0x9AE5, 0xC5FC },
 { 0x9AEE, 0xBC1C },
 { 0x9B1A, 0xC218 },
 { 0x9B27, 0xB8CC },
 { 0x9B2A, 0xD22C },
 { 0x9B31, 0xC6B8 },
 { 0x9B3C, 0xADC0 },
 { 0x9B41, 0xAD34 },
 { 0x9B42, 0xD63C },
 { 0x9B43, 0xBC1C },
 { 0x9B44, 0xBC31 },
 { 0x9B45, 0xB9E4 },
 { 0x9B4F, 0xC704 },
 { 0x9B54, 0xB9C8 },
 { 0x9B5A, 0xC5B4 },
 { 0x9B6F, 0xB178 },
 { 0x9B8E, 0xC810 },
 { 0x9B91, 0xD3EC },
 { 0x9B9F, 0xC548 },
 { 0x9BAB, 0xAD50 },
 { 0x9BAE, 0xC120 },
 { 0x9BC9, 0xB9AC },
 { 0x9BD6, 0xCCAD },
 { 0x9BE4, 0xACE4 },
 { 0x9BE8, 0xACBD },
 { 0x9C0D, 0xCD94 },
 { 0x9C10, 0xC545 },
 { 0x9C12, 0xBCF5 },
 { 0x9C15, 0xD558 },
 { 0x9C25, 0xD658 },
 { 0x9C32, 0xC624 },
 { 0x9C3B, 0xB9CC },
 { 0x9C47, 0xAC15 },
 { 0x9C49, 0xBCC4 },
 { 0x9C57, 0xB9B0 },
 { 0x9CE5, 0xC870 },
 { 0x9CE7, 0xBD80 },
 { 0x9CE9, 0xAD6C },
 { 0x9CF3, 0xBD09 },
 { 0x9CF4, 0xBA85 },
 { 0x9CF6, 0xC5F0 },
 { 0x9D09, 0xC544 },
 { 0x9D1B, 0xC6D0 },
 { 0x9D26, 0xC559 },
 { 0x9D28, 0xC555 },
 { 0x9D3B, 0xD64D },
 { 0x9D51, 0xACAC },
 { 0x9D5D, 0xC544 },
 { 0x9D60, 0xACE1 },
 { 0x9D61, 0xBB34 },
 { 0x9D6C, 0xBD95 },
 { 0x9D72, 0xC791 },
 { 0x9DA9, 0xBAA9 },
 { 0x9DAF, 0xC575 },
 { 0x9DB4, 0xD559 },
 { 0x9DC4, 0xACC4 },
 { 0x9DD7, 0xAD6C },
 { 0x9DF2, 0xCDE8 },
 { 0x9DF8, 0xD73C },
 { 0x9DF9, 0xC751 },
 { 0x9DFA, 0xB178 },
 { 0x9E1A, 0xC575 },
 { 0x9E1E, 0xB09C },
 { 0x9E75, 0xB85C },
 { 0x9E79, 0xD568 },
 { 0x9E7D, 0xC5FC },
 { 0x9E7F, 0xB179 },
 { 0x9E92, 0xAE30 },
 { 0x9E93, 0xB85D },
 { 0x9E97, 0xB824 },
 { 0x9E9D, 0xC0AC },
 { 0x9E9F, 0xB9B0 },
 { 0x9EA5, 0xB9E5 },
 { 0x9EB4, 0xAD6D },
 { 0x9EB5, 0xBA74 },
 { 0x9EBB, 0xB9C8 },
 { 0x9EBE, 0xD718 },
 { 0x9EC3, 0xD669 },
 { 0x9ECD, 0xC11C },
 { 0x9ECE, 0xB824 },
 { 0x9ED1, 0xD751 },
 { 0x9ED4, 0xAC80 },
 { 0x9ED8, 0xBB35 },
 { 0x9EDB, 0xB300 },
 { 0x9EDC, 0xCD9C },
 { 0x9EDE, 0xC810 },
 { 0x9EE8, 0xB2F9 },
 { 0x9EF4, 0xBBF8 },
 { 0x9F07, 0xC624 },
 { 0x9F08, 0xBCC4 },
 { 0x9F0E, 0xC815 },
 { 0x9F13, 0xACE0 },
 { 0x9F20, 0xC11C },
 { 0x9F3B, 0xBE44 },
 { 0x9F4A, 0xC81C },
 { 0x9F4B, 0xC7AC },
 { 0x9F4E, 0xC7AC },
 { 0x9F52, 0xCE58 },
 { 0x9F5F, 0xC800 },
 { 0x9F61, 0xB839 },
 { 0x9F67, 0xC124 },
 { 0x9F6A, 0xCC29 },
 { 0x9F6C, 0xC5B4 },
 { 0x9F77, 0xC545 },
 { 0x9F8D, 0xB8E1 },
 { 0x9F90, 0xBC29 },
 { 0x9F95, 0xAC10 },
 { 0x9F9C, 0xAD6C },
 { 0xF900, 0xAC1C },
 { 0xF901, 0xAC31 },
 { 0xF902, 0xAC70 },
 { 0xF903, 0xACE0 },
 { 0xF904, 0xACE8 },
 { 0xF905, 0xAD00 },
 { 0xF906, 0xADC0 },
 { 0xF907, 0xADC0 },
 { 0xF908, 0xADE0 },
 { 0xF909, 0xAE00 },
 { 0xF90A, 0xAE08 },
 { 0xF90B, 0xB098 },
 { 0xF90C, 0xB098 },
 { 0xF90D, 0xB098 },
 { 0xF90E, 0xB098 },
 { 0xF90F, 0xB098 },
 { 0xF910, 0xB098 },
 { 0xF911, 0xB098 },
 { 0xF912, 0xB098 },
 { 0xF913, 0xB098 },
 { 0xF914, 0xB099 },
 { 0xF915, 0xB099 },
 { 0xF916, 0xB099 },
 { 0xF917, 0xB099 },
 { 0xF918, 0xB099 },
 { 0xF919, 0xB099 },
 { 0xF91A, 0xB099 },
 { 0xF91B, 0xB09C },
 { 0xF91C, 0xB09C },
 { 0xF91D, 0xB09C },
 { 0xF91E, 0xB09C },
 { 0xF91F, 0xB09C },
 { 0xF920, 0xB09C },
 { 0xF921, 0xB0A8 },
 { 0xF922, 0xB0A8 },
 { 0xF923, 0xB0A8 },
 { 0xF924, 0xB0A8 },
 { 0xF925, 0xB0A9 },
 { 0xF926, 0xB0A9 },
 { 0xF927, 0xB0A9 },
 { 0xF928, 0xB0AD },
 { 0xF929, 0xB0AD },
 { 0xF92A, 0xB0AD },
 { 0xF92B, 0xB0AD },
 { 0xF92C, 0xB0AD },
 { 0xF92D, 0xB0B4 },
 { 0xF92E, 0xB0C9 },
 { 0xF92F, 0xB178 },
 { 0xF930, 0xB178 },
 { 0xF931, 0xB178 },
 { 0xF932, 0xB178 },
 { 0xF933, 0xB178 },
 { 0xF934, 0xB178 },
 { 0xF935, 0xB178 },
 { 0xF936, 0xB178 },
 { 0xF937, 0xB178 },
 { 0xF938, 0xB178 },
 { 0xF939, 0xB178 },
 { 0xF93A, 0xB178 },
 { 0xF93B, 0xB179 },
 { 0xF93C, 0xB179 },
 { 0xF93D, 0xB179 },
 { 0xF93E, 0xB179 },
 { 0xF93F, 0xB179 },
 { 0xF940, 0xB179 },
 { 0xF941, 0xB17C },
 { 0xF942, 0xB18D },
 { 0xF943, 0xB18D },
 { 0xF944, 0xB18D },
 { 0xF945, 0xB18D },
 { 0xF946, 0xB1CC },
 { 0xF947, 0xB1CC },
 { 0xF948, 0xB1CC },
 { 0xF949, 0xB1CC },
 { 0xF94A, 0xB204 },
 { 0xF94B, 0xB204 },
 { 0xF94C, 0xB204 },
 { 0xF94D, 0xB204 },
 { 0xF94E, 0xB204 },
 { 0xF94F, 0xB204 },
 { 0xF950, 0xB204 },
 { 0xF951, 0xB204 },
 { 0xF952, 0xB291 },
 { 0xF953, 0xB291 },
 { 0xF954, 0xB2A0 },
 { 0xF955, 0xB2A5 },
 { 0xF956, 0xB2A5 },
 { 0xF957, 0xB2A5 },
 { 0xF958, 0xB2A5 },
 { 0xF959, 0xB2A5 },
 { 0xF95A, 0xB450 },
 { 0xF95B, 0xB77C },
 { 0xF95C, 0xB77D },
 { 0xF95D, 0xB77D },
 { 0xF95E, 0xB780 },
 { 0xF95F, 0xB839 },
 { 0xF960, 0xB85C },
 { 0xF961, 0xB960 },
 { 0xF962, 0xB9AC },
 { 0xF963, 0xBC30 },
 { 0xF964, 0xBC88 },
 { 0xF965, 0xBCC0 },
 { 0xF966, 0xBD80 },
 { 0xF967, 0xBD88 },
 { 0xF968, 0xBE44 },
 { 0xF969, 0xC0AD },
 { 0xF96A, 0xC0AD },
 { 0xF96B, 0xC0BC },
 { 0xF96C, 0xC0C9 },
 { 0xF96D, 0xC0DD },
 { 0xF96E, 0xC12D },
 { 0xF96F, 0xC138 },
 { 0xF970, 0xC1C4 },
 { 0xF971, 0xC2E0 },
 { 0xF972, 0xC2EC },
 { 0xF973, 0xC2ED },
 { 0xF974, 0xC57C },
 { 0xF975, 0xC57D },
 { 0xF976, 0xC57D },
 { 0xF977, 0xC591 },
 { 0xF978, 0xC591 },
 { 0xF979, 0xC591 },
 { 0xF97A, 0xC591 },
 { 0xF97B, 0xC591 },
 { 0xF97C, 0xC591 },
 { 0xF97D, 0xC591 },
 { 0xF97E, 0xC591 },
 { 0xF97F, 0xC5EC },
 { 0xF980, 0xC5EC },
 { 0xF981, 0xC5EC },
 { 0xF982, 0xC5EC },
 { 0xF983, 0xC5EC },
 { 0xF984, 0xC5EC },
 { 0xF985, 0xC5EC },
 { 0xF986, 0xC5EC },
 { 0xF987, 0xC5EC },
 { 0xF988, 0xC5EC },
 { 0xF989, 0xC5EC },
 { 0xF98A, 0xC5ED },
 { 0xF98B, 0xC5ED },
 { 0xF98C, 0xC5ED },
 { 0xF98D, 0xC5ED },
 { 0xF98E, 0xC5F0 },
 { 0xF98F, 0xC5F0 },
 { 0xF990, 0xC5F0 },
 { 0xF991, 0xC5F0 },
 { 0xF992, 0xC5F0 },
 { 0xF993, 0xC5F0 },
 { 0xF994, 0xC5F0 },
 { 0xF995, 0xC5F0 },
 { 0xF996, 0xC5F0 },
 { 0xF997, 0xC5F0 },
 { 0xF998, 0xC5F0 },
 { 0xF999, 0xC5F0 },
 { 0xF99A, 0xC5F0 },
 { 0xF99B, 0xC5F0 },
 { 0xF99C, 0xC5F4 },
 { 0xF99D, 0xC5F4 },
 { 0xF99E, 0xC5F4 },
 { 0xF99F, 0xC5F4 },
 { 0xF9A0, 0xC5F4 },
 { 0xF9A1, 0xC5F4 },
 { 0xF9A2, 0xC5FC },
 { 0xF9A3, 0xC5FC },
 { 0xF9A4, 0xC5FC },
 { 0xF9A5, 0xC5FC },
 { 0xF9A6, 0xC5FC },
 { 0xF9A7, 0xC5FD },
 { 0xF9A8, 0xC601 },
 { 0xF9A9, 0xC601 },
 { 0xF9AA, 0xC601 },
 { 0xF9AB, 0xC601 },
 { 0xF9AC, 0xC601 },
 { 0xF9AD, 0xC601 },
 { 0xF9AE, 0xC601 },
 { 0xF9AF, 0xC601 },
 { 0xF9B0, 0xC601 },
 { 0xF9B1, 0xC601 },
 { 0xF9B2, 0xC601 },
 { 0xF9B3, 0xC601 },
 { 0xF9B4, 0xC601 },
 { 0xF9B5, 0xC608 },
 { 0xF9B6, 0xC608 },
 { 0xF9B7, 0xC608 },
 { 0xF9B8, 0xC608 },
 { 0xF9B9, 0xC624 },
 { 0xF9BA, 0xC694 },
 { 0xF9BB, 0xC694 },
 { 0xF9BC, 0xC694 },
 { 0xF9BD, 0xC694 },
 { 0xF9BE, 0xC694 },
 { 0xF9BF, 0xC694 },
 { 0xF9C0, 0xC694 },
 { 0xF9C1, 0xC694 },
 { 0xF9C2, 0xC694 },
 { 0xF9C3, 0xC694 },
 { 0xF9C4, 0xC6A9 },
 { 0xF9C5, 0xC6B4 },
 { 0xF9C6, 0xC6D0 },
 { 0xF9C7, 0xC720 },
 { 0xF9C8, 0xC720 },
 { 0xF9C9, 0xC720 },
 { 0xF9CA, 0xC720 },
 { 0xF9CB, 0xC720 },
 { 0xF9CC, 0xC720 },
 { 0xF9CD, 0xC720 },
 { 0xF9CE, 0xC720 },
 { 0xF9CF, 0xC720 },
 { 0xF9D0, 0xC720 },
 { 0xF9D1, 0xC721 },
 { 0xF9D2, 0xC721 },
 { 0xF9D3, 0xC721 },
 { 0xF9D4, 0xC724 },
 { 0xF9D5, 0xC724 },
 { 0xF9D6, 0xC724 },
 { 0xF9D7, 0xC724 },
 { 0xF9D8, 0xC728 },
 { 0xF9D9, 0xC728 },
 { 0xF9DA, 0xC728 },
 { 0xF9DB, 0xC728 },
 { 0xF9DC, 0xC735 },
 { 0xF9DD, 0xC774 },
 { 0xF9DE, 0xC774 },
 { 0xF9DF, 0xC774 },
 { 0xF9E0, 0xC774 },
 { 0xF9E1, 0xC774 },
 { 0xF9E2, 0xC774 },
 { 0xF9E3, 0xC774 },
 { 0xF9E4, 0xC774 },
 { 0xF9E5, 0xC774 },
 { 0xF9E6, 0xC774 },
 { 0xF9E7, 0xC774 },
 { 0xF9E8, 0xC774 },
 { 0xF9E9, 0xC774 },
 { 0xF9EA, 0xC774 },
 { 0xF9EB, 0xC775 },
 { 0xF9EC, 0xC775 },
 { 0xF9ED, 0xC778 },
 { 0xF9EE, 0xC778 },
 { 0xF9EF, 0xC778 },
 { 0xF9F0, 0xC778 },
 { 0xF9F1, 0xC778 },
 { 0xF9F2, 0xC778 },
 { 0xF9F3, 0xC778 },
 { 0xF9F4, 0xC784 },
 { 0xF9F5, 0xC784 },
 { 0xF9F6, 0xC784 },
 { 0xF9F7, 0xC785 },
 { 0xF9F8, 0xC785 },
 { 0xF9F9, 0xC785 },
 { 0xF9FA, 0xC7A5 },
 { 0xF9FB, 0xC801 },
 { 0xF9FC, 0xC9C0 },
 { 0xF9FD, 0xC9D1 },
 { 0xF9FE, 0xCC28 },
 { 0xF9FF, 0xCC99 },
 { 0xFA00, 0xCCB4 },
 { 0xFA01, 0xD0C1 },
 { 0xFA02, 0xD0C1 },
 { 0xFA03, 0xD0D5 },
 { 0xFA04, 0xD0DD },
 { 0xFA05, 0xD1B5 },
 { 0xFA06, 0xD3EC },
 { 0xFA07, 0xD3ED },
 { 0xFA08, 0xD56D },
 { 0xFA09, 0xD56D },
 { 0xFA0A, 0xD604 },
 { 0xFA0B, 0xD655 },
 { 0x4E1F, 0xC8FC },
 { 0x4E26, 0xBCD1 },
 { 0x4E2B, 0xC544 },
 { 0x4E30, 0xBD09 },
 { 0x4E31, 0xAD00 },
 { 0x4E67, 0xB458 },
 { 0x4E6C, 0xAC78 },
 { 0x4E7C, 0xC904 },
 { 0x4E8F, 0xC6B0 },
 { 0x4E99, 0xAE0D },
 { 0x4E9F, 0xADF9 },
 { 0x4EB3, 0xBC15 },
 { 0x4EB4, 0xC6B0 },
 { 0x4EB9, 0xBBF8 },
 { 0x4EC6, 0xBD80 },
 { 0x4EDE, 0xC778 },
 { 0x4EE1, 0xD758 },
 { 0x4EEE, 0xAC00 },
 { 0x4EF3, 0xBE44 },
 { 0x4EF5, 0xC624 },
 { 0x4EFF, 0xBC29 },
 { 0x4F00, 0xC885 },
 { 0x4F30, 0xACE0 },
 { 0x4F42, 0xC815 },
 { 0x4F49, 0xAD6C },
 { 0x4F4B, 0xC18C },
 { 0x4F53, 0xBD84 },
 { 0x4F54, 0xC810 },
 { 0x4F56, 0xD544 },
 { 0x4F57, 0xD0C0 },
 { 0x4F5D, 0xAD6C },
 { 0x4F5E, 0xB155 },
 { 0x4F5F, 0xD241 },
 { 0x4F6A, 0xD68C },
 { 0x4F78, 0xAD04 },
 { 0x4F79, 0xADA4 },
 { 0x4F7B, 0xC870 },
 { 0x4F7C, 0xAD50 },
 { 0x4F7D, 0xCC28 },
 { 0x4F89, 0xACFC },
 { 0x4F94, 0xBAA8 },
 { 0x4F97, 0xB3D9 },
 { 0x4F9A, 0xC21C },
 { 0x4F9C, 0xC8FC },
 { 0x4FB2, 0xC9C4 },
 { 0x4FBB, 0xD0C8 },
 { 0x4FC1, 0xC6B0 },
 { 0x4FC5, 0xAD6C },
 { 0x4FCC, 0xBCF4 },
 { 0x4FCF, 0xC18C },
 { 0x4FD0, 0xB9AC },
 { 0x4FD8, 0xBD80 },
 { 0x4FDB, 0xBA74 },
 { 0x4FE3, 0xC624 },
 { 0x4FF0, 0xD654 },
 { 0x4FF4, 0xCC9C },
 { 0x4FF6, 0xC219 },
 { 0x4FFE, 0xBE44 },
 { 0x5000, 0xCC3D },
 { 0x5005, 0xC26C },
 { 0x500F, 0xC219 },
 { 0x5014, 0xAD74 },
 { 0x5018, 0xB2F9 },
 { 0x5022, 0xCCA9 },
 { 0x5025, 0xACF5 },
 { 0x5029, 0xCC9C },
 { 0x502E, 0xB77C },
 { 0x5030, 0xB989 },
 { 0x5044, 0xB09C },
 { 0x504A, 0xC6B0 },
 { 0x504E, 0xC678 },
 { 0x5053, 0xC545 },
 { 0x5056, 0xCC28 },
 { 0x5058, 0xAC04 },
 { 0x5060, 0xC694 },
 { 0x5062, 0xCD08 },
 { 0x506A, 0xD54D },
 { 0x5070, 0xC124 },
 { 0x5072, 0xC2DC },
 { 0x508B, 0xAC15 },
 { 0x5092, 0xD61C },
 { 0x5094, 0xACB8 },
 { 0x5096, 0xCC3D },
 { 0x509A, 0xD6A8 },
 { 0x509B, 0xC6A9 },
 { 0x509D, 0xD0D1 },
 { 0x509E, 0xC0AC },
 { 0x50A2, 0xAC00 },
 { 0x50AE, 0xC870 },
 { 0x50B4, 0xAD6C },
 { 0x50BD, 0xC7A5 },
 { 0x50BF, 0xC5B8 },
 { 0x50C2, 0xB8E8 },
 { 0x50C4, 0xD45C },
 { 0x50DB, 0xAE30 },
 { 0x50E2, 0xCC9C },
 { 0x50E6, 0xCD94 },
 { 0x50E9, 0xD55C },
 { 0x50EC, 0xCD08 },
 { 0x50EE, 0xB3D9 },
 { 0x50F5, 0xAC15 },
 { 0x50FE, 0xC560 },
 { 0x5102, 0xB18D },
 { 0x5103, 0xCC9C },
 { 0x5107, 0xD604 },
 { 0x510B, 0xB2F4 },
 { 0x5110, 0xBE48 },
 { 0x5113, 0xB300 },
 { 0x5114, 0xC8FC },
 { 0x5115, 0xC81C },
 { 0x5117, 0xC758 },
 { 0x5118, 0xC9C4 },
 { 0x511A, 0xB9F9 },
 { 0x511B, 0xBB34 },
 { 0x511C, 0xB155 },
 { 0x5122, 0xB824 },
 { 0x5124, 0xD3EC },
 { 0x5125, 0xC721 },
 { 0x5127, 0xCC2C },
 { 0x512B, 0xD638 },
 { 0x5131, 0xB871 },
 { 0x5133, 0xCC38 },
 { 0x5135, 0xC219 },
 { 0x5139, 0xCC2C },
 { 0x513B, 0xB2F9 },
 { 0x5155, 0xC2DC },
 { 0x5157, 0xC5F0 },
 { 0x515A, 0xB2F9 },
 { 0x5189, 0xC5FC },
 { 0x518F, 0xACBD },
 { 0x5193, 0xAD6C },
 { 0x5198, 0xC720 },
 { 0x51A3, 0xCDE8 },
 { 0x51A4, 0xC6D0 },
 { 0x51B0, 0xBE59 },
 { 0x51B1, 0xD638 },
 { 0x51B2, 0xCDA9 },
 { 0x51C5, 0xACE0 },
 { 0x51CA, 0xCCAD },
 { 0x51D2, 0xC758 },
 { 0x51D4, 0xCC3D },
 { 0x51DB, 0xB984 },
 { 0x51E0, 0xADA4 },
 { 0x51E9, 0xBAA9 },
 { 0x51ED, 0xBE59 },
 { 0x51F3, 0xB4F1 },
 { 0x51F4, 0xBE59 },
 { 0x5201, 0xC870 },
 { 0x5213, 0xC644 },
 { 0x5216, 0xC6D4 },
 { 0x5226, 0xAC81 },
 { 0x5227, 0xAC81 },
 { 0x5231, 0xCC3D },
 { 0x5232, 0xADDC },
 { 0x5233, 0xACE0 },
 { 0x5244, 0xACBD },
 { 0x5249, 0xC88C },
 { 0x5255, 0xBE44 },
 { 0x5257, 0xC794 },
 { 0x525A, 0xC0AC },
 { 0x525E, 0xAE30 },
 { 0x525F, 0xCCA0 },
 { 0x5274, 0xAC1C },
 { 0x5277, 0xC0B0 },
 { 0x527A, 0xB9AC },
 { 0x5282, 0xADA4 },
 { 0x5293, 0xC758 },
 { 0x5297, 0xCC2C },
 { 0x5298, 0xB9C8 },
 { 0x52AC, 0xAD6C },
 { 0x52AD, 0xCD08 },
 { 0x52CC, 0xAD8C },
 { 0x52E0, 0xB959 },
 { 0x52E1, 0xD45C },
 { 0x52E6, 0xCD08 },
 { 0x5301, 0xBB38 },
 { 0x5302, 0xB0B4 },
 { 0x5303, 0xAC1C },
 { 0x530A, 0xAD6D },
 { 0x530C, 0xD569 },
 { 0x5327, 0xD611 },
 { 0x5331, 0xADA4 },
 { 0x5336, 0xAD6C },
 { 0x533D, 0xC5B8 },
 { 0x533E, 0xD3B8 },
 { 0x5345, 0xC0BD },
 { 0x5361, 0xC7A1 },
 { 0x536C, 0xC559 },
 { 0x536E, 0xCE58 },
 { 0x5372, 0xC18C },
 { 0x5373, 0xC989 },
 { 0x5378, 0xC0AC },
 { 0x5379, 0xD73C },
 { 0x537A, 0xADFC },
 { 0x537B, 0xAC01 },
 { 0x537E, 0xC545 },
 { 0x5396, 0xBC29 },
 { 0x539D, 0xC870 },
 { 0x53AA, 0xADFC },
 { 0x53AE, 0xC2DC },
 { 0x53B2, 0xB824 },
 { 0x53B9, 0xAD6C },
 { 0x53C5, 0xCC38 },
 { 0x53DF, 0xC218 },
 { 0x53E6, 0xB839 },
 { 0x53E8, 0xB3C4 },
 { 0x53EE, 0xC815 },
 { 0x53F4, 0xAD6C },
 { 0x53F5, 0xD30C },
 { 0x53F6, 0xD611 },
 { 0x5401, 0xC6B0 },
 { 0x540B, 0xCD0C },
 { 0x542C, 0xC740 },
 { 0x542E, 0xC804 },
 { 0x5436, 0xB20C },
 { 0x543D, 0xD6C4 },
 { 0x5440, 0xD558 },
 { 0x5446, 0xB9E4 },
 { 0x544E, 0xCC99 },
 { 0x545D, 0xC561 },
 { 0x5462, 0xB2C8 },
 { 0x5464, 0xB839 },
 { 0x5466, 0xC720 },
 { 0x546B, 0xCCA9 },
 { 0x5470, 0xC790 },
 { 0x5476, 0xB178 },
 { 0x547F, 0xAC70 },
 { 0x5484, 0xB3CC },
 { 0x5487, 0xD544 },
 { 0x548B, 0xC0AC },
 { 0x548D, 0xD574 },
 { 0x548F, 0xC601 },
 { 0x5496, 0xAC00 },
 { 0x54A0, 0xC9D1 },
 { 0x54A2, 0xC545 },
 { 0x54A5, 0xD76C },
 { 0x54AF, 0xAC01 },
 { 0x54BB, 0xD734 },
 { 0x54BC, 0xAD18 },
 { 0x54BF, 0xC774 },
 { 0x54C2, 0xC2E0 },
 { 0x54C3, 0xB3D9 },
 { 0x54C6, 0xCE58 },
 { 0x54C7, 0xC640 },
 { 0x54E6, 0xC544 },
 { 0x54E9, 0xB9AC },
 { 0x54F1, 0xBC1C },
 { 0x54FD, 0xACBD },
 { 0x54FF, 0xAC00 },
 { 0x5509, 0xC560 },
 { 0x550F, 0xD76C },
 { 0x5514, 0xC624 },
 { 0x552B, 0xAE08 },
 { 0x552E, 0xC218 },
 { 0x5533, 0xB824 },
 { 0x553C, 0xC0BD },
 { 0x5540, 0xC560 },
 { 0x5542, 0xB204 },
 { 0x5557, 0xB2F4 },
 { 0x555C, 0xCCA0 },
 { 0x557B, 0xC2DC },
 { 0x557D, 0xC554 },
 { 0x557E, 0xCD94 },
 { 0x5581, 0xC639 },
 { 0x5583, 0xB0A8 },
 { 0x558B, 0xCCA9 },
 { 0x5591, 0xC74C },
 { 0x5593, 0xC694 },
 { 0x5594, 0xC545 },
 { 0x559E, 0xC989 },
 { 0x559F, 0xC704 },
 { 0x55A3, 0xD6C4 },
 { 0x55A4, 0xD669 },
 { 0x55A8, 0xB7C9 },
 { 0x55B0, 0xC2DD },
 { 0x55C9, 0xC18C },
 { 0x55D1, 0xD569 },
 { 0x55DB, 0xACB8 },
 { 0x55E0, 0xB77D },
 { 0x55E2, 0xC62C },
 { 0x55F7, 0xC624 },
 { 0x55FF, 0xD0D0 },
 { 0x5604, 0xAD50 },
 { 0x5608, 0xC870 },
 { 0x560C, 0xD45C },
 { 0x560D, 0xB8E8 },
 { 0x560E, 0xC54C },
 { 0x5610, 0xAD50 },
 { 0x5612, 0xD61C },
 { 0x5616, 0xCC45 },
 { 0x5629, 0xD654 },
 { 0x562C, 0xCD5C },
 { 0x5637, 0xD638 },
 { 0x5638, 0xBB34 },
 { 0x5639, 0xB8CC },
 { 0x563B, 0xD76C },
 { 0x563F, 0xBB35 },
 { 0x5641, 0xC624 },
 { 0x5642, 0xC900 },
 { 0x5649, 0xB2F4 },
 { 0x564B, 0xD1A4 },
 { 0x564D, 0xCD08 },
 { 0x564E, 0xC5F4 },
 { 0x564F, 0xD761 },
 { 0x5664, 0xAE08 },
 { 0x5665, 0xB18D },
 { 0x5669, 0xC545 },
 { 0x566A, 0xC870 },
 { 0x566C, 0xC11C },
 { 0x566D, 0xAD50 },
 { 0x566F, 0xC560 },
 { 0x5672, 0xCF8C },
 { 0x5676, 0xAC08 },
 { 0x5678, 0xD1A4 },
 { 0x567A, 0xC2E0 },
 { 0x5680, 0xB155 },
 { 0x5684, 0xD68D },
 { 0x5687, 0xD558 },
 { 0x568F, 0xCCB4 },
 { 0x5699, 0xAD50 },
 { 0x569A, 0xC740 },
 { 0x56B3, 0xACE1 },
 { 0x56B6, 0xC575 },
 { 0x56C0, 0xC804 },
 { 0x56C1, 0xC12D },
 { 0x56C2, 0xD6A8 },
 { 0x56C3, 0xC7A1 },
 { 0x56C8, 0xC608 },
 { 0x56C9, 0xB77C },
 { 0x56DF, 0xC2E0 },
 { 0x56E6, 0xC5F0 },
 { 0x56EB, 0xD640 },
 { 0x56EE, 0xC640 },
 { 0x56F7, 0xADE0 },
 { 0x56FF, 0xC720 },
 { 0x5701, 0xC740 },
 { 0x5702, 0xD63C },
 { 0x5709, 0xC5B4 },
 { 0x570A, 0xCCAD },
 { 0x571C, 0xC6D0 },
 { 0x5725, 0xB85D },
 { 0x572C, 0xC624 },
 { 0x572E, 0xBE44 },
 { 0x573E, 0xAE09 },
 { 0x574C, 0xBD84 },
 { 0x574F, 0xBC30 },
 { 0x5767, 0xCC99 },
 { 0x5768, 0xC774 },
 { 0x5769, 0xAC10 },
 { 0x576F, 0xBC30 },
 { 0x5771, 0xC559 },
 { 0x5773, 0xC694 },
 { 0x5777, 0xAC00 },
 { 0x5778, 0xAD6C },
 { 0x577B, 0xC9C0 },
 { 0x578C, 0xB3D9 },
 { 0x5795, 0xD6C4 },
 { 0x579E, 0xD0C0 },
 { 0x57A4, 0xC9C8 },
 { 0x57B8, 0xC644 },
 { 0x57BD, 0xC740 },
 { 0x57C6, 0xAC01 },
 { 0x57C9, 0xD611 },
 { 0x57CF, 0xC5F0 },
 { 0x57D1, 0xCCA0 },
 { 0x57D2, 0xB784 },
 { 0x57DC, 0xC57C },
 { 0x57E4, 0xBE44 },
 { 0x57ED, 0xD0DC },
 { 0x57F3, 0xAC10 },
 { 0x57FB, 0xC900 },
 { 0x57FD, 0xC18C },
 { 0x5803, 0xACE4 },
 { 0x580B, 0xBD95 },
 { 0x5817, 0xB3CC },
 { 0x5819, 0xC778 },
 { 0x581D, 0xACFC },
 { 0x5820, 0xD6C4 },
 { 0x5823, 0xC6B0 },
 { 0x5826, 0xACC4 },
 { 0x5840, 0xBCD1 },
 { 0x5849, 0xCC99 },
 { 0x584C, 0xD0D1 },
 { 0x584D, 0xC2B9 },
 { 0x5850, 0xC18C },
 { 0x5852, 0xC2DC },
 { 0x5859, 0xACE0 },
 { 0x5869, 0xC5FC },
 { 0x587D, 0xC0C1 },
 { 0x5880, 0xC9C0 },
 { 0x5881, 0xB9CC },
 { 0x588A, 0xC810 },
 { 0x588D, 0xAE30 },
 { 0x5890, 0xADFC },
 { 0x589D, 0xC694 },
 { 0x58A1, 0xC120 },
 { 0x58A3, 0xBCF5 },
 { 0x58AB, 0xC900 },
 { 0x58B0, 0xB2F4 },
 { 0x58B1, 0xB4F1 },
 { 0x58C3, 0xAC15 },
 { 0x58D4, 0xB3C4 },
 { 0x58DA, 0xB85C },
 { 0x58DC, 0xB2F4 },
 { 0x58DD, 0xC720 },
 { 0x5906, 0xBD09 },
 { 0x5908, 0xC7AC },
 { 0x5910, 0xD615 },
 { 0x5924, 0xC778 },
 { 0x5925, 0xACFC },
 { 0x592C, 0xCF8C },
 { 0x592F, 0xD56D },
 { 0x5938, 0xACFC },
 { 0x5953, 0xCC28 },
 { 0x5958, 0xC7A5 },
 { 0x5969, 0xB834 },
 { 0x5976, 0xB0B4 },
 { 0x5991, 0xD30C },
 { 0x5992, 0xD22C },
 { 0x599D, 0xC7A5 },
 { 0x59A3, 0xBE44 },
 { 0x59A7, 0xC644 },
 { 0x59AF, 0xCD95 },
 { 0x59B2, 0xB2EC },
 { 0x59C8, 0xB839 },
 { 0x59CD, 0xC0B0 },
 { 0x59D2, 0xC0AC },
 { 0x59DE, 0xAE38 },
 { 0x59E3, 0xAD50 },
 { 0x59E4, 0xAD6C },
 { 0x59E5, 0xBAA8 },
 { 0x59FA, 0xC2E0 },
 { 0x5A09, 0xBE59 },
 { 0x5A13, 0xBBF8 },
 { 0x5A19, 0xD615 },
 { 0x5A23, 0xC81C },
 { 0x5A27, 0xD0DC },
 { 0x5A2D, 0xC560 },
 { 0x5A35, 0xCD94 },
 { 0x5A3F, 0xC544 },
 { 0x5A40, 0xC544 },
 { 0x5A4C, 0xC219 },
 { 0x5A50, 0xC640 },
 { 0x5A63, 0xC778 },
 { 0x5A67, 0xC815 },
 { 0x5A6A, 0xB78C },
 { 0x5A72, 0xD654 },
 { 0x5A77, 0xC815 },
 { 0x5A7F, 0xC11C },
 { 0x5A84, 0xBBF8 },
 { 0x5A8D, 0xBD80 },
 { 0x5A93, 0xD669 },
 { 0x5A95, 0xC554 },
 { 0x5A9F, 0xC124 },
 { 0x5AA2, 0xBAA8 },
 { 0x5AA7, 0xC65C },
 { 0x5AAA, 0xC624 },
 { 0x5AB3, 0xC2DD },
 { 0x5AB5, 0xC789 },
 { 0x5ABA, 0xBBF8 },
 { 0x5ABB, 0xBC18 },
 { 0x5ABC, 0xC628 },
 { 0x5ABD, 0xB9C8 },
 { 0x5ABE, 0xAD6C },
 { 0x5ABF, 0xAD34 },
 { 0x5AC8, 0xC575 },
 { 0x5ACB, 0xB1E8 },
 { 0x5AD6, 0xD45C },
 { 0x5AD7, 0xAD6C },
 { 0x5AD9, 0xC120 },
 { 0x5ADA, 0xB9CC },
 { 0x5ADB, 0xC608 },
 { 0x5AE0, 0xB9AC },
 { 0x5AE2, 0xADDC },
 { 0x5AE3, 0xC5B8 },
 { 0x5AE5, 0xC804 },
 { 0x5AEB, 0xBAA8 },
 { 0x5AED, 0xD638 },
 { 0x5AEE, 0xD638 },
 { 0x5AEF, 0xC624 },
 { 0x5AF6, 0xCD08 },
 { 0x5AFA, 0xD55C },
 { 0x5AFB, 0xD55C },
 { 0x5AFD, 0xB8CC },
 { 0x5B00, 0xADDC },
 { 0x5B08, 0xC694 },
 { 0x5B19, 0xC7A5 },
 { 0x5B1B, 0xD604 },
 { 0x5B25, 0xC870 },
 { 0x5B28, 0xC790 },
 { 0x5B2D, 0xB0B4 },
 { 0x5B30, 0xC601 },
 { 0x5B32, 0xB1E8 },
 { 0x5B34, 0xC601 },
 { 0x5B3E, 0xB780 },
 { 0x5B3F, 0xC5F0 },
 { 0x5B45, 0xC12C },
 { 0x5B4C, 0xB828 },
 { 0x5B56, 0xC790 },
 { 0x5B5B, 0xD328 },
 { 0x5B61, 0xD0DC },
 { 0x5B65, 0xB178 },
 { 0x5B76, 0xC790 },
 { 0x5B7F, 0xC0B0 },
 { 0x5B81, 0xC800 },
 { 0x5B82, 0xC6A9 },
 { 0x5B8A, 0xAC00 },
 { 0x5BD6, 0xCE68 },
 { 0x5BD8, 0xCE58 },
 { 0x5BED, 0xD61C },
 { 0x5BF0, 0xD658 },
 { 0x5C05, 0xADF9 },
 { 0x5C0C, 0xC8FC },
 { 0x5C14, 0xC774 },
 { 0x5C1F, 0xC120 },
 { 0x5C20, 0xC120 },
 { 0x5C2B, 0xC655 },
 { 0x5C41, 0xBE44 },
 { 0x5C47, 0xC804 },
 { 0x5C50, 0xADF9 },
 { 0x5C68, 0xAD6C },
 { 0x5C73, 0xC120 },
 { 0x5C7A, 0xAE30 },
 { 0x5C88, 0xD558 },
 { 0x5C8A, 0xC808 },
 { 0x5C8C, 0xAE09 },
 { 0x5C8F, 0xC644 },
 { 0x5C92, 0xACB8 },
 { 0x5C94, 0xCC28 },
 { 0x5C9D, 0xC791 },
 { 0x5CA3, 0xAD6C },
 { 0x5CA5, 0xD30C },
 { 0x5CA6, 0xB9BD },
 { 0x5CA7, 0xCD08 },
 { 0x5CA8, 0xC800 },
 { 0x5CAA, 0xBD88 },
 { 0x5CAD, 0xB839 },
 { 0x5CD2, 0xB3D9 },
 { 0x5CE9, 0xC544 },
 { 0x5CED, 0xCD08 },
 { 0x5CF1, 0xB178 },
 { 0x5D06, 0xACF5 },
 { 0x5D10, 0xACE4 },
 { 0x5D15, 0xC560 },
 { 0x5D1B, 0xAD74 },
 { 0x5D1F, 0xC74C },
 { 0x5D22, 0xC7C1 },
 { 0x5D24, 0xD6A8 },
 { 0x5D26, 0xC5C4 },
 { 0x5D34, 0xC678 },
 { 0x5D3D, 0xC7AC },
 { 0x5D41, 0xAC10 },
 { 0x5D42, 0xB960 },
 { 0x5D4E, 0xC6B0 },
 { 0x5D53, 0xC554 },
 { 0x5D54, 0xC678 },
 { 0x5D71, 0xC6A9 },
 { 0x5D81, 0xB8E8 },
 { 0x5D82, 0xC7A5 },
 { 0x5D84, 0xCC38 },
 { 0x5D92, 0xC99D },
 { 0x5D94, 0xAE08 },
 { 0x5D95, 0xCD08 },
 { 0x5D99, 0xB9B0 },
 { 0x5DA7, 0xC5ED },
 { 0x5DAB, 0xC5C5 },
 { 0x5DAE, 0xD5D8 },
 { 0x5DB0, 0xD574 },
 { 0x5DB7, 0xC758 },
 { 0x5DC9, 0xCC38 },
 { 0x5DCB, 0xADDC },
 { 0x5DD1, 0xCC2C },
 { 0x5DD3, 0xC804 },
 { 0x5DDA, 0xD5CC },
 { 0x5DF5, 0xCE58 },
 { 0x5DF9, 0xADFC },
 { 0x5E09, 0xBD84 },
 { 0x5E15, 0xB9D0 },
 { 0x5E1A, 0xCD94 },
 { 0x5E28, 0xC138 },
 { 0x5E37, 0xC720 },
 { 0x5E43, 0xC704 },
 { 0x5E4E, 0xBA71 },
 { 0x5E54, 0xB9CC },
 { 0x5E58, 0xCC45 },
 { 0x5E5E, 0xBCF5 },
 { 0x5E68, 0xCCA8 },
 { 0x5E6A, 0xBABD },
 { 0x5E6B, 0xBC29 },
 { 0x5E6C, 0xC8FC },
 { 0x5E70, 0xD5CC },
 { 0x5E71, 0xB780 },
 { 0x5E7A, 0xC694 },
 { 0x5E7F, 0xC5C4 },
 { 0x5E8B, 0xAE30 },
 { 0x5E96, 0xD3EC },
 { 0x5EB3, 0xBE44 },
 { 0x5EBD, 0xC6B0 },
 { 0x5EC1, 0xCE21 },
 { 0x5ECB, 0xC218 },
 { 0x5ED1, 0xADFC },
 { 0x5ED5, 0xC74C },
 { 0x5EDD, 0xC2DC },
 { 0x5EDE, 0xD760 },
 { 0x5EE1, 0xBB34 },
 { 0x5EE5, 0xAD34 },
 { 0x5EE7, 0xC7A5 },
 { 0x5EE8, 0xD574 },
 { 0x5EE9, 0xB984 },
 { 0x5EF1, 0xC639 },
 { 0x5EFF, 0xC785 },
 { 0x5F03, 0xAE30 },
 { 0x5F07, 0xAC10 },
 { 0x5F08, 0xD601 },
 { 0x5F0B, 0xC775 },
 { 0x5F22, 0xB3C4 },
 { 0x5F25, 0xBBF8 },
 { 0x5F2D, 0xBBF8 },
 { 0x5F34, 0xB3C8 },
 { 0x5F40, 0xAD6C },
 { 0x5F46, 0xBCC4 },
 { 0x5F54, 0xB85D },
 { 0x5F58, 0xCCB4 },
 { 0x5F5D, 0xC774 },
 { 0x5F64, 0xB3D9 },
 { 0x5F82, 0xC870 },
 { 0x5F88, 0xD754 },
 { 0x5F89, 0xC591 },
 { 0x5F9C, 0xC0C1 },
 { 0x5FA3, 0xCC28 },
 { 0x5FA7, 0xD3B8 },
 { 0x5FAC, 0xBC29 },
 { 0x5FAD, 0xC694 },
 { 0x5FAF, 0xD61C },
 { 0x5FBC, 0xC694 },
 { 0x5FC9, 0xB3C4 },
 { 0x5FD0, 0xD0D0 },
 { 0x5FD2, 0xD2B9 },
 { 0x5FDD, 0xCCA8 },
 { 0x5FDE, 0xBBFC },
 { 0x5FDF, 0xBBFC },
 { 0x5FE1, 0xCDA9 },
 { 0x5FE4, 0xC624 },
 { 0x5FE8, 0xC644 },
 { 0x5FED, 0xBCC0 },
 { 0x5FEE, 0xAE30 },
 { 0x5FF1, 0xCE68 },
 { 0x5FF8, 0xB274 },
 { 0x5FFB, 0xD754 },
 { 0x5FFC, 0xAC15 },
 { 0x600A, 0xCD08 },
 { 0x600D, 0xC791 },
 { 0x6014, 0xC815 },
 { 0x6015, 0xD30C },
 { 0x6017, 0xCCA9 },
 { 0x6019, 0xD638 },
 { 0x601B, 0xB2EC },
 { 0x6026, 0xD3C9 },
 { 0x6029, 0xB2C8 },
 { 0x6033, 0xD669 },
 { 0x6046, 0xD56D },
 { 0x6047, 0xAD11 },
 { 0x6048, 0xBAA8 },
 { 0x604A, 0xD611 },
 { 0x605A, 0xC5D0 },
 { 0x605F, 0xD749 },
 { 0x6060, 0xAD34 },
 { 0x606B, 0xD1B5 },
 { 0x6081, 0xC5F0 },
 { 0x6083, 0xACE4 },
 { 0x6084, 0xCD08 },
 { 0x608B, 0xB9B0 },
 { 0x608F, 0xD611 },
 { 0x6092, 0xC74D },
 { 0x6095, 0xD76C },
 { 0x6097, 0xBB38 },
 { 0x60A2, 0xB7C9 },
 { 0x60B1, 0xBE44 },
 { 0x60B5, 0xCC3D },
 { 0x60BB, 0xD589 },
 { 0x60BE, 0xACF5 },
 { 0x60C4, 0xB141 },
 { 0x60C6, 0xCD94 },
 { 0x60C9, 0xCCA8 },
 { 0x60CB, 0xC644 },
 { 0x60CF, 0xB78C },
 { 0x60D3, 0xAD8C },
 { 0x60D5, 0xCC99 },
 { 0x60D8, 0xB9DD },
 { 0x60D9, 0xCCA0 },
 { 0x60DB, 0xD63C },
 { 0x60DD, 0xCC3D },
 { 0x60E2, 0xC1C4 },
 { 0x60F2, 0xC6B4 },
 { 0x60F4, 0xCDCC },
 { 0x60F7, 0xC900 },
 { 0x60F8, 0xACBD },
 { 0x60FC, 0xD3B8 },
 { 0x6100, 0xCD08 },
 { 0x6103, 0xC120 },
 { 0x6114, 0xC74C },
 { 0x611C, 0xD611 },
 { 0x612C, 0xC18C },
 { 0x612D, 0xAE30 },
 { 0x6146, 0xB3C4 },
 { 0x614D, 0xC628 },
 { 0x6152, 0xC885 },
 { 0x615A, 0xCC38 },
 { 0x6160, 0xC624 },
 { 0x6171, 0xB2E8 },
 { 0x6173, 0xAC04 },
 { 0x6174, 0xC2B5 },
 { 0x6175, 0xC6A9 },
 { 0x617A, 0xB8E8 },
 { 0x6183, 0xC6A9 },
 { 0x618D, 0xAD50 },
 { 0x6192, 0xADA4 },
 { 0x6196, 0xC740 },
 { 0x61A8, 0xAC10 },
 { 0x61AF, 0xCC38 },
 { 0x61B8, 0xC12C },
 { 0x61BB, 0xD0C4 },
 { 0x61BC, 0xACBD },
 { 0x61C6, 0xC870 },
 { 0x61CC, 0xC5ED },
 { 0x61D5, 0xC5FC },
 { 0x61D7, 0xD558 },
 { 0x61DE, 0xBABD },
 { 0x61DF, 0xB300 },
 { 0x61E3, 0xBB38 },
 { 0x61FD, 0xD658 },
 { 0x61FE, 0xC12D },
 { 0x6214, 0xC794 },
 { 0x6215, 0xC7A5 },
 { 0x621E, 0xC54C },
 { 0x6222, 0xC9D1 },
 { 0x6227, 0xCC3D },
 { 0x6229, 0xC804 },
 { 0x6232, 0xD76C },
 { 0x6233, 0xCC29 },
 { 0x6235, 0xAD6C },
 { 0x6239, 0xC561 },
 { 0x6243, 0xACBD },
 { 0x624A, 0xC5FC },
 { 0x624E, 0xCC30 },
 { 0x6251, 0xBCF5 },
 { 0x6252, 0xBC30 },
 { 0x6257, 0xC7AC },
 { 0x625B, 0xAC15 },
 { 0x6263, 0xAD6C },
 { 0x6268, 0xC778 },
 { 0x6273, 0xBC18 },
 { 0x627A, 0xC9C0 },
 { 0x627E, 0xC870 },
 { 0x6283, 0xBCC0 },
 { 0x6286, 0xBB38 },
 { 0x6294, 0xBD80 },
 { 0x6296, 0xB450 },
 { 0x62A8, 0xD3C9 },
 { 0x62AC, 0xD0DC },
 { 0x62C4, 0xC8FC },
 { 0x62C5, 0xB2E8 },
 { 0x62C6, 0xD0C1 },
 { 0x62CA, 0xBD80 },
 { 0x62D1, 0xACB8 },
 { 0x62D5, 0xD0C0 },
 { 0x62F5, 0xC874 },
 { 0x62F6, 0xCC30 },
 { 0x62FD, 0xC608 },
 { 0x6302, 0xAD18 },
 { 0x630C, 0xACA9 },
 { 0x6310, 0xB098 },
 { 0x6312, 0xB82C },
 { 0x6328, 0xC560 },
 { 0x632A, 0xB098 },
 { 0x6339, 0xC74D },
 { 0x6342, 0xC624 },
 { 0x6343, 0xAD70 },
 { 0x6344, 0xAD6C },
 { 0x6346, 0xACE4 },
 { 0x634D, 0xD55C },
 { 0x634E, 0xC18C },
 { 0x6357, 0xCC99 },
 { 0x6369, 0xB82C },
 { 0x636B, 0xBB38 },
 { 0x6371, 0xC560 },
 { 0x6376, 0xCD94 },
 { 0x6380, 0xD754 },
 { 0x6384, 0xB95C },
 { 0x6387, 0xCCA0 },
 { 0x638A, 0xBD80 },
 { 0x638E, 0xAE30 },
 { 0x638F, 0xB3C4 },
 { 0x639C, 0xC608 },
 { 0x639F, 0xC815 },
 { 0x63AC, 0xAD6D },
 { 0x63BE, 0xC5F0 },
 { 0x63C3, 0xC804 },
 { 0x63C9, 0xC720 },
 { 0x63DF, 0xC11C },
 { 0x63E0, 0xC54C },
 { 0x63E3, 0xCDCC },
 { 0x63E9, 0xAC1C },
 { 0x63EB, 0xCD94 },
 { 0x63F2, 0xC124 },
 { 0x63F5, 0xAC74 },
 { 0x6406, 0xAD6C },
 { 0x6409, 0xAC01 },
 { 0x6412, 0xBC29 },
 { 0x6424, 0xC561 },
 { 0x6425, 0xCD94 },
 { 0x6428, 0xD0D1 },
 { 0x642A, 0xB2F9 },
 { 0x642B, 0xBC18 },
 { 0x642F, 0xB3C4 },
 { 0x6430, 0xACE8 },
 { 0x6436, 0xCC3D },
 { 0x6467, 0xCD5C },
 { 0x646D, 0xCC99 },
 { 0x6473, 0xAD6C },
 { 0x647B, 0xC12C },
 { 0x647D, 0xD45C },
 { 0x6485, 0xADA4 },
 { 0x6490, 0xD0F1 },
 { 0x6495, 0xC11C },
 { 0x6499, 0xC900 },
 { 0x649D, 0xD718 },
 { 0x649F, 0xAD50 },
 { 0x64A9, 0xB8CC },
 { 0x64BC, 0xAC10 },
 { 0x64BE, 0xACFC },
 { 0x64BF, 0xAC80 },
 { 0x64C9, 0xCC29 },
 { 0x64D0, 0xD658 },
 { 0x64D5, 0xD734 },
 { 0x64D7, 0xBCBD },
 { 0x64E0, 0xC81C },
 { 0x64E3, 0xB3C4 },
 { 0x64ED, 0xD654 },
 { 0x64EF, 0xBE48 },
 { 0x64F1, 0xAC01 },
 { 0x6502, 0xB8B0 },
 { 0x6507, 0xD5CC },
 { 0x650A, 0xB825 },
 { 0x650F, 0xB871 },
 { 0x6514, 0xB780 },
 { 0x6519, 0xCC38 },
 { 0x6522, 0xCC2C },
 { 0x6524, 0xD0C4 },
 { 0x6532, 0xAE30 },
 { 0x653D, 0xBC18 },
 { 0x6549, 0xBBF8 },
 { 0x6554, 0xC5B4 },
 { 0x6555, 0xCE59 },
 { 0x655D, 0xD3D0 },
 { 0x6581, 0xB450 },
 { 0x659D, 0xAC00 },
 { 0x65B1, 0xC791 },
 { 0x65B2, 0xCC29 },
 { 0x65BF, 0xC720 },
 { 0x65C2, 0xAE30 },
 { 0x65C3, 0xC804 },
 { 0x65C4, 0xBAA8 },
 { 0x65C6, 0xD328 },
 { 0x65CA, 0xBC29 },
 { 0x65F2, 0xB300 },
 { 0x6600, 0xC724 },
 { 0x6603, 0xCE21 },
 { 0x661B, 0xAC70 },
 { 0x6621, 0xD604 },
 { 0x6624, 0xB839 },
 { 0x662B, 0xAD6C },
 { 0x6635, 0xB2C9 },
 { 0x6637, 0xC628 },
 { 0x6648, 0xAD50 },
 { 0x664C, 0xC0C1 },
 { 0x664E, 0xD64D },
 { 0x6661, 0xD3EC },
 { 0x6662, 0xC808 },
 { 0x666C, 0xC218 },
 { 0x6670, 0xC11D },
 { 0x667B, 0xC554 },
 { 0x668B, 0xBBFC },
 { 0x668C, 0xADDC },
 { 0x669E, 0xAD50 },
 { 0x66A3, 0xAE30 },
 { 0x66AC, 0xC124 },
 { 0x66B1, 0xB2C9 },
 { 0x66B5, 0xD55C },
 { 0x66B8, 0xB8CC },
 { 0x66BF, 0xD76C },
 { 0x66C0, 0xC5D0 },
 { 0x66C1, 0xAE30 },
 { 0x66C5, 0xC5FD },
 { 0x66D3, 0xD3EC },
 { 0x66DA, 0xBABD },
 { 0x66DB, 0xD6C8 },
 { 0x66E3, 0xC5F0 },
 { 0x66E8, 0xB871 },
 { 0x66E9, 0xB0AD },
 { 0x66EA, 0xB77C },
 { 0x66EC, 0xC1C4 },
 { 0x66EE, 0xC5C4 },
 { 0x66EF, 0xCD09 },
 { 0x6705, 0xAC78 },
 { 0x670C, 0xBC18 },
 { 0x6723, 0xB3D9 },
 { 0x6733, 0xD314 },
 { 0x6745, 0xC6B0 },
 { 0x6760, 0xAC15 },
 { 0x676A, 0xCD08 },
 { 0x676C, 0xC6D0 },
 { 0x6772, 0xACE0 },
 { 0x677C, 0xC800 },
 { 0x678C, 0xBD84 },
 { 0x6798, 0xC608 },
 { 0x6799, 0xC640 },
 { 0x67B2, 0xC2DC },
 { 0x67C1, 0xD0C0 },
 { 0x67C5, 0xB2C8 },
 { 0x67CA, 0xC885 },
 { 0x67D7, 0xC1A1 },
 { 0x67D8, 0xC790 },
 { 0x67D9, 0xD569 },
 { 0x67DE, 0xC791 },
 { 0x67DF, 0xB0A8 },
 { 0x67E2, 0xC800 },
 { 0x67F9, 0xC2DC },
 { 0x6803, 0xD68C },
 { 0x681D, 0xAD04 },
 { 0x681E, 0xAC04 },
 { 0x682E, 0xC774 },
 { 0x6831, 0xACF5 },
 { 0x6832, 0xACE0 },
 { 0x6834, 0xC804 },
 { 0x683B, 0xC2DD },
 { 0x6844, 0xAD11 },
 { 0x6846, 0xAD11 },
 { 0x6849, 0xC548 },
 { 0x686F, 0xC815 },
 { 0x6874, 0xBD80 },
 { 0x6877, 0xAC01 },
 { 0x687E, 0xAD70 },
 { 0x6883, 0xC815 },
 { 0x6886, 0xBC29 },
 { 0x689B, 0xB098 },
 { 0x68A0, 0xB824 },
 { 0x68A3, 0xC2EC },
 { 0x68CA, 0xAE30 },
 { 0x68CC, 0xCC44 },
 { 0x68D0, 0xBE44 },
 { 0x68D6, 0xC815 },
 { 0x68E3, 0xCCB4 },
 { 0x68EC, 0xAD8C },
 { 0x68FB, 0xBD84 },
 { 0x68FC, 0xBD84 },
 { 0x6901, 0xACFD },
 { 0x6904, 0xC811 },
 { 0x690B, 0xB7C9 },
 { 0x6911, 0xBE44 },
 { 0x6919, 0xCC3D },
 { 0x6934, 0xB2E8 },
 { 0x6936, 0xC885 },
 { 0x6939, 0xCE68 },
 { 0x6942, 0xC0AC },
 { 0x694F, 0xADDC },
 { 0x695D, 0xB828 },
 { 0x6965, 0xC6D0 },
 { 0x696A, 0xC811 },
 { 0x696C, 0xAC08 },
 { 0x6973, 0xB9E4 },
 { 0x698E, 0xAC00 },
 { 0x6991, 0xBD80 },
 { 0x6994, 0xB791 },
 { 0x6998, 0xAD6C },
 { 0x699F, 0xC7AC },
 { 0x69A4, 0xAC78 },
 { 0x69A6, 0xAC04 },
 { 0x69AD, 0xC0AC },
 { 0x69B0, 0xC9C0 },
 { 0x69B1, 0xCD5C },
 { 0x69B7, 0xAD50 },
 { 0x69BC, 0xD569 },
 { 0x69BE, 0xACE8 },
 { 0x69C0, 0xACE0 },
 { 0x69C7, 0xC804 },
 { 0x69CA, 0xC0AD },
 { 0x69CE, 0xCC28 },
 { 0x69D3, 0xACF5 },
 { 0x69E5, 0xD61C },
 { 0x69E6, 0xC6A9 },
 { 0x69E7, 0xCC38 },
 { 0x69E9, 0xAC1C },
 { 0x69ED, 0xC0C9 },
 { 0x69F2, 0xACE1 },
 { 0x69F9, 0xACE0 },
 { 0x6A00, 0xC801 },
 { 0x6A05, 0xC885 },
 { 0x6A0B, 0xD1B5 },
 { 0x6A12, 0xBC00 },
 { 0x6A14, 0xC18C },
 { 0x6A1B, 0xADDC },
 { 0x6A2B, 0xACAC },
 { 0x6A3B, 0xADA4 },
 { 0x6A43, 0xBC8C },
 { 0x6A47, 0xCDE8 },
 { 0x6A4A, 0xB958 },
 { 0x6A50, 0xD0C1 },
 { 0x6A5A, 0xC219 },
 { 0x6A62, 0xD0C0 },
 { 0x6A66, 0xB3D9 },
 { 0x6A75, 0xC0B0 },
 { 0x6A7F, 0xAC15 },
 { 0x6A90, 0xCCA8 },
 { 0x6A94, 0xB2F9 },
 { 0x6A9D, 0xC999 },
 { 0x6A9F, 0xAC00 },
 { 0x6AA0, 0xACBD },
 { 0x6AAE, 0xB3C4 },
 { 0x6AB6, 0xC5FC },
 { 0x6ABC, 0xC740 },
 { 0x6ABF, 0xC5FC },
 { 0x6ADC, 0xACE0 },
 { 0x6ADD, 0xB3C5 },
 { 0x6ADE, 0xC5F0 },
 { 0x6ADF, 0xB825 },
 { 0x6AEA, 0xB825 },
 { 0x6AEC, 0xCE5C },
 { 0x6AFC, 0xCCA8 },
 { 0x6B02, 0xBC15 },
 { 0x6B03, 0xCC38 },
 { 0x6B11, 0xCC2C },
 { 0x6B1E, 0xB839 },
 { 0x6B2C, 0xD574 },
 { 0x6B37, 0xD76C },
 { 0x6B38, 0xC560 },
 { 0x6B39, 0xC758 },
 { 0x6B3B, 0xD6CC },
 { 0x6B3F, 0xAC10 },
 { 0x6B43, 0xC0BD },
 { 0x6B49, 0xACB8 },
 { 0x6B54, 0xD5C8 },
 { 0x6B59, 0xD761 },
 { 0x6B5A, 0xC120 },
 { 0x6B5B, 0xAC10 },
 { 0x6B60, 0xCCA0 },
 { 0x6B67, 0xAE30 },
 { 0x6B80, 0xC694 },
 { 0x6B82, 0xC870 },
 { 0x6B8D, 0xD45C },
 { 0x6B91, 0xAE0D },
 { 0x6BA2, 0xCCB4 },
 { 0x6BA4, 0xC0C1 },
 { 0x6BAB, 0xD0C4 },
 { 0x6BAD, 0xAC15 },
 { 0x6BB3, 0xC218 },
 { 0x6BBD, 0xD6A8 },
 { 0x6BC4, 0xACA9 },
 { 0x6BDA, 0xCC38 },
 { 0x6BEF, 0xB2F4 },
 { 0x6BF3, 0xCDE8 },
 { 0x6BFF, 0xC0B0 },
 { 0x6C05, 0xCC3D },
 { 0x6C10, 0xC800 },
 { 0x6C1B, 0xBD84 },
 { 0x6C24, 0xC778 },
 { 0x6C33, 0xC628 },
 { 0x6C3F, 0xADA4 },
 { 0x6C43, 0xD314 },
 { 0x6C4B, 0xC791 },
 { 0x6C4F, 0xB300 },
 { 0x6C52, 0xB9DD },
 { 0x6C54, 0xD758 },
 { 0x6C59, 0xC624 },
 { 0x6C5B, 0xC2E0 },
 { 0x6C5C, 0xC0AC },
 { 0x6C74, 0xD310 },
 { 0x6C84, 0xC6B4 },
 { 0x6C89, 0xCE68 },
 { 0x6C8D, 0xD638 },
 { 0x6C98, 0xBE44 },
 { 0x6CAC, 0xB9E4 },
 { 0x6CB0, 0xD0C1 },
 { 0x6CB1, 0xD0C0 },
 { 0x6CC5, 0xC218 },
 { 0x6CC6, 0xC77C },
 { 0x6CD0, 0xB975 },
 { 0x6CD1, 0xC720 },
 { 0x6CD4, 0xAC10 },
 { 0x6CD9, 0xD3C9 },
 { 0x6CDA, 0xC790 },
 { 0x6CDC, 0xC9C0 },
 { 0x6CDD, 0xC18C },
 { 0x6CE0, 0xB839 },
 { 0x6CF1, 0xC559 },
 { 0x6D04, 0xD68C },
 { 0x6D07, 0xC778 },
 { 0x6D0A, 0xCC9C },
 { 0x6D12, 0xC138 },
 { 0x6D1F, 0xC774 },
 { 0x6D2B, 0xD601 },
 { 0x6D2C, 0xC18D },
 { 0x6D2E, 0xB3C4 },
 { 0x6D3A, 0xBA85 },
 { 0x6D3C, 0xC640 },
 { 0x6D61, 0xBC1C },
 { 0x6D64, 0xAD49 },
 { 0x6D65, 0xC74D },
 { 0x6D67, 0xC601 },
 { 0x6D82, 0xB3C4 },
 { 0x6D94, 0xC7A0 },
 { 0x6D96, 0xB9AC },
 { 0x6DAA, 0xBD80 },
 { 0x6DAB, 0xAD00 },
 { 0x6DB4, 0xC644 },
 { 0x6DB7, 0xB3D9 },
 { 0x6DB8, 0xD6C4 },
 { 0x6DBC, 0xB7C9 },
 { 0x6DBF, 0xD0C1 },
 { 0x6DC8, 0xAD74 },
 { 0x6DD2, 0xCC98 },
 { 0x6DD6, 0xB1E8 },
 { 0x6DDD, 0xBE44 },
 { 0x6DE0, 0xBE44 },
 { 0x6DE2, 0xC5ED },
 { 0x6DE4, 0xC5B4 },
 { 0x6DE5, 0xB85D },
 { 0x6DE6, 0xAC10 },
 { 0x6DEC, 0xC26C },
 { 0x6DF6, 0xB798 },
 { 0x6DFC, 0xBB18 },
 { 0x6E1D, 0xD22C },
 { 0x6E28, 0xC678 },
 { 0x6E37, 0xC5F0 },
 { 0x6E45, 0xB828 },
 { 0x6E4E, 0xBA74 },
 { 0x6E51, 0xC11C },
 { 0x6E53, 0xBD84 },
 { 0x6E54, 0xC804 },
 { 0x6E63, 0xBBFC },
 { 0x6E8F, 0xB2F9 },
 { 0x6E95, 0xBABD },
 { 0x6E98, 0xD569 },
 { 0x6EA7, 0xB960 },
 { 0x6EB2, 0xC218 },
 { 0x6EB5, 0xC740 },
 { 0x6EB7, 0xD63C },
 { 0x6EBD, 0xC695 },
 { 0x6EC3, 0xC639 },
 { 0x6EC8, 0xD638 },
 { 0x6ED5, 0xB4F1 },
 { 0x6ED9, 0xD68C },
 { 0x6EEC, 0xD638 },
 { 0x6EED, 0xD544 },
 { 0x6EF5, 0xBC00 },
 { 0x6EF7, 0xB85C },
 { 0x6F04, 0xC560 },
 { 0x6F09, 0xB85D },
 { 0x6F0C, 0xADFC },
 { 0x6F13, 0xB9AC },
 { 0x6F19, 0xB2E8 },
 { 0x6F1A, 0xAD6C },
 { 0x6F2A, 0xC758 },
 { 0x6F2D, 0xB9DD },
 { 0x6F30, 0xBD95 },
 { 0x6F33, 0xC7A5 },
 { 0x6F3C, 0xCD5C },
 { 0x6F3E, 0xC591 },
 { 0x6F4F, 0xC728 },
 { 0x6F59, 0xADDC },
 { 0x6F5C, 0xC7A0 },
 { 0x6F5D, 0xD761 },
 { 0x6F61, 0xB3C8 },
 { 0x6F66, 0xB8CC },
 { 0x6F6F, 0xC2EC },
 { 0x6F74, 0xC800 },
 { 0x6F78, 0xC0B0 },
 { 0x6F82, 0xC9D5 },
 { 0x6F86, 0xC694 },
 { 0x6F89, 0xAC10 },
 { 0x6F8B, 0xD6A1 },
 { 0x6F8C, 0xC2DC },
 { 0x6F92, 0xD64D },
 { 0x6F98, 0xC0B0 },
 { 0x6F9F, 0xB984 },
 { 0x6FA1, 0xC870 },
 { 0x6FA8, 0xC11C },
 { 0x6FB6, 0xC804 },
 { 0x6FC6, 0xBD84 },
 { 0x6FC7, 0xC0C9 },
 { 0x6FC9, 0xC218 },
 { 0x6FD8, 0xB155 },
 { 0x6FDB, 0xBABD },
 { 0x6FDE, 0xBE44 },
 { 0x6FEE, 0xBCF5 },
 { 0x6FF3, 0xC7A0 },
 { 0x6FF4, 0xC601 },
 { 0x6FFA, 0xCC9C },
 { 0x7007, 0xC655 },
 { 0x7012, 0xC0C9 },
 { 0x7013, 0xC9D5 },
 { 0x7026, 0xC800 },
 { 0x7030, 0xBBF8 },
 { 0x7032, 0xB834 },
 { 0x7038, 0xCCA8 },
 { 0x703C, 0xC591 },
 { 0x7044, 0xC12D },
 { 0x704E, 0xC5FC },
 { 0x7053, 0xB780 },
 { 0x705E, 0xD30C },
 { 0x7069, 0xC5FC },
 { 0x706F, 0xC815 },
 { 0x707E, 0xC7AC },
 { 0x7081, 0xAE30 },
 { 0x7086, 0xBB38 },
 { 0x7095, 0xD56D },
 { 0x70AE, 0xD3EC },
 { 0x70B0, 0xD3EC },
 { 0x70CA, 0xC591 },
 { 0x70D3, 0xACC4 },
 { 0x70DC, 0xD6E4 },
 { 0x711C, 0xD63C },
 { 0x7120, 0xC26C },
 { 0x712D, 0xACBD },
 { 0x712E, 0xD754 },
 { 0x712F, 0xC791 },
 { 0x7131, 0xD601 },
 { 0x7146, 0xD558 },
 { 0x7152, 0xC704 },
 { 0x7157, 0xB09C },
 { 0x7162, 0xACBD },
 { 0x7168, 0xC678 },
 { 0x7180, 0xC5FD },
 { 0x7185, 0xC628 },
 { 0x7187, 0xD6A8 },
 { 0x719B, 0xD45C },
 { 0x71A0, 0xC2B5 },
 { 0x71A4, 0xC775 },
 { 0x71A8, 0xC704 },
 { 0x71AF, 0xC120 },
 { 0x71B2, 0xACBD },
 { 0x71CB, 0xCD08 },
 { 0x71D6, 0xC2EC },
 { 0x71D9, 0xD0D5 },
 { 0x71E0, 0xC6B1 },
 { 0x71EC, 0xD6FC },
 { 0x71F9, 0xD76C },
 { 0x7207, 0xC124 },
 { 0x720B, 0xD6C8 },
 { 0x720C, 0xAD11 },
 { 0x720D, 0xC0AD },
 { 0x7214, 0xD76C },
 { 0x7217, 0xC5FD },
 { 0x721A, 0xC57D },
 { 0x721F, 0xAD00 },
 { 0x7225, 0xCD09 },
 { 0x7228, 0xCC2C },
 { 0x7238, 0xD30C },
 { 0x7239, 0xB2E4 },
 { 0x7242, 0xC7A5 },
 { 0x724B, 0xC804 },
 { 0x7253, 0xBC29 },
 { 0x7254, 0xBC15 },
 { 0x7256, 0xC720 },
 { 0x725A, 0xD0F1 },
 { 0x726F, 0xACE0 },
 { 0x7274, 0xC800 },
 { 0x7278, 0xC790 },
 { 0x7287, 0xBD84 },
 { 0x728D, 0xAC74 },
 { 0x7292, 0xD638 },
 { 0x7296, 0xB77D },
 { 0x72B3, 0xC791 },
 { 0x72B4, 0xC548 },
 { 0x72B5, 0xD790 },
 { 0x72C9, 0xBE44 },
 { 0x72D2, 0xBE44 },
 { 0x72E2, 0xD559 },
 { 0x72E5, 0xC21C },
 { 0x72E8, 0xC735 },
 { 0x72F4, 0xD3D0 },
 { 0x72F7, 0xACAC },
 { 0x72FA, 0xC740 },
 { 0x72FB, 0xC0B0 },
 { 0x7309, 0xAE30 },
 { 0x7313, 0xACFC },
 { 0x7317, 0xC758 },
 { 0x7318, 0xC81C },
 { 0x7319, 0xC7C1 },
 { 0x7322, 0xD638 },
 { 0x7327, 0xC640 },
 { 0x7328, 0xC6D0 },
 { 0x7331, 0xB178 },
 { 0x7334, 0xD6C4 },
 { 0x7343, 0xC560 },
 { 0x7358, 0xD3D0 },
 { 0x735C, 0xB9B0 },
 { 0x7360, 0xB8CC },
 { 0x7369, 0xC608 },
 { 0x736B, 0xD5D8 },
 { 0x736C, 0xD574 },
 { 0x736F, 0xD6C8 },
 { 0x7377, 0xAD11 },
 { 0x737C, 0xBBF8 },
 { 0x7381, 0xD5D8 },
 { 0x7388, 0xB85C },
 { 0x7395, 0xAC04 },
 { 0x739E, 0xBD80 },
 { 0x73A0, 0xAC1C },
 { 0x73A6, 0xACB0 },
 { 0x73AB, 0xB9E4 },
 { 0x73B7, 0xC810 },
 { 0x73BB, 0xD30C },
 { 0x73C6, 0xC774 },
 { 0x73C8, 0xAC00 },
 { 0x73D2, 0xC9C4 },
 { 0x73E2, 0xC740 },
 { 0x73EB, 0xCDA9 },
 { 0x73EE, 0xD328 },
 { 0x73F5, 0xC815 },
 { 0x7404, 0xD604 },
 { 0x7424, 0xC7C1 },
 { 0x7431, 0xC870 },
 { 0x7432, 0xBC30 },
 { 0x7445, 0xC81C },
 { 0x7446, 0xC131 },
 { 0x744E, 0xD574 },
 { 0x745D, 0xD669 },
 { 0x7463, 0xC1C4 },
 { 0x746D, 0xB2F9 },
 { 0x7470, 0xAD34 },
 { 0x7471, 0xC804 },
 { 0x7472, 0xCC3D },
 { 0x7473, 0xCC28 },
 { 0x747D, 0xC885 },
 { 0x7480, 0xCD5C },
 { 0x7486, 0xAD6C },
 { 0x748A, 0xBB38 },
 { 0x749D, 0xAD34 },
 { 0x74AB, 0xB2F9 },
 { 0x74DB, 0xD658 },
 { 0x74E9, 0xCC9C },
 { 0x74F2, 0xD1A4 },
 { 0x7503, 0xCD94 },
 { 0x750C, 0xAD6C },
 { 0x750D, 0xB9F9 },
 { 0x750E, 0xC804 },
 { 0x7513, 0xBCBD },
 { 0x751E, 0xC0C1 },
 { 0x753F, 0xB9F9 },
 { 0x754A, 0xACBD },
 { 0x754B, 0xC804 },
 { 0x754E, 0xACAC },
 { 0x755A, 0xBD84 },
 { 0x7560, 0xC804 },
 { 0x7564, 0xCE58 },
 { 0x7567, 0xB7B5 },
 { 0x756B, 0xD654 },
 { 0x7588, 0xBCBD },
 { 0x7594, 0xC815 },
 { 0x7599, 0xD758 },
 { 0x759A, 0xAD6C },
 { 0x75A3, 0xC6B0 },
 { 0x75A9, 0xCDCC },
 { 0x75B4, 0xC544 },
 { 0x75C0, 0xAD6C },
 { 0x75C3, 0xD604 },
 { 0x75CA, 0xC804 },
 { 0x75CE, 0xD574 },
 { 0x75D3, 0xCE58 },
 { 0x75DE, 0xBE44 },
 { 0x75E3, 0xC9C0 },
 { 0x75E4, 0xC88C },
 { 0x75E7, 0xC0AC },
 { 0x75F3, 0xB9BC },
 { 0x75F9, 0xBE44 },
 { 0x75FF, 0xC704 },
 { 0x7601, 0xCDCC },
 { 0x7607, 0xC885 },
 { 0x7608, 0xACC4 },
 { 0x7609, 0xC720 },
 { 0x760B, 0xD48D },
 { 0x7610, 0xC720 },
 { 0x7615, 0xD558 },
 { 0x7627, 0xD559 },
 { 0x762F, 0xC871 },
 { 0x7630, 0xB77C },
 { 0x7633, 0xCD94 },
 { 0x7634, 0xC7A5 },
 { 0x7643, 0xB96D },
 { 0x7646, 0xB85C },
 { 0x7648, 0xD3D0 },
 { 0x7649, 0xB2E8 },
 { 0x7655, 0xC639 },
 { 0x7658, 0xB824 },
 { 0x765C, 0xC804 },
 { 0x7662, 0xC591 },
 { 0x7665, 0xC9D5 },
 { 0x7667, 0xB825 },
 { 0x7668, 0xACFD },
 { 0x766D, 0xC601 },
 { 0x766E, 0xC740 },
 { 0x766F, 0xAD6C },
 { 0x7671, 0xD0C4 },
 { 0x7680, 0xAE09 },
 { 0x7681, 0xC870 },
 { 0x7683, 0xBAA8 },
 { 0x768B, 0xACE0 },
 { 0x7696, 0xD658 },
 { 0x769A, 0xC560 },
 { 0x769B, 0xD6A8 },
 { 0x769C, 0xD638 },
 { 0x769E, 0xD638 },
 { 0x76A4, 0xD30C },
 { 0x76AC, 0xD559 },
 { 0x76B4, 0xC900 },
 { 0x76B6, 0xC0AC },
 { 0x76B8, 0xAD70 },
 { 0x76CC, 0xC644 },
 { 0x76CD, 0xD569 },
 { 0x76CE, 0xC559 },
 { 0x76D4, 0xD68C },
 { 0x76D9, 0xBCF4 },
 { 0x76E5, 0xAD00 },
 { 0x76EA, 0xD0D5 },
 { 0x76EC, 0xACE0 },
 { 0x76F1, 0xC6B0 },
 { 0x76F9, 0xC21C },
 { 0x76FB, 0xD61C },
 { 0x7707, 0xBB18 },
 { 0x770A, 0xBAA8 },
 { 0x770C, 0xD604 },
 { 0x771A, 0xC0DD },
 { 0x771B, 0xB9E4 },
 { 0x7725, 0xC790 },
 { 0x7726, 0xC790 },
 { 0x7728, 0xC7A1 },
 { 0x7734, 0xC21C },
 { 0x773B, 0xC591 },
 { 0x773E, 0xC911 },
 { 0x7746, 0xD658 },
 { 0x7747, 0xC81C },
 { 0x7752, 0xC12C },
 { 0x775A, 0xC560 },
 { 0x775F, 0xC218 },
 { 0x7760, 0xAD8C },
 { 0x7762, 0xC218 },
 { 0x7765, 0xBE44 },
 { 0x7768, 0xC608 },
 { 0x7777, 0xAC74 },
 { 0x777D, 0xADDC },
 { 0x778D, 0xC218 },
 { 0x778E, 0xD560 },
 { 0x7796, 0xC608 },
 { 0x7799, 0xB9C9 },
 { 0x77A0, 0xB2F9 },
 { 0x77A2, 0xBABD },
 { 0x77AA, 0xC9D5 },
 { 0x77B9, 0xC560 },
 { 0x77BD, 0xACE0 },
 { 0x77C7, 0xBABD },
 { 0x77C9, 0xBE48 },
 { 0x77CD, 0xD655 },
 { 0x77D9, 0xAC10 },
 { 0x77DA, 0xCD09 },
 { 0x77DE, 0xC728 },
 { 0x77E1, 0xD655 },
 { 0x77E7, 0xC2E0 },
 { 0x77F0, 0xC99D },
 { 0x77FA, 0xD0C1 },
 { 0x77FB, 0xACE8 },
 { 0x77FC, 0xAC15 },
 { 0x77FD, 0xC11D },
 { 0x780C, 0xCCB4 },
 { 0x780F, 0xBD84 },
 { 0x7811, 0xC544 },
 { 0x7822, 0xB77C },
 { 0x782D, 0xD3C4 },
 { 0x7830, 0xD33D },
 { 0x7843, 0xC8FC },
 { 0x785C, 0xAC31 },
 { 0x7860, 0xB791 },
 { 0x7867, 0xC6A9 },
 { 0x7868, 0xCC28 },
 { 0x786E, 0xD559 },
 { 0x788F, 0xC791 },
 { 0x7893, 0xB300 },
 { 0x789F, 0xC124 },
 { 0x78A4, 0xC601 },
 { 0x78A8, 0xC678 },
 { 0x78AD, 0xD0D5 },
 { 0x78BE, 0xB144 },
 { 0x78C8, 0xC678 },
 { 0x78CF, 0xB834 },
 { 0x78D1, 0xC560 },
 { 0x78D4, 0xCC45 },
 { 0x78D5, 0xAC1C },
 { 0x78DA, 0xC804 },
 { 0x78E0, 0xB1E8 },
 { 0x78E7, 0xC801 },
 { 0x78EA, 0xCD5C },
 { 0x78F4, 0xB4F1 },
 { 0x78F7, 0xB9B0 },
 { 0x78FA, 0xAD11 },
 { 0x78FC, 0xC7A1 },
 { 0x78FD, 0xAD50 },
 { 0x790C, 0xB8B0 },
 { 0x790F, 0xAE09 },
 { 0x7911, 0xB2F9 },
 { 0x7912, 0xC758 },
 { 0x7919, 0xC560 },
 { 0x7927, 0xB8B0 },
 { 0x792D, 0xD655 },
 { 0x7931, 0xB871 },
 { 0x7945, 0xC694 },
 { 0x7946, 0xCC9C },
 { 0x794A, 0xD33D },
 { 0x7953, 0xBD88 },
 { 0x7954, 0xBD80 },
 { 0x7955, 0xBE44 },
 { 0x795F, 0xC218 },
 { 0x7962, 0xB2C8 },
 { 0x7967, 0xC870 },
 { 0x797C, 0xAD00 },
 { 0x7980, 0xD488 },
 { 0x798A, 0xACC4 },
 { 0x798B, 0xC778 },
 { 0x7994, 0xC81C },
 { 0x799D, 0xC9C1 },
 { 0x79A8, 0xAE30 },
 { 0x79AB, 0xB2F4 },
 { 0x79B0, 0xB2C8 },
 { 0x79B4, 0xC57D },
 { 0x79BA, 0xC639 },
 { 0x79BB, 0xB9AC },
 { 0x79C6, 0xAC04 },
 { 0x79E2, 0xB839 },
 { 0x79E3, 0xB9D0 },
 { 0x79EA, 0xC9C0 },
 { 0x79EB, 0xCD9C },
 { 0x79EC, 0xAC70 },
 { 0x79F8, 0xAC08 },
 { 0x7A01, 0xACE0 },
 { 0x7A02, 0xB791 },
 { 0x7A0A, 0xC81C },
 { 0x7A0C, 0xB3C4 },
 { 0x7A1E, 0xACFC },
 { 0x7A27, 0xC124 },
 { 0x7A39, 0xC9C4 },
 { 0x7A45, 0xAC15 },
 { 0x7A60, 0xB18D },
 { 0x7A75, 0xC54C },
 { 0x7A7E, 0xC694 },
 { 0x7A80, 0xB454 },
 { 0x7A82, 0xB85C },
 { 0x7A83, 0xC808 },
 { 0x7A85, 0xC694 },
 { 0x7A86, 0xD3C4 },
 { 0x7A8A, 0xC640 },
 { 0x7A96, 0xAD50 },
 { 0x7A97, 0xCC3D },
 { 0x7A9B, 0xAD6C },
 { 0x7AA0, 0xACFC },
 { 0x7AA3, 0xC194 },
 { 0x7AAC, 0xC720 },
 { 0x7AB3, 0xC720 },
 { 0x7AB6, 0xAD6C },
 { 0x7AB9, 0xC624 },
 { 0x7ABB, 0xCC3D },
 { 0x7ABE, 0xAD00 },
 { 0x7ABF, 0xB96D },
 { 0x7AC8, 0xC870 },
 { 0x7AE2, 0xC0AC },
 { 0x7AE6, 0xC1A1 },
 { 0x7AF8, 0xACBD },
 { 0x7AFD, 0xC6B0 },
 { 0x7B06, 0xD30C },
 { 0x7B08, 0xAE09 },
 { 0x7B0A, 0xC870 },
 { 0x7B0C, 0xC544 },
 { 0x7B12, 0xAE08 },
 { 0x7B18, 0xC810 },
 { 0x7B25, 0xC0AC },
 { 0x7B27, 0xCC45 },
 { 0x7B28, 0xBD84 },
 { 0x7B2F, 0xB178 },
 { 0x7B33, 0xAC00 },
 { 0x7B35, 0xBC94 },
 { 0x7B45, 0xC120 },
 { 0x7B48, 0xAD04 },
 { 0x7B53, 0xACC4 },
 { 0x7B5F, 0xBD80 },
 { 0x7B65, 0xAC70 },
 { 0x7B66, 0xAD00 },
 { 0x7B67, 0xACAC },
 { 0x7B69, 0xD1B5 },
 { 0x7B71, 0xC18C },
 { 0x7B73, 0xC815 },
 { 0x7B8D, 0xACE0 },
 { 0x7B8E, 0xD638 },
 { 0x7B92, 0xCD94 },
 { 0x7B99, 0xBCF5 },
 { 0x7B9C, 0xACF5 },
 { 0x7BA0, 0xCD94 },
 { 0x7BBE, 0xC18C },
 { 0x7BCA, 0xD64D },
 { 0x7BCB, 0xD611 },
 { 0x7BCC, 0xD6C4 },
 { 0x7BD4, 0xC6B4 },
 { 0x7BD9, 0xACE0 },
 { 0x7BDB, 0xC57D },
 { 0x7BDD, 0xAD6C },
 { 0x7BE6, 0xBE44 },
 { 0x7BEA, 0xC9C0 },
 { 0x7BF3, 0xD544 },
 { 0x7BF7, 0xBD09 },
 { 0x7BFE, 0xBA78 },
 { 0x7C00, 0xCC45 },
 { 0x7C09, 0xCD94 },
 { 0x7C0B, 0xADA4 },
 { 0x7C0F, 0xB85D },
 { 0x7C1F, 0xC810 },
 { 0x7C20, 0xBCF4 },
 { 0x7C37, 0xCCA8 },
 { 0x7C38, 0xD30C },
 { 0x7C50, 0xB4F1 },
 { 0x7C52, 0xC8FC },
 { 0x7C54, 0xC218 },
 { 0x7C5B, 0xC804 },
 { 0x7C5C, 0xD0C1 },
 { 0x7C5F, 0xB8B0 },
 { 0x7C65, 0xC57D },
 { 0x7C67, 0xAC70 },
 { 0x7C69, 0xBCC0 },
 { 0x7C72, 0xC720 },
 { 0x7C7E, 0xC778 },
 { 0x7C81, 0xCC9C },
 { 0x7C8D, 0xBAA8 },
 { 0x7CA2, 0xC790 },
 { 0x7CA4, 0xC6D4 },
 { 0x7CA6, 0xB9B0 },
 { 0x7CA8, 0xBC31 },
 { 0x7CBC, 0xB9B0 },
 { 0x7CBD, 0xC885 },
 { 0x7CC5, 0xC720 },
 { 0x7CC6, 0xBA74 },
 { 0x7CD5, 0xACE0 },
 { 0x7CD7, 0xAD6C },
 { 0x7CD9, 0xC870 },
 { 0x7CDA, 0xC7A5 },
 { 0x7CDC, 0xBBF8 },
 { 0x7CDD, 0xC0BC },
 { 0x7CE2, 0xBAA8 },
 { 0x7CE5, 0xB098 },
 { 0x7CEF, 0xB098 },
 { 0x7CF2, 0xB824 },
 { 0x7CF4, 0xC801 },
 { 0x7CF6, 0xC870 },
 { 0x7CF8, 0xBA71 },
 { 0x7CFA, 0xADDC },
 { 0x7D03, 0xC21C },
 { 0x7D13, 0xC11C },
 { 0x7D15, 0xBE44 },
 { 0x7D1C, 0xC6B4 },
 { 0x7D31, 0xBD88 },
 { 0x7D32, 0xC124 },
 { 0x7D49, 0xC220 },
 { 0x7D4D, 0xC784 },
 { 0x7D56, 0xAD11 },
 { 0x7D59, 0xD658 },
 { 0x7D5B, 0xC870 },
 { 0x7D5C, 0xD608 },
 { 0x7D63, 0xBCD1 },
 { 0x7D69, 0xC870 },
 { 0x7D70, 0xC9C8 },
 { 0x7D7A, 0xCE58 },
 { 0x7D83, 0xCD08 },
 { 0x7D86, 0xACBD },
 { 0x7D98, 0xBD09 },
 { 0x7DA3, 0xAD8C },
 { 0x7DA5, 0xAE30 },
 { 0x7DA6, 0xAE30 },
 { 0x7DA7, 0xC900 },
 { 0x7DA9, 0xC6D0 },
 { 0x7DAB, 0xC120 },
 { 0x7DAE, 0xACC4 },
 { 0x7DB0, 0xAD00 },
 { 0x7DC4, 0xACE4 },
 { 0x7DCC, 0xC720 },
 { 0x7DCD, 0xBBFC },
 { 0x7DD7, 0xC0C1 },
 { 0x7DDC, 0xBA74 },
 { 0x7DE6, 0xC2DC },
 { 0x7DF6, 0xD3B8 },
 { 0x7DF9, 0xC81C },
 { 0x7E08, 0xC601 },
 { 0x7E0B, 0xCD94 },
 { 0x7E10, 0xCD94 },
 { 0x7E11, 0xACB8 },
 { 0x7E17, 0xCD5C },
 { 0x7E18, 0xACC4 },
 { 0x7E20, 0xACE1 },
 { 0x7E22, 0xB4F1 },
 { 0x7E32, 0xB958 },
 { 0x7E33, 0xACAC },
 { 0x7E35, 0xB9CC },
 { 0x7E39, 0xD45C },
 { 0x7E3B, 0xBBF8 },
 { 0x7E44, 0xC608 },
 { 0x7E45, 0xC18C },
 { 0x7E48, 0xAC15 },
 { 0x7E50, 0xC138 },
 { 0x7E56, 0xC0B0 },
 { 0x7E57, 0xB9B0 },
 { 0x7E59, 0xBC88 },
 { 0x7E5A, 0xB8CC },
 { 0x7E62, 0xADA4 },
 { 0x7E6F, 0xD604 },
 { 0x7E76, 0xC5B5 },
 { 0x7E7D, 0xBE48 },
 { 0x7E7E, 0xACAC },
 { 0x7E81, 0xD6C8 },
 { 0x7E87, 0xB8B0 },
 { 0x7E88, 0xD790 },
 { 0x7E8A, 0xAD11 },
 { 0x7E8D, 0xB958 },
 { 0x7E94, 0xC7AC },
 { 0x7F37, 0xC0AC },
 { 0x7F3E, 0xBCD1 },
 { 0x7F3F, 0xD56D },
 { 0x7F43, 0xC575 },
 { 0x7F44, 0xACBD },
 { 0x7F45, 0xD558 },
 { 0x7F47, 0xC900 },
 { 0x7F4B, 0xC639 },
 { 0x7F4D, 0xB8B0 },
 { 0x7F4E, 0xB2F4 },
 { 0x7F58, 0xBD80 },
 { 0x7F5D, 0xC800 },
 { 0x7F5F, 0xACE0 },
 { 0x7F60, 0xBBFC },
 { 0x7F61, 0xAC15 },
 { 0x7F63, 0xAD18 },
 { 0x7F66, 0xBD80 },
 { 0x7F68, 0xC5C4 },
 { 0x7F78, 0xBC8C },
 { 0x7F7C, 0xD544 },
 { 0x7F7D, 0xACC4 },
 { 0x7F7E, 0xC99D },
 { 0x7F82, 0xACAC },
 { 0x7F86, 0xBE44 },
 { 0x7F87, 0xAE30 },
 { 0x7F96, 0xACE0 },
 { 0x7F97, 0xAC15 },
 { 0x7F98, 0xC7A5 },
 { 0x7F9D, 0xC800 },
 { 0x7FAB, 0xAC15 },
 { 0x7FAF, 0xAC08 },
 { 0x7FB6, 0xC804 },
 { 0x7FD6, 0xD761 },
 { 0x7FDB, 0xC18C },
 { 0x7FE3, 0xC0BD },
 { 0x7FE4, 0xC2DC },
 { 0x7FE6, 0xC804 },
 { 0x7FEC, 0xD718 },
 { 0x7FEE, 0xD575 },
 { 0x7FF3, 0xC608 },
 { 0x7FFA, 0xACE0 },
 { 0x7FFB, 0xBC88 },
 { 0x7FFE, 0xD604 },
 { 0x8004, 0xBAA8 },
 { 0x8011, 0xB2E8 },
 { 0x8012, 0xB8B0 },
 { 0x8014, 0xC790 },
 { 0x8016, 0xCD08 },
 { 0x8019, 0xD30C },
 { 0x801E, 0xAC00 },
 { 0x8021, 0xC11C },
 { 0x8026, 0xC6B0 },
 { 0x8028, 0xB204 },
 { 0x802C, 0xB8E8 },
 { 0x8030, 0xC6B0 },
 { 0x8052, 0xAD04 },
 { 0x8055, 0xD638 },
 { 0x8071, 0xC624 },
 { 0x8075, 0xC678 },
 { 0x8076, 0xC12D },
 { 0x8093, 0xD669 },
 { 0x8098, 0xC8FC },
 { 0x809A, 0xB450 },
 { 0x80A6, 0xBD84 },
 { 0x80AB, 0xC21C },
 { 0x80AD, 0xB20C },
 { 0x80B5, 0xAE30 },
 { 0x80C6, 0xB2E8 },
 { 0x80CA, 0xAD6C },
 { 0x80D5, 0xBD80 },
 { 0x80D6, 0xBC18 },
 { 0x80D9, 0xC870 },
 { 0x80E0, 0xAC70 },
 { 0x80EF, 0xACE0 },
 { 0x80F7, 0xD749 },
 { 0x80FE, 0xC790 },
 { 0x8109, 0xB9E5 },
 { 0x8116, 0xBC1C },
 { 0x8124, 0xC2E0 },
 { 0x8127, 0xCD5C },
 { 0x8130, 0xB450 },
 { 0x813A, 0xC218 },
 { 0x8141, 0xBCC0 },
 { 0x8146, 0xC804 },
 { 0x814A, 0xC11D },
 { 0x8152, 0xAC70 },
 { 0x8153, 0xBE44 },
 { 0x8160, 0xC8FC },
 { 0x8164, 0xC554 },
 { 0x816D, 0xC545 },
 { 0x8174, 0xC720 },
 { 0x8176, 0xB2E8 },
 { 0x8177, 0xD53D },
 { 0x8182, 0xB824 },
 { 0x8183, 0xC62C },
 { 0x8184, 0xC218 },
 { 0x8186, 0xC18C },
 { 0x818B, 0xB8CC },
 { 0x819E, 0xC804 },
 { 0x81A9, 0xB2C8 },
 { 0x81B0, 0xBC88 },
 { 0x81B4, 0xBB34 },
 { 0x81B8, 0xC218 },
 { 0x81B9, 0xBD84 },
 { 0x81CA, 0xC870 },
 { 0x81CF, 0xBE48 },
 { 0x81D1, 0xB178 },
 { 0x81D9, 0xC5F0 },
 { 0x81DA, 0xB824 },
 { 0x81DD, 0xB77C },
 { 0x81E0, 0xB828 },
 { 0x81EC, 0xC5BC },
 { 0x81F6, 0xCC9C },
 { 0x8201, 0xC5EC },
 { 0x8202, 0xC6A9 },
 { 0x8203, 0xC11D },
 { 0x8210, 0xC9C0 },
 { 0x8216, 0xD3EC },
 { 0x8233, 0xCD95 },
 { 0x8238, 0xAC00 },
 { 0x823D, 0xBC29 },
 { 0x8251, 0xD3B8 },
 { 0x8258, 0xC18C },
 { 0x825A, 0xC870 },
 { 0x825F, 0xB3D9 },
 { 0x8268, 0xBABD },
 { 0x826A, 0xB85C },
 { 0x826B, 0xB85C },
 { 0x8274, 0xBD88 },
 { 0x828A, 0xCC9C },
 { 0x8294, 0xD6FC },
 { 0x8298, 0xBE44 },
 { 0x82A1, 0xAC80 },
 { 0x82A2, 0xC778 },
 { 0x82A3, 0xBD80 },
 { 0x82A7, 0xC11C },
 { 0x82A8, 0xAE09 },
 { 0x82AA, 0xAE30 },
 { 0x82AB, 0xC6D0 },
 { 0x82B0, 0xAE30 },
 { 0x82BE, 0xBE44 },
 { 0x82C5, 0xC608 },
 { 0x82D3, 0xB839 },
 { 0x82DC, 0xBAA9 },
 { 0x82E0, 0xBBFC },
 { 0x82E3, 0xAC70 },
 { 0x82EB, 0xC810 },
 { 0x82F3, 0xB3D9 },
 { 0x82F4, 0xC800 },
 { 0x82F9, 0xD3C9 },
 { 0x82FA, 0xB9E4 },
 { 0x82FB, 0xBD80 },
 { 0x82FF, 0xBBF8 },
 { 0x8300, 0xBD88 },
 { 0x8308, 0xC790 },
 { 0x831B, 0xAC04 },
 { 0x831C, 0xCC9C },
 { 0x831D, 0xCC44 },
 { 0x8323, 0xC624 },
 { 0x832A, 0xAD11 },
 { 0x8332, 0xC790 },
 { 0x8333, 0xAC15 },
 { 0x833C, 0xB3D9 },
 { 0x8343, 0xC804 },
 { 0x8350, 0xCC9C },
 { 0x8385, 0xB9AC },
 { 0x8392, 0xAC70 },
 { 0x8393, 0xB9E4 },
 { 0x839A, 0xC5F0 },
 { 0x839B, 0xC815 },
 { 0x839D, 0xC88C },
 { 0x839F, 0xD568 },
 { 0x83A0, 0xC720 },
 { 0x83A8, 0xB791 },
 { 0x83BE, 0xB9DD },
 { 0x83BF, 0xC790 },
 { 0x83C2, 0xC801 },
 { 0x83D1, 0xCE58 },
 { 0x83D4, 0xBCF5 },
 { 0x83D8, 0xC22D },
 { 0x83DF, 0xB3C4 },
 { 0x83E0, 0xD30C },
 { 0x83E1, 0xD568 },
 { 0x83F6, 0xBD09 },
 { 0x8406, 0xBE44 },
 { 0x8407, 0xC7A5 },
 { 0x840B, 0xCC98 },
 { 0x8435, 0xC640 },
 { 0x8439, 0xD3B8 },
 { 0x843C, 0xC545 },
 { 0x8446, 0xBCF4 },
 { 0x844A, 0xC554 },
 { 0x8451, 0xD48D },
 { 0x845A, 0xC2EC },
 { 0x8469, 0xD30C },
 { 0x846A, 0xACC4 },
 { 0x846D, 0xAC00 },
 { 0x8471, 0xCD1D },
 { 0x8473, 0xC704 },
 { 0x8477, 0xD6C8 },
 { 0x8482, 0xCCB4 },
 { 0x849F, 0xAD6C },
 { 0x84A8, 0xCC9C },
 { 0x84AD, 0xCD94 },
 { 0x84AF, 0xAD34 },
 { 0x84B4, 0xC0AD },
 { 0x84B9, 0xACB8 },
 { 0x84BA, 0xC9C8 },
 { 0x84C1, 0xC9C4 },
 { 0x84CA, 0xC639 },
 { 0x84D0, 0xC695 },
 { 0x84D3, 0xBC30 },
 { 0x84D6, 0xBE44 },
 { 0x84ED, 0xC554 },
 { 0x84EF, 0xCD1D },
 { 0x84F1, 0xD3C9 },
 { 0x84FD, 0xD544 },
 { 0x8500, 0xBD80 },
 { 0x8506, 0xB989 },
 { 0x8515, 0xCCB4 },
 { 0x851F, 0xC8FC },
 { 0x8526, 0xC870 },
 { 0x852A, 0xC810 },
 { 0x8546, 0xCC9C },
 { 0x854A, 0xC608 },
 { 0x854B, 0xC608 },
 { 0x8555, 0xC720 },
 { 0x8556, 0xAC70 },
 { 0x8558, 0xC694 },
 { 0x855E, 0xCCB4 },
 { 0x8561, 0xBD84 },
 { 0x8564, 0xC720 },
 { 0x856B, 0xB3D9 },
 { 0x857A, 0xC999 },
 { 0x857E, 0xB8B0 },
 { 0x8580, 0xC628 },
 { 0x8586, 0xC560 },
 { 0x8589, 0xC608 },
 { 0x858A, 0xACC4 },
 { 0x858C, 0xD5A5 },
 { 0x8597, 0xC6D0 },
 { 0x8599, 0xCE58 },
 { 0x859D, 0xB2F4 },
 { 0x85A4, 0xD574 },
 { 0x85A5, 0xCD09 },
 { 0x85D0, 0xBB18 },
 { 0x85F9, 0xC560 },
 { 0x85FD, 0xCE5C },
 { 0x8600, 0xD0C1 },
 { 0x8604, 0xAE30 },
 { 0x860B, 0xBE48 },
 { 0x8618, 0xC591 },
 { 0x8622, 0xB871 },
 { 0x8627, 0xAC70 },
 { 0x8629, 0xBC88 },
 { 0x862F, 0xD0D5 },
 { 0x863C, 0xBBF8 },
 { 0x8641, 0xAE30 },
 { 0x8653, 0xD6A8 },
 { 0x866B, 0xD6FC },
 { 0x866C, 0xADDC },
 { 0x866F, 0xADDC },
 { 0x8671, 0xC2AC },
 { 0x8678, 0xC790 },
 { 0x867A, 0xD6FC },
 { 0x867B, 0xB9F9 },
 { 0x868B, 0xC608 },
 { 0x868D, 0xBE44 },
 { 0x8695, 0xCC9C },
 { 0x86A8, 0xBD80 },
 { 0x86AA, 0xB450 },
 { 0x86AF, 0xAD6C },
 { 0x86B0, 0xC720 },
 { 0x86B1, 0xCC45 },
 { 0x86B4, 0xC720 },
 { 0x86C0, 0xC8FC },
 { 0x86C5, 0xC810 },
 { 0x86C6, 0xC800 },
 { 0x86C9, 0xB839 },
 { 0x86DE, 0xD65C },
 { 0x86E3, 0xAE38 },
 { 0x86E9, 0xACF5 },
 { 0x86EC, 0xACF5 },
 { 0x86F8, 0xC18C },
 { 0x86F9, 0xC6A9 },
 { 0x86FB, 0xC138 },
 { 0x8706, 0xD604 },
 { 0x8709, 0xBD80 },
 { 0x870A, 0xB9AC },
 { 0x870B, 0xB791 },
 { 0x8711, 0xB2E8 },
 { 0x871D, 0xAE30 },
 { 0x8721, 0xC0AC },
 { 0x8725, 0xC11D },
 { 0x8728, 0xC811 },
 { 0x8729, 0xC870 },
 { 0x8734, 0xCC99 },
 { 0x8735, 0xC5F0 },
 { 0x8737, 0xAD8C },
 { 0x873A, 0xC608 },
 { 0x873B, 0xCCAD },
 { 0x873F, 0xC6D0 },
 { 0x8740, 0xB3D9 },
 { 0x874C, 0xACFC },
 { 0x8759, 0xD3B8 },
 { 0x8760, 0xBCF5 },
 { 0x8764, 0xC720 },
 { 0x8765, 0xBAA8 },
 { 0x876E, 0xBCF5 },
 { 0x8783, 0xBC29 },
 { 0x878C, 0xBC18 },
 { 0x8798, 0xC758 },
 { 0x879E, 0xB9C8 },
 { 0x87A3, 0xB4F1 },
 { 0x87AD, 0xB9AC },
 { 0x87B4, 0xC9C4 },
 { 0x87BB, 0xB8E8 },
 { 0x87BD, 0xC885 },
 { 0x87C0, 0xC194 },
 { 0x87C7, 0xB9C8 },
 { 0x87CA, 0xBAA8 },
 { 0x87CB, 0xC2E4 },
 { 0x87D2, 0xB9DD },
 { 0x87DA, 0xD33D },
 { 0x87DB, 0xD33D },
 { 0x87E3, 0xAE30 },
 { 0x87F7, 0xB2F9 },
 { 0x880D, 0xAC08 },
 { 0x8811, 0xC601 },
 { 0x8815, 0xC5F0 },
 { 0x8821, 0xB824 },
 { 0x8832, 0xACAC },
 { 0x8839, 0xB450 },
 { 0x8844, 0xB275 },
 { 0x884A, 0xBA78 },
 { 0x884E, 0xAC04 },
 { 0x885E, 0xC704 },
 { 0x886E, 0xACE4 },
 { 0x887D, 0xC784 },
 { 0x889B, 0xC800 },
 { 0x889D, 0xBD80 },
 { 0x88A2, 0xBC88 },
 { 0x88AA, 0xAC70 },
 { 0x88C0, 0xC778 },
 { 0x88C5, 0xC7A5 },
 { 0x88CA, 0xB1E8 },
 { 0x88CD, 0xACE4 },
 { 0x88D2, 0xBD80 },
 { 0x88D8, 0xAD6C },
 { 0x88EF, 0xC8FC },
 { 0x88F1, 0xD45C },
 { 0x88F9, 0xACFC },
 { 0x88FE, 0xAC70 },
 { 0x8904, 0xCC98 },
 { 0x890A, 0xD3B8 },
 { 0x890C, 0xACE4 },
 { 0x8915, 0xC720 },
 { 0x891A, 0xC800 },
 { 0x8927, 0xACBD },
 { 0x892B, 0xCE58 },
 { 0x8930, 0xAC74 },
 { 0x893A, 0xCCA9 },
 { 0x894D, 0xC7A1 },
 { 0x8956, 0xC624 },
 { 0x8958, 0xAD34 },
 { 0x895C, 0xCCA8 },
 { 0x895E, 0xBCBD },
 { 0x8960, 0xB2F9 },
 { 0x896D, 0xD790 },
 { 0x896F, 0xCE5C },
 { 0x8974, 0xB780 },
 { 0x8988, 0xD575 },
 { 0x8989, 0xAE30 },
 { 0x8997, 0xC0AC },
 { 0x8998, 0xC810 },
 { 0x89A8, 0xC545 },
 { 0x89AC, 0xAE30 },
 { 0x89AF, 0xAD6C },
 { 0x89B7, 0xCC98 },
 { 0x89BF, 0xC801 },
 { 0x89D4, 0xADFC },
 { 0x89D6, 0xACB0 },
 { 0x89DA, 0xACE0 },
 { 0x89DC, 0xC790 },
 { 0x89DD, 0xC800 },
 { 0x89E5, 0xAD49 },
 { 0x89F1, 0xD544 },
 { 0x89F3, 0xACE1 },
 { 0x8A07, 0xAD49 },
 { 0x8A09, 0xBC94 },
 { 0x8A10, 0xC54C },
 { 0x8A11, 0xC774 },
 { 0x8A14, 0xC740 },
 { 0x8A15, 0xC0B0 },
 { 0x8A22, 0xD754 },
 { 0x8A3C, 0xC815 },
 { 0x8A3E, 0xC790 },
 { 0x8A41, 0xACE0 },
 { 0x8A46, 0xC800 },
 { 0x8A4B, 0xC8FC },
 { 0x8A51, 0xC774 },
 { 0x8A56, 0xD53C },
 { 0x8A57, 0xD615 },
 { 0x8A58, 0xAD74 },
 { 0x8A61, 0xD6C4 },
 { 0x8A6B, 0xD0C0 },
 { 0x8A6C, 0xAD6C },
 { 0x8A7C, 0xD68C },
 { 0x8A7F, 0xAD18 },
 { 0x8A82, 0xC870 },
 { 0x8A84, 0xB8B0 },
 { 0x8A86, 0xAD11 },
 { 0x8A91, 0xAD11 },
 { 0x8A9A, 0xCD08 },
 { 0x8AA2, 0xD604 },
 { 0x8AA7, 0xD3EC },
 { 0x8AC9, 0xC704 },
 { 0x8ADE, 0xD3B8 },
 { 0x8ADF, 0xC2DC },
 { 0x8AE0, 0xD6E4 },
 { 0x8AE4, 0xC545 },
 { 0x8AF0, 0xC2DC },
 { 0x8AF3, 0xC554 },
 { 0x8AF4, 0xD568 },
 { 0x8AFC, 0xD6E4 },
 { 0x8B07, 0xAC74 },
 { 0x8B0C, 0xAC00 },
 { 0x8B11, 0xD61C },
 { 0x8B1C, 0xC6D0 },
 { 0x8B23, 0xC6B0 },
 { 0x8B26, 0xACBD },
 { 0x8B37, 0xC624 },
 { 0x8B3E, 0xB9CC },
 { 0x8B43, 0xD6C4 },
 { 0x8B4C, 0xC640 },
 { 0x8B54, 0xC120 },
 { 0x8B56, 0xCC38 },
 { 0x8B59, 0xCD08 },
 { 0x8B5F, 0xC870 },
 { 0x8B6B, 0xC12C },
 { 0x8B71, 0xC120 },
 { 0x8B7F, 0xD61C },
 { 0x8B89, 0xC720 },
 { 0x8B8C, 0xC5F0 },
 { 0x8B8E, 0xC218 },
 { 0x8B9C, 0xB2F9 },
 { 0x8B9E, 0xC5B8 },
 { 0x8C47, 0xAC15 },
 { 0x8C49, 0xC2DC },
 { 0x8C4E, 0xC218 },
 { 0x8C56, 0xCD95 },
 { 0x8C62, 0xD658 },
 { 0x8C68, 0xD76C },
 { 0x8C6C, 0xC800 },
 { 0x8C73, 0xBE48 },
 { 0x8C78, 0xCE58 },
 { 0x8C83, 0xB9E5 },
 { 0x8C8D, 0xB9AC },
 { 0x8C93, 0xBB18 },
 { 0x8C94, 0xBE44 },
 { 0x8C98, 0xB9E5 },
 { 0x8CB2, 0xC790 },
 { 0x8CBA, 0xD669 },
 { 0x8CC6, 0xBCC0 },
 { 0x8CD9, 0xC8FC },
 { 0x8CE1, 0xAC31 },
 { 0x8CEC, 0xC7A5 },
 { 0x8D0B, 0xC548 },
 { 0x8D0F, 0xC601 },
 { 0x8D10, 0xC2E0 },
 { 0x8D14, 0xBE44 },
 { 0x8D1B, 0xACF5 },
 { 0x8D67, 0xB09C },
 { 0x8D6D, 0xC790 },
 { 0x8D76, 0xAC04 },
 { 0x8D81, 0xC9C4 },
 { 0x8D8E, 0xC8FC },
 { 0x8D90, 0xD608 },
 { 0x8DA0, 0xCD08 },
 { 0x8DAB, 0xAD50 },
 { 0x8DB2, 0xCC2C },
 { 0x8DC2, 0xAE30 },
 { 0x8DCE, 0xD0C0 },
 { 0x8DD5, 0xC811 },
 { 0x8DD6, 0xCC99 },
 { 0x8DD7, 0xBD80 },
 { 0x8DDF, 0xADFC },
 { 0x8DEA, 0xADA4 },
 { 0x8DEB, 0xACF5 },
 { 0x8DEC, 0xADDC },
 { 0x8DF1, 0xCE58 },
 { 0x8DFC, 0xAD6D },
 { 0x8E06, 0xC900 },
 { 0x8E08, 0xC18C },
 { 0x8E09, 0xB7C9 },
 { 0x8E14, 0xD0C1 },
 { 0x8E1D, 0xACFC },
 { 0x8E1F, 0xC9C0 },
 { 0x8E34, 0xC6A9 },
 { 0x8E36, 0xC81C },
 { 0x8E3D, 0xC6B0 },
 { 0x8E40, 0xC811 },
 { 0x8E4C, 0xCC3D },
 { 0x8E4F, 0xC81C },
 { 0x8E55, 0xD544 },
 { 0x8E5C, 0xCD95 },
 { 0x8E63, 0xB9CC },
 { 0x8E64, 0xC885 },
 { 0x8E72, 0xC900 },
 { 0x8E7B, 0xAD50 },
 { 0x8E85, 0xCD09 },
 { 0x8E89, 0xB3C8 },
 { 0x8E8B, 0xC81C },
 { 0x8E90, 0xB835 },
 { 0x8E91, 0xCC99 },
 { 0x8E93, 0xC9C0 },
 { 0x8E94, 0xC804 },
 { 0x8E99, 0xB9B0 },
 { 0x8EA1, 0xC12D },
 { 0x8EA9, 0xACFD },
 { 0x8EB1, 0xD0C0 },
 { 0x8EB3, 0xAD81 },
 { 0x8EC6, 0xCCB4 },
 { 0x8EE8, 0xB839 },
 { 0x8EF9, 0xC9C0 },
 { 0x8EFA, 0xCD08 },
 { 0x8EFC, 0xC77C },
 { 0x8F07, 0xC804 },
 { 0x8F08, 0xC8FC },
 { 0x8F28, 0xAD00 },
 { 0x8F30, 0xC591 },
 { 0x8F40, 0xC628 },
 { 0x8F42, 0xACE1 },
 { 0x8F46, 0xB85D },
 { 0x8F52, 0xBD84 },
 { 0x8F54, 0xB9B0 },
 { 0x8F57, 0xAC10 },
 { 0x8F58, 0xD658 },
 { 0x8F5E, 0xD568 },
 { 0x8F61, 0xBE44 },
 { 0x8F63, 0xB825 },
 { 0x8F64, 0xB85C },
 { 0x8F9F, 0xD53C },
 { 0x8FA2, 0xB784 },
 { 0x8FAE, 0xBCC0 },
 { 0x8FBB, 0xC2ED },
 { 0x8FBF, 0xCC9C },
 { 0x8FC0, 0xAC04 },
 { 0x8FC3, 0xC624 },
 { 0x8FC4, 0xD758 },
 { 0x8FCD, 0xB454 },
 { 0x8FD3, 0xC544 },
 { 0x8FD5, 0xC624 },
 { 0x8FE2, 0xCD08 },
 { 0x8FE4, 0xC774 },
 { 0x8FE5, 0xD615 },
 { 0x8FE8, 0xD0DC },
 { 0x8FEC, 0xC655 },
 { 0x8FF4, 0xD68C },
 { 0x8FF8, 0xBCD1 },
 { 0x8FFA, 0xB0B4 },
 { 0x900C, 0xC720 },
 { 0x9016, 0xC801 },
 { 0x9034, 0xD0C1 },
 { 0x9036, 0xC704 },
 { 0x9049, 0xC815 },
 { 0x904F, 0xC54C },
 { 0x9052, 0xC8FC },
 { 0x9058, 0xAD6C },
 { 0x905B, 0xB958 },
 { 0x9068, 0xC624 },
 { 0x906C, 0xC18D },
 { 0x9076, 0xC694 },
 { 0x9085, 0xC804 },
 { 0x908C, 0xB824 },
 { 0x9090, 0xB9AC },
 { 0x90A0, 0xBE48 },
 { 0x90B3, 0xBE44 },
 { 0x90BD, 0xADDC },
 { 0x90C3, 0xD569 },
 { 0x90C4, 0xADF9 },
 { 0x90C5, 0xC9C8 },
 { 0x90DC, 0xACE0 },
 { 0x90DD, 0xD559 },
 { 0x90E2, 0xC601 },
 { 0x90EA, 0xCC98 },
 { 0x90EB, 0xBE44 },
 { 0x90EF, 0xB2F4 },
 { 0x90F4, 0xCE68 },
 { 0x911E, 0xC740 },
 { 0x9122, 0xC5B8 },
 { 0x9123, 0xC7A5 },
 { 0x9130, 0xB9B0 },
 { 0x9131, 0xD30C },
 { 0x9134, 0xC5C5 },
 { 0x913D, 0xC804 },
 { 0x9148, 0xB825 },
 { 0x9156, 0xD0D0 },
 { 0x9157, 0xD6C4 },
 { 0x9163, 0xAC10 },
 { 0x9164, 0xACE0 },
 { 0x9165, 0xC18C },
 { 0x9172, 0xC815 },
 { 0x9174, 0xB3C4 },
 { 0x9176, 0xB9E4 },
 { 0x9179, 0xB8B0 },
 { 0x9183, 0xC5C4 },
 { 0x9190, 0xD638 },
 { 0x919E, 0xC628 },
 { 0x91A2, 0xD574 },
 { 0x91AA, 0xB8CC },
 { 0x91B2, 0xB18D },
 { 0x91C1, 0xD754 },
 { 0x91C3, 0xC2DC },
 { 0x91C5, 0xC5FC },
 { 0x91E4, 0xC0BC },
 { 0x91E6, 0xAD6C },
 { 0x91ED, 0xACF5 },
 { 0x91FF, 0xADFC },
 { 0x9207, 0xBD80 },
 { 0x9214, 0xCD08 },
 { 0x921C, 0xD6A1 },
 { 0x921F, 0xC870 },
 { 0x9235, 0xBCD1 },
 { 0x9237, 0xACE0 },
 { 0x9238, 0xBC1C },
 { 0x9241, 0xC9C4 },
 { 0x924B, 0xD3EC },
 { 0x924E, 0xC0DD },
 { 0x924F, 0xC11C },
 { 0x9250, 0xC11D },
 { 0x9252, 0xC8FC },
 { 0x9277, 0xD64D },
 { 0x9278, 0xAD50 },
 { 0x927C, 0xBCD1 },
 { 0x927E, 0xBAA8 },
 { 0x928B, 0xC784 },
 { 0x9295, 0xCCA0 },
 { 0x9299, 0xACFC },
 { 0x929A, 0xC870 },
 { 0x929B, 0xC12C },
 { 0x92C6, 0xC724 },
 { 0x92D1, 0xC804 },
 { 0x92D7, 0xD604 },
 { 0x92DF, 0xCE68 },
 { 0x92E5, 0xC815 },
 { 0x92F2, 0xBCD1 },
 { 0x92F9, 0xCC3D },
 { 0x92FA, 0xC6D0 },
 { 0x92FD, 0xC870 },
 { 0x9306, 0xCC3D },
 { 0x930F, 0xC544 },
 { 0x9315, 0xACE4 },
 { 0x9319, 0xCE58 },
 { 0x9323, 0xCCA0 },
 { 0x9327, 0xAD00 },
 { 0x932A, 0xC804 },
 { 0x932C, 0xB3D9 },
 { 0x9349, 0xC2DC },
 { 0x9357, 0xC81C },
 { 0x935C, 0xD558 },
 { 0x935D, 0xC6B0 },
 { 0x9364, 0xC0BD },
 { 0x936B, 0xCD08 },
 { 0x936C, 0xCD08 },
 { 0x9370, 0xD658 },
 { 0x938A, 0xBC29 },
 { 0x9397, 0xC7C1 },
 { 0x939B, 0xBC15 },
 { 0x939F, 0xC0C1 },
 { 0x93A1, 0xC790 },
 { 0x93A4, 0xD669 },
 { 0x93C7, 0xC120 },
 { 0x93C8, 0xB828 },
 { 0x93CA, 0xC624 },
 { 0x93CB, 0xB9CC },
 { 0x93CC, 0xB9C9 },
 { 0x93D6, 0xC624 },
 { 0x93D7, 0xAC31 },
 { 0x93D8, 0xC7A5 },
 { 0x93DC, 0xB2F9 },
 { 0x93DD, 0xB9CC },
 { 0x93DF, 0xC0B0 },
 { 0x93E2, 0xD45C },
 { 0x93E6, 0xCD1D },
 { 0x93E8, 0xCC38 },
 { 0x93FB, 0xB9B0 },
 { 0x9403, 0xB1E8 },
 { 0x940F, 0xC900 },
 { 0x9410, 0xB8CC },
 { 0x9413, 0xB300 },
 { 0x9414, 0xC2EC },
 { 0x9419, 0xB4F1 },
 { 0x942A, 0xB85C },
 { 0x9436, 0xD658 },
 { 0x943A, 0xB2F9 },
 { 0x9442, 0xD6C8 },
 { 0x944A, 0xD655 },
 { 0x944C, 0xBE48 },
 { 0x9455, 0xC9C8 },
 { 0x945E, 0xB78D },
 { 0x9460, 0xC0AD },
 { 0x9462, 0xB824 },
 { 0x9463, 0xD45C },
 { 0x946A, 0xB85C },
 { 0x9470, 0xC57D },
 { 0x9471, 0xCC38 },
 { 0x9472, 0xC591 },
 { 0x9475, 0xAD00 },
 { 0x9477, 0xC12D },
 { 0x947C, 0xB77C },
 { 0x947E, 0xB780 },
 { 0x9588, 0xD55C },
 { 0x958A, 0xC0B0 },
 { 0x958E, 0xAD49 },
 { 0x959F, 0xBE44 },
 { 0x95A9, 0xBBFC },
 { 0x95AB, 0xACE4 },
 { 0x95AC, 0xB791 },
 { 0x95B6, 0xCC3D },
 { 0x95B9, 0xC5C4 },
 { 0x95BD, 0xD63C },
 { 0x95BE, 0xC5ED },
 { 0x95C3, 0xACA9 },
 { 0x95C8, 0xC704 },
 { 0x95CB, 0xACB0 },
 { 0x95CC, 0xB780 },
 { 0x95CD, 0xB3C4 },
 { 0x95D3, 0xAC1C },
 { 0x95DA, 0xADDC },
 { 0x95DE, 0xD568 },
 { 0x95E0, 0xADA4 },
 { 0x95E5, 0xB2EC },
 { 0x9624, 0xCE58 },
 { 0x9628, 0xC561 },
 { 0x962F, 0xC9C0 },
 { 0x9658, 0xD615 },
 { 0x9661, 0xB450 },
 { 0x966C, 0xCD94 },
 { 0x9672, 0xC218 },
 { 0x9674, 0xBE44 },
 { 0x967C, 0xC800 },
 { 0x9684, 0xC81C },
 { 0x9691, 0xAE30 },
 { 0x9697, 0xC678 },
 { 0x96A4, 0xD1F4 },
 { 0x96A9, 0xC624 },
 { 0x96B0, 0xC2B5 },
 { 0x96B3, 0xD734 },
 { 0x96B4, 0xB871 },
 { 0x96B6, 0xC774 },
 { 0x96B8, 0xC608 },
 { 0x96B9, 0xCD94 },
 { 0x96BC, 0xC900 },
 { 0x96DA, 0xAD00 },
 { 0x96DD, 0xC639 },
 { 0x96DE, 0xACC4 },
 { 0x96FA, 0xBABD },
 { 0x9704, 0xC18C },
 { 0x9705, 0xC0BD },
 { 0x9708, 0xD328 },
 { 0x9709, 0xB9E4 },
 { 0x970C, 0xC8FC },
 { 0x970D, 0xACFD },
 { 0x970E, 0xC0BD },
 { 0x970F, 0xBE44 },
 { 0x9714, 0xC8FC },
 { 0x9723, 0xC6B4 },
 { 0x9726, 0xBE48 },
 { 0x972A, 0xC74C },
 { 0x9738, 0xD328 },
 { 0x9746, 0xCCB4 },
 { 0x9749, 0xC560 },
 { 0x974C, 0xBCF4 },
 { 0x975A, 0xC815 },
 { 0x975B, 0xC804 },
 { 0x975D, 0xCC9C },
 { 0x9760, 0xACE0 },
 { 0x9766, 0xC804 },
 { 0x9768, 0xC5FD },
 { 0x9773, 0xADFC },
 { 0x9775, 0xB274 },
 { 0x977C, 0xB2EC },
 { 0x9780, 0xB3C4 },
 { 0x9781, 0xD53C },
 { 0x9784, 0xD3EC },
 { 0x9785, 0xC559 },
 { 0x9798, 0xCD08 },
 { 0x97A3, 0xC720 },
 { 0x97A6, 0xCD94 },
 { 0x97AC, 0xAC74 },
 { 0x97B1, 0xB3C4 },
 { 0x97B4, 0xBE44 },
 { 0x97B8, 0xD544 },
 { 0x97B9, 0xACFD },
 { 0x97C1, 0xAC15 },
 { 0x97C3, 0xB2EC },
 { 0x97CD, 0xBD88 },
 { 0x97D0, 0xAC11 },
 { 0x97D9, 0xC704 },
 { 0x97DD, 0xAD6C },
 { 0x97DE, 0xC628 },
 { 0x97E0, 0xD544 },
 { 0x97E1, 0xC704 },
 { 0x97ED, 0xAD6C },
 { 0x97EE, 0xAD6C },
 { 0x97F1, 0xC12C },
 { 0x97F2, 0xC81C },
 { 0x97F5, 0xC6B4 },
 { 0x97FA, 0xC601 },
 { 0x980D, 0xADDC },
 { 0x980E, 0xAE30 },
 { 0x980F, 0xD56D },
 { 0x9816, 0xBC18 },
 { 0x981E, 0xC54C },
 { 0x9821, 0xD790 },
 { 0x9823, 0xC2E0 },
 { 0x9826, 0xD574 },
 { 0x982B, 0xBD80 },
 { 0x982E, 0xD68C },
 { 0x9837, 0xC554 },
 { 0x983F, 0xC790 },
 { 0x9842, 0xB8B0 },
 { 0x9847, 0xCDCC },
 { 0x9848, 0xACBD },
 { 0x984B, 0xC2DC },
 { 0x9852, 0xC639 },
 { 0x9853, 0xC804 },
 { 0x9859, 0xC0C1 },
 { 0x9866, 0xCD08 },
 { 0x9870, 0xBE48 },
 { 0x9871, 0xB85C },
 { 0x9873, 0xC12D },
 { 0x9874, 0xAD00 },
 { 0x98AD, 0xC810 },
 { 0x98B2, 0xB82C },
 { 0x98B6, 0xAD6C },
 { 0x98BA, 0xC591 },
 { 0x98BC, 0xC218 },
 { 0x98BF, 0xBC94 },
 { 0x98C2, 0xB8CC },
 { 0x98C9, 0xB8CC },
 { 0x98CB, 0xC2E4 },
 { 0x98E6, 0xC804 },
 { 0x98E7, 0xC190 },
 { 0x98EA, 0xC784 },
 { 0x98EB, 0xC5B4 },
 { 0x98F1, 0xC190 },
 { 0x98FB, 0xCCA0 },
 { 0x9912, 0xB1CC },
 { 0x9914, 0xD3EC },
 { 0x9915, 0xC900 },
 { 0x991A, 0xD6A8 },
 { 0x9926, 0xC7A5 },
 { 0x9927, 0xC704 },
 { 0x992A, 0xB09C },
 { 0x992C, 0xD638 },
 { 0x992E, 0xCCA0 },
 { 0x9930, 0xC804 },
 { 0x9931, 0xD6C4 },
 { 0x9933, 0xB2F9 },
 { 0x9939, 0xB2F9 },
 { 0x993C, 0xD76C },
 { 0x993D, 0xADA4 },
 { 0x9942, 0xC628 },
 { 0x9948, 0xC218 },
 { 0x9950, 0xC758 },
 { 0x9955, 0xB3C4 },
 { 0x995C, 0xC5FC },
 { 0x995E, 0xCC38 },
 { 0x9997, 0xADDC },
 { 0x9998, 0xAD35 },
 { 0x99A3, 0xC554 },
 { 0x99B0, 0xC801 },
 { 0x99C4, 0xD0DC },
 { 0x99C8, 0xAD6C },
 { 0x99C9, 0xACBD },
 { 0x99CF, 0xAC70 },
 { 0x99D8, 0xD0DC },
 { 0x99DB, 0xC0AC },
 { 0x99DC, 0xD544 },
 { 0x99DE, 0xD0C0 },
 { 0x99E2, 0xBCC0 },
 { 0x99E8, 0xC21C },
 { 0x99EA, 0xC2E0 },
 { 0x99EE, 0xBC15 },
 { 0x99F8, 0xCE68 },
 { 0x99FB, 0xD55C },
 { 0x99FD, 0xD604 },
 { 0x9A02, 0xC131 },
 { 0x9A03, 0xC560 },
 { 0x9A04, 0xB85D },
 { 0x9A05, 0xCD94 },
 { 0x9A0B, 0xB798 },
 { 0x9A0D, 0xACFC },
 { 0x9A11, 0xBE44 },
 { 0x9A16, 0xBB34 },
 { 0x9A18, 0xCD1D },
 { 0x9A1B, 0xBE44 },
 { 0x9A2D, 0xC990 },
 { 0x9A35, 0xC6D0 },
 { 0x9A38, 0xC120 },
 { 0x9A3E, 0xB77C },
 { 0x9A41, 0xC624 },
 { 0x9A42, 0xCC38 },
 { 0x9A44, 0xCD1D },
 { 0x9A4A, 0xD654 },
 { 0x9A4C, 0xC219 },
 { 0x9A4E, 0xB9B0 },
 { 0x9A4F, 0xC794 },
 { 0x9A52, 0xD0C4 },
 { 0x9A58, 0xB77C },
 { 0x9A64, 0xC591 },
 { 0x9AAA, 0xC704 },
 { 0x9AB0, 0xD22C },
 { 0x9AB9, 0xAD50 },
 { 0x9ABC, 0xACA9 },
 { 0x9ABF, 0xBCC0 },
 { 0x9AC0, 0xBE44 },
 { 0x9AC6, 0xBC15 },
 { 0x9ACF, 0xB8E8 },
 { 0x9AD1, 0xCD09 },
 { 0x9AD6, 0xAD00 },
 { 0x9AD7, 0xB85C },
 { 0x9ADF, 0xD45C },
 { 0x9AE1, 0xACE4 },
 { 0x9AE3, 0xBC29 },
 { 0x9AE6, 0xBAA8 },
 { 0x9AEB, 0xCD08 },
 { 0x9AED, 0xC790 },
 { 0x9AF0, 0xCCB4 },
 { 0x9AF2, 0xD53C },
 { 0x9AF4, 0xBD88 },
 { 0x9AF9, 0xD734 },
 { 0x9AFA, 0xAD04 },
 { 0x9AFB, 0xACC4 },
 { 0x9AFD, 0xC88C },
 { 0x9B02, 0xBE48 },
 { 0x9B05, 0xBD95 },
 { 0x9B06, 0xC1A1 },
 { 0x9B0A, 0xC21C },
 { 0x9B0B, 0xC804 },
 { 0x9B0D, 0xD638 },
 { 0x9B10, 0xAE30 },
 { 0x9B12, 0xC9C4 },
 { 0x9B16, 0xC0BC },
 { 0x9B18, 0xB9CC },
 { 0x9B19, 0xC2B9 },
 { 0x9B1F, 0xD658 },
 { 0x9B22, 0xBE48 },
 { 0x9B23, 0xB835 },
 { 0x9B28, 0xD64D },
 { 0x9B29, 0xD601 },
 { 0x9B2E, 0xAD6C },
 { 0x9B2F, 0xCC3D },
 { 0x9B32, 0xACA9 },
 { 0x9B3A, 0xC0C1 },
 { 0x9B48, 0xC18C },
 { 0x9B4B, 0xCD94 },
 { 0x9B4D, 0xB9DD },
 { 0x9B4E, 0xB7C9 },
 { 0x9B51, 0xB9AC },
 { 0x9B58, 0xC5FC },
 { 0x9B66, 0xC0AC },
 { 0x9B74, 0xBC29 },
 { 0x9B80, 0xD0C0 },
 { 0x9B83, 0xD3C9 },
 { 0x9B92, 0xBD80 },
 { 0x9B93, 0xC790 },
 { 0x9B97, 0xB3D9 },
 { 0x9BA7, 0xC81C },
 { 0x9BA8, 0xC9C0 },
 { 0x9BAA, 0xC720 },
 { 0x9BAD, 0xD574 },
 { 0x9BB9, 0xC18C },
 { 0x9BC1, 0xACBD },
 { 0x9BC6, 0xD3EC },
 { 0x9BCA, 0xC0AC },
 { 0x9BD4, 0xCE58 },
 { 0x9BDB, 0xC870 },
 { 0x9BE2, 0xC608 },
 { 0x9BF7, 0xC81C },
 { 0x9C08, 0xC811 },
 { 0x9C0A, 0xB828 },
 { 0x9C0C, 0xCD94 },
 { 0x9C13, 0xC0C8 },
 { 0x9C24, 0xC0AC },
 { 0x9C2D, 0xAE30 },
 { 0x9C2E, 0xC628 },
 { 0x9C2F, 0xC57D },
 { 0x9C31, 0xB828 },
 { 0x9C35, 0xBBFC },
 { 0x9C39, 0xACAC },
 { 0x9C3A, 0xC18C },
 { 0x9C3E, 0xD45C },
 { 0x9C4F, 0xC2EC },
 { 0x9C52, 0xC900 },
 { 0x9C53, 0xC120 },
 { 0x9C60, 0xD68C },
 { 0x9C63, 0xC804 },
 { 0x9C67, 0xB840 },
 { 0x9C78, 0xB85C },
 { 0x9C7B, 0xC120 },
 { 0x9CE6, 0xC744 },
 { 0x9D03, 0xACA9 },
 { 0x9D06, 0xC9D0 },
 { 0x9D07, 0xBCF4 },
 { 0x9D08, 0xC548 },
 { 0x9D0C, 0xBD09 },
 { 0x9D12, 0xB839 },
 { 0x9D15, 0xD0C0 },
 { 0x9D18, 0xBCC0 },
 { 0x9D19, 0xCE58 },
 { 0x9D1F, 0xCE58 },
 { 0x9D23, 0xACE0 },
 { 0x9D30, 0xAD04 },
 { 0x9D36, 0xC54C },
 { 0x9D41, 0xAD50 },
 { 0x9D42, 0xD734 },
 { 0x9D44, 0xCE58 },
 { 0x9D53, 0xBC1C },
 { 0x9D54, 0xC900 },
 { 0x9D5E, 0xC544 },
 { 0x9D69, 0xBCF5 },
 { 0x9D6F, 0xD544 },
 { 0x9D70, 0xC870 },
 { 0x9D77, 0xC6D0 },
 { 0x9D7B, 0xCD94 },
 { 0x9D7E, 0xACE4 },
 { 0x9D84, 0xCCAD },
 { 0x9D89, 0xC21C },
 { 0x9D96, 0xCD94 },
 { 0x9D9A, 0xC545 },
 { 0x9DA1, 0xAC08 },
 { 0x9DA4, 0xACE4 },
 { 0x9DAC, 0xCC3D },
 { 0x9DB5, 0xCD94 },
 { 0x9DB8, 0xC57D },
 { 0x9DB9, 0xB958 },
 { 0x9DBB, 0xACE8 },
 { 0x9DBF, 0xC790 },
 { 0x9DC1, 0xC775 },
 { 0x9DC2, 0xC694 },
 { 0x9DC7, 0xAD6C },
 { 0x9DD3, 0xC790 },
 { 0x9DD6, 0xC608 },
 { 0x9DD9, 0xC9C0 },
 { 0x9DE6, 0xCD08 },
 { 0x9DE9, 0xBCC4 },
 { 0x9DEA, 0xC575 },
 { 0x9DEB, 0xC219 },
 { 0x9DF0, 0xC5F0 },
 { 0x9DF1, 0xACE0 },
 { 0x9DF3, 0xD55C },
 { 0x9DFD, 0xD559 },
 { 0x9DFF, 0xBCBD },
 { 0x9E07, 0xC804 },
 { 0x9E0F, 0xBABD },
 { 0x9E15, 0xB85C },
 { 0x9E1B, 0xAD00 },
 { 0x9E1C, 0xAD6C },
 { 0x9E77, 0xB839 },
 { 0x9E7B, 0xAC10 },
 { 0x9E80, 0xC6B0 },
 { 0x9E82, 0xADA4 },
 { 0x9E84, 0xCD94 },
 { 0x9E8B, 0xBBF8 },
 { 0x9E8C, 0xC6B0 },
 { 0x9E8F, 0xADE0 },
 { 0x9E91, 0xC608 },
 { 0x9E9E, 0xC7A5 },
 { 0x9EA4, 0xCD94 },
 { 0x9EA9, 0xBD80 },
 { 0x9EAA, 0xBA74 },
 { 0x9EAF, 0xAD6D },
 { 0x9EBD, 0xB9C8 },
 { 0x9EBF, 0xB9C8 },
 { 0x9EC5, 0xAE08 },
 { 0x9ECC, 0xD6A1 },
 { 0x9ECF, 0xC810 },
 { 0x9ED0, 0xB9AC },
 { 0x9ED6, 0xAE30 },
 { 0x9EDA, 0xACB8 },
 { 0x9EDD, 0xC720 },
 { 0x9EE0, 0xD790 },
 { 0x9EE5, 0xACBD },
 { 0x9EEE, 0xB2F4 },
 { 0x9EEF, 0xC554 },
 { 0x9EF2, 0xCC38 },
 { 0x9EF5, 0xB2F4 },
 { 0x9EF6, 0xC5FC },
 { 0x9EF7, 0xB3C5 },
 { 0x9EFA, 0xBD84 },
 { 0x9EFB, 0xBD88 },
 { 0x9EFC, 0xBCF4 },
 { 0x9EFD, 0xBBFC },
 { 0x9EFF, 0xC6D0 },
 { 0x9F00, 0xCD95 },
 { 0x9F02, 0xC870 },
 { 0x9F04, 0xC8FC },
 { 0x9F09, 0xD0C0 },
 { 0x9F0A, 0xBCBD },
 { 0x9F10, 0xB0B4 },
 { 0x9F17, 0xB3C4 },
 { 0x9F19, 0xBE44 },
 { 0x9F22, 0xBD84 },
 { 0x9F2B, 0xC11D },
 { 0x9F2C, 0xC720 },
 { 0x9F2F, 0xC624 },
 { 0x9F34, 0xC5B8 },
 { 0x9F38, 0xACB8 },
 { 0x9F39, 0xC5B8 },
 { 0x9F3E, 0xD55C },
 { 0x9F54, 0xCE5C },
 { 0x9F55, 0xD758 },
 { 0x9F57, 0xC740 },
 { 0x9F60, 0xCD08 },
 { 0x9F66, 0xAC04 },
 { 0x9F69, 0xAD50 },
 { 0x9F6B, 0xACE4 },
 { 0x9F72, 0xC6B0 },
 { 0x9F76, 0xC545 },
 { 0x9F8E, 0xB871 },
 { 0x9F92, 0xB8E1 },
 { 0x9F94, 0xACF5 },
 { 0x9F9D, 0xCD94 },
 { 0x9FA0, 0xC57D },
 { 0x9FA2, 0xD654 },
 { 0x9FA5, 0xC720 },

};

static int hanja_table_size = sizeof(hanja_to_hangul) / sizeof(ConvTbl);

UINT PASCAL
hanja_search(UINT code)
{
    int min, max,mid;
    /* binary search in table */
    min = 0;
    max = hanja_table_size;
    while (max >= min) {
        mid = (min + max) / 2;
        if (hanja_to_hangul[mid].key < code)
            min = mid + 1;
        else if (hanja_to_hangul[mid].key > code)
            max = mid - 1;
        else
            return hanja_to_hangul[mid].code;
    }
    return 0;
}
